
/////////////////////// Qt includes
#include <QApplication>
#include <QTranslator>
#include <QtGlobal>
#include <QTimer>
#include <QRegExp>


/////////////////////// Std includes
#include <iostream>
#include <string>
#include <algorithm>


////////////////////// Command line stuff include
#include <tclap/CmdLine.h>


/////////////////////// Local includes
#include "config.h"
#include "globals.hpp"
#include "MassSpecSqlite3Handler.hpp"
// #include "MassSpecDataFileLoaderMzml.hpp"
#include "MassSpecDataFileLoaderPwiz.hpp"

using std::cout;
using namespace std;
using namespace TCLAP;


using msXpertSuite::MassSpecSqlite3Handler;
//using msXpertSuite::MassSpecDataFileLoaderMzml;
using msXpertSuite::MassSpecDataFileLoaderPwiz;

void printGreetings();
void printVersion();
void printConfig(const QString & = QString());

bool mzmlFilenameArgHandler(QString optionValue);
bool dbFilenameArgHandler(QString optionValue);
void craftDbFileName(const QString &output);

void printFeedback(QString msg, int type);

QString mzmlFileName;
bool mzmlFileNameOk;

QString dbFileName;
bool dbFileNameOk;


	int
main(int argc, char **argv)
{
	// Command line parsing using tclap.

	QString cmdLineProgUsageMsg = "convXpert is a program that can be used \n"
		"to parse an mzML file and create a database out of its data.\n";

	CmdLine cmdLine(cmdLineProgUsageMsg.toStdString(), ' ', VERSION);

	// license switch
	SwitchArg printLicenseArg("l", "license",
			"Print the license of this software.", cmdLine, false);

	// version switch
	SwitchArg printVersionArg(
			"V", "Version", "Print the contextual version of this software.",
			cmdLine, false);

	// config switch
	SwitchArg printConfigArg("c", "config",
			"Print the configuration of this software.",
			cmdLine, false);

	ValueArg<std::string> mzmlFilenameArg("m", "mzML",
			"Filename of the mzML file.",
			false, "", "file name");

	ValueArg<std::string> dbFilenameArg("b", "db",
			"SQLite database filename or directory where to create the db file.",
			false, "", "file name or directory name");

	cmdLine.add(mzmlFilenameArg);
	cmdLine.add(dbFilenameArg);

	cmdLine.parse(argc, argv);

	QString error = "";
	QString optionStringValue = "";

	// Start working on the various parameters.

	bool isPrintLicense = printLicenseArg.getValue();
	bool isPrintVersion = printVersionArg.getValue();
	bool isPrintConfig  = printConfigArg.getValue();

	// The user might ask for the license to be shown:
	if(isPrintLicense)
	{
		printGreetings();
	}

	// The user might ask for the version to be shown:
	if(isPrintVersion)
	{
		printVersion();
		return 0;
	}

	// The user might ask for the config to be shown:
	if(isPrintConfig)
	{
		printConfig(QString("%1 (run as %2)")
				.arg(CONVXPERT_TARGET)
				.arg(argv[0]));
		return 0;
	}

	// At this point, check if we are willing to convert a mzML file into a SQLite
	// Db file.

	optionStringValue = QString::fromStdString(mzmlFilenameArg.getValue());
	if(!optionStringValue.isEmpty())
	{
		mzmlFileNameOk = mzmlFilenameArgHandler(optionStringValue);
		if(!mzmlFileNameOk)
		{
			qDebug() << __FILE__ << __LINE__
				<< "Failed to handle the mzML file name argument.";
			exit(1);
		}
	}
	else
	{

		qDebug() << __FILE__ << __LINE__
			<< "Error, no mzML filename provided, please type convxpert --help.\n";
		exit(1);
	}

	optionStringValue = QString::fromStdString(dbFilenameArg.getValue());
	dbFileNameOk = dbFilenameArgHandler(optionStringValue);
	if(!dbFileNameOk)
	{
		qDebug() << __FILE__ << __LINE__
			<< "Failed to handle the database file name argument.";
		exit(1);
	}

	// At this point check if the program is started to make an mzML->Db conversion
	// or not.

	if(mzmlFileNameOk && dbFileNameOk)
	{
		// MassSpecDataFileLoaderMzml loader(mzmlFileName);
		MassSpecDataFileLoaderPwiz loader(mzmlFileName);
		loader.m_sqlite3Handler.m_dbFileName = dbFileName;

		// The default feedback mechanism is based on QObject::emit(signal). But we
		// cannot rely on that mechanism here. So we have this peculiar provision:
		loader.m_feedbackToStdout = true;

		bool ok = loader.loadData(dbFileName);
		if(!ok)
		{
			qDebug() << __FILE__ << __LINE__
				<< "Failure";
			return 1;
		}

		qDebug() << __FILE__ << __LINE__
			<< "Success";
		return 0;
	}


	// Qt stuff starts here.
	QCoreApplication application(argc, argv);

	return application.exec();
}


	void
printLicense()
{
	QString version = VERSION;

	QString greetings = QObject::tr("convXpert, version %1\n\n")
		.arg(version);

	greetings += QObject::tr("Type 'convxpert --help' for help\n\n");

	greetings += QObject::tr("convXpert is Copyright 2016 \n"
			"by Filippo Rusconi.\n\n"
			"convXpert comes with ABSOLUTELY NO WARRANTY.\n"
			"convXpert is free software, "
			"covered by the GNU General\n"
			"Public License Version 3 or later, "
			"and you are welcome to change it\n"
			"and/or distribute copies of it under "
			"certain conditions.\n"
			"Check the file COPYING in the distribution.\n"
			"\nHappy convXpert'ing!\n\n");

	cout << greetings.toStdString();

	exit(0);
}


bool mzmlFilenameArgHandler(QString optionValue)
{
	QFileInfo mzmlFileInfo(optionValue);

	if(!mzmlFileInfo.isFile())
	{
		qDebug() << __FILE__ << __LINE__
			<< "File " << optionValue << "does not exist. Exiting.\n";

		exit(1);
	}

	// Now convert the mzml filename into an absolute one
	// if it is not already.

	if(!mzmlFileInfo.isAbsolute())
	{
		mzmlFileName = mzmlFileInfo.absoluteFilePath();
	}
	else
	{
		// Since the name is already absolute, let's set it the way it is, with no
		// change whatsoever.
		mzmlFileName = optionValue;
	}

	// Make sure that the file exists:
	QFile file(mzmlFileName);
	if(!file.exists())
	{
		qDebug() << __FILE__ << __LINE__
			<< "File" << mzmlFileName << "does not exist. Exiting.";

		exit(1);
	}

	return true;
}


	bool
dbFilenameArgHandler(QString optionValue)
{

	// The db file name will be stored in global dbFileName

	craftDbFileName(optionValue);

	// Check if the database file exists.
	QFile dbFile(dbFileName);

	if(dbFile.exists())
	{
		qDebug() << __FILE__ << __LINE__
			<< "File exists already: "
			<< dbFileName;

		qDebug() << __FILE__ << __LINE__
			<< "Should that file be overwritten (y/n)?";

		QFile fileIn;
		fileIn.open(stdin, QIODevice::ReadOnly);
		QByteArray answer = fileIn.readLine();
		fileIn.close();

		if(answer[0] == 'n' || answer[0] == 'N')
		{
			qDebug() << __FILE__ << __LINE__
				<< "Operation aborted by the user.";
			exit(1);
		}
		else
		{
			if(!dbFile.remove())
			{
				qDebug() << __FILE__ << __LINE__
					<< "Failed to remove the database file. Exiting.\n";
				exit(1);
			}
		}
	}
	return true;
}


void craftDbFileName(const QString &output)
{
	if(output.isEmpty())
	{
		// If the output is empty, craft a file name using the mzml file name as
		// template.

		dbFileName = mzmlFileName;
		dbFileName.replace(QRegExp("\\.mzml$"), ".db");

		if(dbFileName == mzmlFileName)
		{
			// This means that the replace did not work out.
			// We do not want to overwrite the original mzml file
			// and thus we craft a brand new file by appending .db.
			dbFileName += ".db";
		}

		//qDebug() << __FILE__ << __LINE__
			//<< "Crafted new db filename on the basis of the mzml name:" << dbFileName;

		return;
	}

	// At this point output might be a directory or a file name.

	QDir dir(output);

	if(dir.exists())
	{
		// The output is a directory. We need to write the database file into it.
		// Craft a db file name using the mzml file name as a template.

		QString	tempDbFileName = mzmlFileName;
		tempDbFileName.replace(QRegExp("\\.mzml$"), ".db");

		if(tempDbFileName == mzmlFileName)
		{
			// This means that the replace did not work out.
			// We do not want to overwrite the original mzml file
			// and thus we craft a brand new file by appending .db.
			tempDbFileName += ".db";
		}

		//qDebug() << __FILE__ << __LINE__
			//<< "Crafted new db filename on the basis of the mzml name:" << tempDbFileName;

		// If the mzml filename was an absolute path filename, we need to only use
		// its filename part, not the path. QFileInfo helps.

		QFileInfo fileInfo(tempDbFileName);

		dbFileName = QString("%1/%2").arg(output).arg(fileInfo.fileName());

		//qDebug() << __FILE__ << __LINE__
			//<< "Crafted new db filename on the basis of the mzml name and dir name:" << dbFileName;

		return;
	}
	else
	{
		// output is the requeste db file name. OK
		dbFileName = output;

		//qDebug() << __FILE__ << __LINE__
			//<< "Provided db file name:" << dbFileName;
	}

	return;
}


	void
printFeedback(QString msg, int type)
{


}


	void
printGreetings()
{
	QString greetings = QObject::tr("convXpert, version %1\n\n").arg(VERSION);

	greetings += "Type 'convxpert --help' for help\n\n";

	greetings += "convXpert is Copyright 2016-2017 \n"
		"by Filippo Rusconi.\n\n"
		"convXpert comes with ABSOLUTELY NO WARRANTY.\n"
		"convXpert is free software, "
		"covered by the GNU General\n"
		"Public License Version 3 or later, "
		"and you are welcome to change it\n"
		"and/or distribute copies of it under "
		"certain conditions.\n"
		"Check the file COPYING in the distribution "
		"and/or the\n"
		"'Help/About(Ctrl+H)' menu item of the program.\n"
		"\nHappy convXpert'ing!\n\n";

	cout << greetings.toStdString();
}


	void
printVersion()
{
	QString version = QObject::tr("convXpert, version %1 -- "
			"Compiled against Qt, version %2\n")
		.arg(VERSION)
		.arg(QT_VERSION_STR);

	cout << version.toStdString();
}


	void
printConfig(const QString &execName)
{
	QString config = QObject::tr("convXpert: "
			"Compiled with the following configuration:\n"
			"EXECUTABLE BINARY FILE: = %1\n"
			"MSXPERTSUITE_BIN_DIR = %2\n"
			"MSXPERTSUITE_DATA_DIR = %3\n"
			"MSXPERTSUITE_DOC_DIR = %5\n")
		.arg(execName)
		.arg(MSXPERTSUITE_BIN_DIR)
		.arg(MSXPERTSUITE_DATA_DIR)
		.arg(MSXPERTSUITE_DOC_DIR);

	cout << config.toStdString();
}
