/*   msXpertSuite - mass spectrometry software suite
 *   -----------------------------------------------
 *   Copyright(C) 2016 Filippo Rusconi
 *
 *   http://www.{site}.org
 *
 *   This file is part of the msXpertSuite project.
 *
 *   The msXpertSuite project is the successor of the massXpert project. This
 *   project now includes various independent modules:
 *   - massXpert, the original mass spectrometry environment software (greatly
 *     refactored code);
 *   - mobXpert, a module to process and mine mobility mass spectrometry data;
 *   - viewXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef ISOTOPE_HPP
#define ISOTOPE_HPP


/////////////////////// Qt includes
///////////////////////
#include <QtGlobal>
#include <QtXml>


namespace msXpertSuite
{

  //! The Isotope class provides an isotope.
  /*! Isotopes are the basic elements of chemical elements(also known
    as atoms). An isotope is characterized by two data:

    \li the isotope mass;

    \li the isotope abundance.

    The carbon(symbol 'C') element, for example is mainly found in
    nature in the form of two isotopes: \f$\mathrm{^{12}C}\f$ and
    \f$\mathrm{^{13}C}\f$ . The \f$\mathrm{^{12}C}\f$ has an abundance
    of 0.99(that is 99 percent) while the heavier isotope has an
    abundance of 0.01(that is 1 percent).

    Consequently, the main member data of this class are the mass value
    and the abundance value.
  */
  class Isotope
  {
    protected:
    //! Mass.
    double m_mass;

    //! Abundance.
    double m_abundance;

    public:
    Isotope(double = 0, double = 0);
    Isotope(const Isotope &);
    virtual ~Isotope()
    {
    }

    virtual Isotope *clone() const;
    virtual void     clone(Isotope *) const;
    virtual void     mold(const Isotope &);
    virtual Isotope &operator=(const Isotope &);

    void    setMass(double);
    double  mass() const;
    QString massString(int decimalPlaces = 10) const;

    void    setAbundance(double);
    void    incrementAbundance(double);
    void    decrementAbundance(double);
    double  abundance() const;
    QString abundanceString(int decimalPlaces = 10) const;

    virtual bool operator==(const Isotope &) const;
    virtual bool operator!=(const Isotope &) const;
    virtual bool operator<(const Isotope &) const;
    static bool  lessThan(const Isotope *, const Isotope *);

    bool     renderXmlIsotopeElement(const QDomElement &, int);
    QString *formatXmlIsotopeElement(int, const QString & = QString("  "));
  };

} // namespace msXpertSuite


#endif // ISOTOPE_HPP
