/*   msXpertSuite - mass spectrometry software suite
 *   -----------------------------------------------
 *   Copyright(C) 2016 Filippo Rusconi
 *
 *   http://www.{site}.org
 *
 *   This file is part of the msXpertSuite project.
 *
 *   The msXpertSuite project is the successor of the massXpert project. This
 *   project now includes various independent modules:
 *   - massXpert, the original mass spectrometry environment software (greatly
 *     refactored code);
 *   - mineXpert, a powerful TIC chromatogram/mass/drift spectrum viewer/miner;
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <QtSql>
#include <QSqlDatabase>
#include <QStringList>
#include <QFile>
#include <QByteArray>

#include "MassSpecDataFileLoaderMzml.hpp"
#include "MassSpecFileMetaData.hpp"
#include "Base64EncDec.hpp"


namespace msXpertSuite
{

	MassSpecDataFileLoaderMzml::MassSpecDataFileLoaderMzml(
			const QString &fileName)
		: MassSpecDataFileLoader{fileName}
	{
		// Store the mzML filename for the metaData table of the database:

		m_msFileMetaData.setStringItem("fileName", fileName);
	}


	MassSpecDataFileLoaderMzml::~MassSpecDataFileLoaderMzml()
	{
		// Do not delete mp_newSpectrum: it might have been transferred *as is* to
		// the mp_massSpecDataSet!!!
		//
		// NO! delete mp_newSpectrum;
	}


	void
		MassSpecDataFileLoaderMzml::setFileName(const QString &fileName)
		{
			m_fileName = fileName;
		}


	void
		MassSpecDataFileLoaderMzml::setMassSpecDataSet(
				MassSpecDataSet *massSpecDataSet)
		{
			if(massSpecDataSet == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			mp_massSpecDataSet = massSpecDataSet;
		}


	void
		MassSpecDataFileLoaderMzml::setParseProgressionNumber(int value)
		{
			m_parseProgressionNumber = value;
		}


	QString
		MassSpecDataFileLoaderMzml::formatDescription() const
		{
			QString description =
				QString("The mzml file format describes mass data with lines\n"
						"formatted according to the mzML format specification.\n");

			return description;
		}


	bool
		MassSpecDataFileLoaderMzml::canLoadData(QString fileName)
		{
			QFile file(fileName);

			if(!file.open(QFile::ReadOnly | QFile::Text))
			{
				qDebug() << __FILE__ << __LINE__ << "Failed to open file" << fileName;

				return false;
			}

			QXmlStreamReader xmlStreamReader;
			xmlStreamReader.setDevice(&file);

			// The mzML files may contain an index or not. Depending on that, the first
			// element varies. We need to take this into account.

			if(xmlStreamReader.readNextStartElement())
			{
				if(xmlStreamReader.name() == "indexedmzML" ||
						xmlStreamReader.name() == "mzML")
				{
					// There, we know this is a mzML file.
					// qDebug() << __FILE__ << __LINE__
					//<< "name:" << xmlStreamReader.name();

					file.close();
					return true;
				}
			}

			return false;
		}


	void
		MassSpecDataFileLoaderMzml::resetSpectralData()
		{
			// Spectral stuff

			// Do not delete mp_newSpectrum, we may not own that spectrum anymore if
			// it has been appended to the MassSpecDataSet instance!

			m_spectrumIndex                    = -1;
			m_msLevel                          = -1;
			m_spectrumTitle                    = "NOT_SET";
			m_specPeakCount                    = -1;
			m_scanCount                        = -1;
			m_tic                              = -1;
			m_scanStartTime                    = -1;
			m_spectrumBinaryDataArrayListCount = -1;

			// Precursor ion stuff
			m_precursorIonId = -1;

			m_precursorMzList.clear();
			m_precursorListCount   = -1;
			m_precursorMz          = -1;
			m_selectedIonListCount = -1;
			m_selectedIonMz        = -1;

			// Scan stuff
			m_scanWindowListCount  = -1;
			m_scanWindowLowerLimit = -1;
			m_scanWindowUpperLimit = -1;
			m_collisionEnergy      = -1;
			m_encodedLength        = -1;
			m_mzQByteArray.clear();
			m_iQByteArray.clear();
		}


	void
		MassSpecDataFileLoaderMzml::resetChromatographicData()
		{
			m_chromatogramTitle.clear();
			m_chromatogramListCount                = -1;
			m_chromatogramIndex                    = -1;
			m_chromatogramBinaryDataArrayListCount = -1;
			m_chromTimeQByteArray.clear();
			m_chromIQByteArray.clear();
		}


	void
		MassSpecDataFileLoaderMzml::resetMobilityData()
		{
			// Mobility stuff
			m_driftDataId = -1;

			m_isMobilityExperiment = false;
			m_curDriftTime         = -1;
			m_driftTimeMin         = -1;
			m_driftTimeMax         = -1;
			m_driftTimeStep        = -1;
			m_prevDriftTime        = -1;
			m_curDriftBin          = 0;
			m_driftBinCount        = 0;
		}


	void
		MassSpecDataFileLoaderMzml::seedMobilityData()
		{

			// Depending on the index of the spectrum we have just worked on, we may
			// want to compute some mobility data for later export to the database.

			// Only deal with mobility data if there are mobility data.
			if(m_curDriftTime == -1)
				return;

			if(m_spectrumIndex == 0)
			{
				m_driftTimeMin = m_curDriftTime;

				// qDebug() << __FILE__ << __LINE__
				// << "First spectrum extraction with minimum drift time: "
				// << m_driftTimeMin;
			}
			else if(m_spectrumIndex == 1)
			{
				m_driftTimeStep = (m_curDriftTime - m_prevDriftTime);

				qDebug()
					<< __FILE__ << __LINE__
					<< "Second spectrum extraction allows calculation of drift time step:"
					<< m_driftTimeStep;
			}
			else
			{
				// qDebug() << __FILE__ << __LINE__
				// << "Spectrum index: " << m_spectrumIndex
				// << "with m_curDriftTime: " << m_curDriftTime
				// << "and m_curDriftBin: " << m_curDriftBin;

				if(m_curDriftTime == m_driftTimeMin)
				{
					// We are starting a new drift cycle because we are not
					// at the 0-index iteration (specIndex != 0) and the
					// drift time is nonetheless the same as the one on the
					// first iteration.

					// We thus can store the drift bin count, that is the number of drift
					// bins in a mobility cycle. This count is set to the current spectrum
					// index that will match the number of spectra before this one that
					// initiates a new drift cycle.

					m_driftBinCount = m_spectrumIndex;
					m_curDriftBin   = 1;

					// qDebug() << __FILE__ << __LINE__
					// << "Encountered a new instance of the m_driftTimeMin"
					// << "Set m_curDriftBin to: " << m_curDriftBin;

					// And also set the max drift time, since we stored that value
					// during the previous iteration.
					m_driftTimeMax = m_prevDriftTime;

					// At this point, we have all the data required, we can thus set the
					// boolean value below to true so that we do not need to reperform
					// this calculation over all the spectra.

					m_isMobilityExperiment = true;

					emit dataLoadMessageSignal(
							"The data are for a mobility experiment.", LOG_TO_BOTH);

					if(m_driftBinCount == 0)
					{
						qFatal("Fatal error at %s@%d. The value of driftBinCount "
								"is 0. "
								"Fatal error. Program aborted.",
								__FILE__, __LINE__);
					}
					else
					{
						// qDebug() << __FILE__ << __LINE__
						// << "The driftBinCount is " << m_driftBinCount;
					}
				}
			}
			// Store the drift time value for the next round, if we need it.
			m_prevDriftTime = m_curDriftTime;
		}


	bool
		MassSpecDataFileLoaderMzml::containsMobilityData()
		{
			// We want to know if this file contains mobility data. Let's go very
			// quickly to the first <spectrum> element where that data should be
			// located.

			QFile mzmlFile(m_fileName);

			if(!mzmlFile.open(QFile::ReadOnly | QFile::Text))
			{
				qDebug() << __FILE__ << __LINE__ << "Failed to open file" << m_fileName;

				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}

			m_xmlStreamReader.setDevice(&mzmlFile);

			// The mzML files may contain an index or not. Depending on that, the
			// first
			// element varies. We need to take this into account.

			m_xmlStreamReader.readNextStartElement();

			if(m_xmlStreamReader.name() == "indexedmzML")
			{
				// qDebug() << __FILE__ << __LINE__
				// << "The mzML file indexed.";

				m_xmlStreamReader.readNextStartElement();
			}
			else if(m_xmlStreamReader.name() == "mzML")
			{
				// qDebug() << __FILE__ << __LINE__
				// << "The mzML file is not indexed.";
			}
			else
			{
				qDebug() << __FILE__ << __LINE__ << "Failed parsing the mzML file.";

				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}

			// At this point we are in the mzML element. Go on.

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() != "run")
				{
					m_xmlStreamReader.skipCurrentElement();
					continue;
				}
				else
					break;
			}

			// Make sure we are actually at the "run" element.
			if(!m_xmlStreamReader.isStartElement() || m_xmlStreamReader.name() != "run")
			{
				qDebug() << __FILE__ << __LINE__ << "Failed parsing the mzML file.";

				return false;
			}

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() != "spectrumList")
				{
					m_xmlStreamReader.skipCurrentElement();
					continue;
				}
				else
					break;
			}

			// Make sure we are actually at the "spectrumList" element.
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "spectrumList")
			{
				qDebug() << __FILE__ << __LINE__ << "Failed parsing the mzML file.";

				return false;
			}

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() != "spectrum")
				{
					m_xmlStreamReader.skipCurrentElement();
					continue;
				}
				else
					break;
			}

			// Make sure we are actually at the "spectrum" element.
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "spectrum")
			{
				qDebug() << __FILE__ << __LINE__ << "Failed parsing the mzML file.";

				return false;
			}

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() != "scanList")
				{
					m_xmlStreamReader.skipCurrentElement();
					continue;
				}
				else
					break;
			}

			// Make sure we are actually at the "scanList" element.
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "scanList")
			{
				qDebug() << __FILE__ << __LINE__ << "Failed parsing the mzML file.";

				return false;
			}

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() != "scan")
				{
					m_xmlStreamReader.skipCurrentElement();
					continue;
				}
				else
					break;
			}


			// Make sure we are actually at the "scan" element.
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "scan")
			{
				qDebug() << __FILE__ << __LINE__ << "Failed parsing the mzML file.";

				return false;
			}

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() != "cvParam")
				{
					m_xmlStreamReader.skipCurrentElement();
					continue;
				}
				else
				{
					QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

					QString cvRef = attributes.value(QString(), "cvRef").toString();
					QString accession =
						attributes.value(QString(), "accession").toString();
					QString name = attributes.value(QString(), "name").toString();

					if(cvRef == "MS" && accession == "MS:1002476" &&
							name == "ion mobility drift time")
					{
						// Yeah, this is a mobility experiment mass data file.
						// At this point we should close the database and the mzML
						// file.
						mzmlFile.close();
						return true;
					}
					else
					{
						m_xmlStreamReader.skipCurrentElement();
						continue;
					}
				}
			}
			// End of
			// while(m_xmlStreamReader.readNextStartElement())

			// At this point we should be confident that the file is not a
			// mobility
			// experiment file.
			mzmlFile.close();
			return false;
		}


	int
		MassSpecDataFileLoaderMzml::readSpectrumCount()
		{
			return spectrumCountNoFunction2();
		}


	int
		MassSpecDataFileLoaderMzml::spectrumCountNoFunction2()
		{
			// We want to know how many spectra are stored in this file.

			QFile mzmlFile(m_fileName);

			if(!mzmlFile.open(QFile::ReadOnly | QFile::Text))
			{
				qDebug() << __FILE__ << __LINE__ << "Failed to open file" << m_fileName;

				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}

			m_xmlStreamReader.setDevice(&mzmlFile);

			// The mzML files may contain an index or not. Depending on that, the
			// first
			// element varies. We need to take this into account.

			m_xmlStreamReader.readNextStartElement();

			if(m_xmlStreamReader.name() == "indexedmzML")
			{
				// qDebug() << __FILE__ << __LINE__
				// << "The mzML file indexed.";

				m_xmlStreamReader.readNextStartElement();
			}
			else if(m_xmlStreamReader.name() == "mzML")
			{
				// qDebug() << __FILE__ << __LINE__
				// << "The mzML file is not indexed.";
			}
			else
			{
				qDebug() << __FILE__ << __LINE__ << "Failed parsing the mzML file.";

				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}

			// At this point we are in the mzML element. Go on.

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() != "run")
				{
					m_xmlStreamReader.skipCurrentElement();
					continue;
				}
				else
					break;
			}

			// Make sure we are actually at the "run" element.
			if(!m_xmlStreamReader.isStartElement() || m_xmlStreamReader.name() != "run")
			{
				qDebug() << __FILE__ << __LINE__ << "Failed parsing the mzML file.";
				return -1;
			}

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() != "spectrumList")
				{
					m_xmlStreamReader.skipCurrentElement();
					continue;
				}
				else
					break;
			}

			// Make sure we are actually at the "spectrumList" element.
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "spectrumList")
			{
				qDebug() << __FILE__ << __LINE__ << "Failed parsing the mzML file.";
				return std::numeric_limits<uint>::max();
			}

			// We do no want to simply read the "count" attribute, as this attribute
			// takes into account also the spurious Function2 spectra recorded by
			// ProteoWizard when dealing with SynaptMS files. Thus we will go through
			// all the <spectrum> elements, and stop counting at the first spurious
			// spectrum.
			//
			// QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

			// QString spectrumCount = attributes.value(QString(),
			// "count").toString();
			// bool ok;
			// int count = -1;
			// count = spectrumCount.toInt(&ok);

			// if(!ok)
			// qDebug() << __FILE__ << __LINE__
			// << "Error converting " << spectrumCount << "to integer.";

			int count = 0;

			// Now start to run through all the spectrum elements.
			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() == "spectrum")
				{
					QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

					// this is the format of the element.
					// <spectrum index="100001" id="function=2 process=0 scan=2"
					// defaultArrayLength="0">

					// So we want to check if the id attribute contains the
					// "function2"
					// substring.

					QString value = attributes.value(QString(), "id").toString();
					if(value.isEmpty())
						qFatal("Fatal error at %s@%d. Program aborted.", __FILE__,
								__LINE__);

					if(value.contains("function=2"))
					{
						return count;
					}
					else
					{
						++count;
						m_xmlStreamReader.skipCurrentElement();
					}
				}
				else
					break;
			}

			m_spectrumCount = count;

			return count;
		}


	int
		MassSpecDataFileLoaderMzml::loadSpectrumToMassSpecDataSet()
		{
			// At this point, there are two possibilities: either the file
			// is being read in full mode, the we append the spectrum to
			// the mass spec data set and thus we occupy more memory at
			// each read spectrum. Or we are reading the file in streamed
			// mode, in which case, we need to process the data a
			// bit and then let them go.

			// We will need the mass spectrum that was parsed lastly, right before
			// calling this function. The pointer to that spectrum cannot be
			// nullptr.

			if(mp_newSpectrum == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			if(mp_massSpecDataSet->m_isStreamed)
			{

				// We are reading in streamed mode. This means that we are not
				// loading
				// all the data in memory. Instead, we look at each new spectrum
				// and we
				// compute its TIC which we store as a (rt,tic) pair in the
				// mp_massSpecDataSet->m_ticChromMap. This is the chance we have
				// to
				// compute the TIC for the spectrum:

				double newTic = mp_newSpectrum->tic();

				// Had already another spectrum have the same rt ? If so another
				// TIC was
				// already computed.

				double oldTic =
					mp_massSpecDataSet->m_ticChromMap.value(m_scanStartTime, qSNaN());

				if(!qIsNaN(oldTic))
				{

					// A rt by the same value alreay existed, update the
					// tic value and insert it, that will erase the old
					// value, because the map is NOT a multimap.

					newTic += oldTic;
				}

				// And now store the key/value pair for later to use to
				// plot the TIC chromatogram.

				mp_massSpecDataSet->m_ticChromMap.insert(m_scanStartTime, newTic);

				// At this point, handle the case of the colorMap that
				// relates all the dt with the corresponding (mz,i) pairs,
				// that is, each dt has a single mass spectrum that
				// corresponds to the combination of all the spectra having
				// that dt.

				if(m_curDriftTime != -1)
				{
					// qDebug() << __FILE__ << __LINE__
					// << "This file is from a mobility experiment."
					// << "m_curDriftBin: " << m_curDriftBin
					// << "m_curDriftTime:" << m_curDriftTime;

					// Let's see if for the current drift time, a spectrum
					// was already encountered (and possibly undergoing
					// combinations).

					MassSpectrum *massSpectrum =
						mp_massSpecDataSet->m_dtHash.value(m_curDriftTime, nullptr);

					if(massSpectrum == nullptr)
					{
						MassSpectrum *newSpectrum = new MassSpectrum;

						// This is the first spectrum that we set to the hash that has the
						// current dt value. We want to combine all the spectra that have
						// the same dt.  We need to store the spectrum so that we can later
						// refer to it, to fill in the colormap. But remember, we want to
						// make a calculation such that there are less MassPeak elements in
						// the spectra. So we need to allocated a new spectrum and combine
						// with rounding (0 decimals) the newSpectrum in it.

						newSpectrum->m_requiredDecimals = 0; // See comment above.
						newSpectrum->combine(*mp_newSpectrum);
						mp_massSpecDataSet->m_massSpectra.append(newSpectrum);

						// Delete the newSpectrum that we do not use anymore, since we
						// already computed the TIC value and also made the color map point.

						delete mp_newSpectrum;
						mp_newSpectrum = nullptr;

						// And now make the hash association with the current
						// drift time
						// value. This is the "degraded" resolution spectrum
						// because we
						// want to compute a low resolution map.

						mp_massSpecDataSet->m_dtHash.insert(m_curDriftTime,
								newSpectrum);

						// Finally, let's store the first mz value and the last
						// one.
						// Only if the values need an update.

						double mz_first = newSpectrum->first()->m_mz;
						if(mz_first < mp_massSpecDataSet->m_mzRangeStart)
							mp_massSpecDataSet->m_mzRangeStart = mz_first;

						double mz_last = newSpectrum->last()->m_mz;
						if(mz_last > mp_massSpecDataSet->m_mzRangeEnd)
							mp_massSpecDataSet->m_mzRangeEnd = mz_last;

						// We will need to know what is the number of cells on the
						// mz
						// axis of the colormap matrix.

						int spectrumSize = newSpectrum->size();
						if(spectrumSize > mp_massSpecDataSet->m_mzCellCount)
							mp_massSpecDataSet->m_mzCellCount = spectrumSize;
					}
					// End of
					// if(massSpectrum == nullptr)
					else
					{

						// We already have inserted a mass spectrum by the current dt.  So
						// use it to combine the newly generated one. Remember we want to do
						// a 0-decimals combination so as to have a low-resolution map.

						massSpectrum->m_requiredDecimals = 0;
						massSpectrum->combine(*mp_newSpectrum);

						// Since we do not use this spectrum anymore, we can
						// delete it.
						delete mp_newSpectrum;
						mp_newSpectrum = nullptr;

						// Finally, let's store the first mz value and the last
						// one.
						// Only if the values need an update.

						double mz_first = massSpectrum->first()->m_mz;
						if(mz_first < mp_massSpecDataSet->m_mzRangeStart)
							mp_massSpecDataSet->m_mzRangeStart = mz_first;

						double mz_last = massSpectrum->last()->m_mz;
						if(mz_last > mp_massSpecDataSet->m_mzRangeEnd)
							mp_massSpecDataSet->m_mzRangeEnd = mz_last;

						// We will need to know what is the number of cells
						// on the mz
						// axis of the colormap matrix.
						int spectrumSize = massSpectrum->size();
						if(spectrumSize > mp_massSpecDataSet->m_mzCellCount)
							mp_massSpecDataSet->m_mzCellCount = spectrumSize;
					}
				}
				// End of
				// if(m_curDriftTime != -1)
			}
			// End of
			// if(mp_massSpecDataSet->m_isStreamed)
			else
			{

				// We are loading the data in full. Just store the spectrum
				// after having set to it the rt and dt values.

				// qDebug() << __FILE__ << __LINE__
				// << "Appending new mass spectrum to the data set with
				// retention time:" << m_scanStartTime;

				mp_massSpecDataSet->appendMassSpectrum(mp_newSpectrum);
				mp_massSpecDataSet->insertRtHash(m_scanStartTime, mp_newSpectrum);

				//// FIXME Only for debugging purposes:
				//double tic = mp_newSpectrum->tic();

				//qDebug() << __FILE__ << __LINE__
				//<< "TIC value:" << tic;

				// Do not forget to set that retention time to the spectrum itself !

				mp_newSpectrum->m_rt = m_scanStartTime;

				// If the spectrum is from a drift time experiment, then store that data
				// also.

				if(m_curDriftTime != -1)
				{
					// qDebug() << __FILE__ << __LINE__
					// << "This file is from a mobility experiment."
					// << "m_curDriftBin: " << m_curDriftBin
					// << "m_curDriftTime:" << m_curDriftTime;

					mp_massSpecDataSet->insertDtHash(m_curDriftTime, mp_newSpectrum);

					// Do not forget to set that drift time to the spectrum itself !

					mp_newSpectrum->m_dt = m_curDriftTime;
				}

				// Do not destroy the spectrum: as we append it, its owner changes.
			}

			return 1;
		}


	int
		MassSpecDataFileLoaderMzml::loadSpectrumToVectors()
		{

			// At this point, we do have the mass spectrum in the form of
			// &m_mzQByteArray and &m_iQByteArray. Depending on the integration type
			// that was asked we'll do differentiated processing with these.

			if(m_integrationType == IntegrationType::ANY_TO_RT)
			{
				// We want to compute a TIC chrom.
				if(!integrateToRt())
					return -1;
			}
			else if(m_integrationType == IntegrationType::ANY_TO_MZ)
			{
				// We want to compute a mass spectrum.
				if(!integrateToMz())
					return -1;
			}
			else if(m_integrationType == IntegrationType::ANY_TO_DT)
			{
				if(m_curDriftTime == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

				// We want to compute a drift spectrum.
				if(!integrateToDt())
					return -1;
			}
			else
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			return 1;
		}


	int
		MassSpecDataFileLoaderMzml::integrateToRt()
		{

			// We are creating a TIC chromatogram. We may have a history to take into
			// account.

			// RT integration.
			// ===============
			bool   rtIntegration = false;
			double rtStart       = -1;
			double rtEnd         = -1;

			rtIntegration = m_history.innermostRtRange(&rtStart, &rtEnd);
			if(rtIntegration)
			{
				// qDebug() << __FILE__ << __LINE__ << "rt integration accounted for:"
				//<< QString("[%1-%2]\n").arg(rtStart).arg(rtEnd) << "\n";
			}

			// MZ integration.
			// ===============
			bool   mzIntegration = false;
			double mzStart       = -1;
			double mzEnd         = -1;

			mzIntegration = m_history.innermostMzRange(&mzStart, &mzEnd);
			if(mzIntegration)
			{
				// qDebug() << __FILE__ << __LINE__ << "mz integration accounted for:"
				//<< QString("[%1-%2]\n").arg(mzStart).arg(mzEnd) << "\n";
			}

			// DT integration.
			// ===============
			bool   dtIntegration = false;
			double dtStart       = -1;
			double dtEnd         = -1;

			// If the history is comprised of many successive integrations on the
			// basis of dt value ranges, then we should make sure we do take into
			// account the more internal dt value range.
			dtIntegration = m_history.innermostDtRange(&dtStart, &dtEnd);
			if(dtIntegration)
			{
				// qDebug() << __FILE__ << __LINE__ << "dt integration accounted for:"
				//<< QString("[%1-%2]\n").arg(dtStart).arg(dtEnd) << "\n";
			}

			// We just have parsed a spectrum and we have all there relevant data
			// in the
			// various data members.

			// We can immediately check for rt and dt integration:
			if(rtIntegration)
			{
				if(m_scanStartTime < rtStart || m_scanStartTime > rtEnd)
					return 0;
			}
			if(dtIntegration)
			{
				if(m_curDriftTime < dtStart || m_curDriftTime > dtEnd)
					return 0;
			}

			// At this point we know that we'll need the data:
			MassSpectrum *mp_newSpectrum = MassSpectrum::createFromBase64EncByteArrays(
					&m_mzQByteArray, &m_iQByteArray, m_dataCompression);

			if(mp_newSpectrum == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			double newTic = 0;

			// Now the more specific mz integration
			if(mzIntegration)
			{
				// We need to integrate the intensity of the various mass peaks of
				// the
				// spectrum into the TIC, but only for the mz integration range.

				newTic = mp_newSpectrum->tic(mzStart, mzEnd);
			}
			else
				newTic = mp_newSpectrum->tic();

			// Now check if there was already a mass spectrum that was processed with
			// the same m_scanStartTime (which should occur if the mass spectra were
			// acquired for a mobility experiment).

			double oldTic = m_keyValMap.value(m_scanStartTime, qSNaN());

			if(!qIsNaN(oldTic))
			{

				// There was a same retention time processed already. Simply make the
				// intensity increment.

				newTic += oldTic;
			}

			// Now, either overwrite the previous pair (not a QMultiMap) with the
			// incremented tic or simply insert the new one.

			// qDebug() << __FILE__ << __LINE__
			//<< "Inserting new m_keyValMap pair:" << m_scanStartTime << "-"
			//<< newTic;

			m_keyValMap.insert(m_scanStartTime, newTic);

			return 1;
		}


	int
		MassSpecDataFileLoaderMzml::integrateToMz()
		{

			// We are creating a mass spectrum. We may have a history to take into
			// account.

			// RT integration.
			// ===============
			bool   rtIntegration = false;
			double rtStart       = -1;
			double rtEnd         = -1;

			rtIntegration = m_history.innermostRtRange(&rtStart, &rtEnd);
			if(rtIntegration)
			{
				// qDebug() << __FILE__ << __LINE__ << "rt integration accounted for:
				// \n"
				//<< QString("[%1-%2]\n").arg(rtStart).arg(rtEnd);
			}

			// MZ integration.
			// ===============
			bool   mzIntegration = false;
			double mzStart       = -1;
			double mzEnd         = -1;

			mzIntegration = m_history.innermostMzRange(&mzStart, &mzEnd);
			if(mzIntegration)
			{
				// qDebug() << __FILE__ << __LINE__ << "mz integration accounted for:
				// \n"
				//<< QString("[%1-%2]\n").arg(mzStart).arg(mzEnd);
			}

			// DT integration.
			// ===============
			bool   dtIntegration = false;
			double dtStart       = -1;
			double dtEnd         = -1;

			// If the history is comprised of many successive integrations on the
			// basis of dt value ranges, then we should make sure we do take into
			// account the more internal dt value range.
			dtIntegration = m_history.innermostDtRange(&dtStart, &dtEnd);
			if(dtIntegration)
			{
				// qDebug() << __FILE__ << __LINE__ << "dt integration accounted for:
				// \n"
				//<< QString("[%1-%2]\n").arg(dtStart).arg(dtEnd);
			}

			// We just have parsed a spectrum and we have all there relevant data in the
			// various data members.

			// We can immediately check for rt and dt integration:

			if(rtIntegration)
			{
				if(m_scanStartTime < rtStart || m_scanStartTime > rtEnd)
				{
					// qDebug() << __FILE__ << __LINE__ << "returning because "
					//"m_scanStartTime < rtStart || "
					//"m_scanStartTime > rtEnd";

					return 0;
				}
			}
			if(dtIntegration)
			{
				if(m_curDriftTime < dtStart || m_curDriftTime > dtEnd)
				{
					// qDebug() << __FILE__ << __LINE__ << "returning because "
					//"m_curDriftTime < dtStart || "
					//"m_curDriftTime > dtEnd";

					return 0;
				}
			}

			// At this point we know that we'll need the data:
			MassSpectrum *mp_newSpectrum = MassSpectrum::createFromBase64EncByteArrays(
					&m_mzQByteArray, &m_iQByteArray, m_dataCompression);

			if(mp_newSpectrum == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// qDebug() << __FILE__ << __LINE__
			//<< "mp_newSpectrum has size:" << mp_newSpectrum->size();

			// Now the more specific mz integration
			if(mzIntegration)
			{
				m_massSpectrum.combine(*mp_newSpectrum, mzStart, mzEnd);
			}
			else
				m_massSpectrum.combine(*mp_newSpectrum);

			// qDebug() << __FILE__ << __LINE__
			//<< "m_massSpectrum has size:" << m_massSpectrum.size();

			return 1;
		}


	int
		MassSpecDataFileLoaderMzml::integrateToDt()
		{

			// We are creating a drift spectrum. We may have a history to take into
			// account.

			// RT integration.
			// ===============
			bool   rtIntegration = false;
			double rtStart       = -1;
			double rtEnd         = -1;

			rtIntegration = m_history.innermostRtRange(&rtStart, &rtEnd);
			if(rtIntegration)
			{
				//qDebug() << __FILE__ << __LINE__ << "rt integration accounted for:"
				//<< QString("[%1-%2]\n").arg(rtStart).arg(rtEnd) << "\n";

				//qDebug() << __FILE__ << __LINE__
				//<< "m_scanStartTime:" << m_scanStartTime
				//<< "rtStart:" << rtStart << "rtEnd:" << rtEnd;
			}

			// MZ integration.
			// ===============
			bool   mzIntegration = false;
			double mzStart       = -1;
			double mzEnd         = -1;

			mzIntegration = m_history.innermostMzRange(&mzStart, &mzEnd);
			if(mzIntegration)
			{
				// qDebug() << __FILE__ << __LINE__ << "mz integration accounted for:"
				//<< QString("[%1-%2]\n").arg(mzStart).arg(mzEnd) << "\n";

				// qDebug() << __FILE__ << __LINE__ << "mzStart:" << mzStart
				//<< "mzEnd:" << mzEnd;
			}

			// DT integration.
			// ===============
			bool   dtIntegration = false;
			double dtStart       = -1;
			double dtEnd         = -1;

			dtIntegration = m_history.innermostDtRange(&dtStart, &dtEnd);
			if(dtIntegration)
			{
				// qDebug() << __FILE__ << __LINE__ << "dt integration accounted for:"
				//<< QString("[%1-%2]\n").arg(dtStart).arg(dtEnd) << "\n";

				// qDebug() << __FILE__ << __LINE__ << "m_curDriftTime:" <<
				// m_curDriftTime
				//<< "dtStart:" << dtStart << "dtEnd:" << dtEnd;
			}

			// Now that we have all the start/end pairs, we can deal with the spectrum
			// that we have just parsed.

			// We can immediately check for rt and dt integration:
			if(rtIntegration)
			{
				if(m_scanStartTime < rtStart || m_scanStartTime > rtEnd)
				{
					// qDebug() << __FILE__ << __LINE__
					//<< "returning because m_scanStartTime "
					//"< rtStart || m_scanStartTime > "
					//"rtEnd";
					return 0;
				}
				else
				{
					// qDebug() << __FILE__ << __LINE__ << "Correct rt integration
					// value:"
					//<< "m_scanStartTime:" << m_scanStartTime
					//<< "rtStart:" << rtStart << "rtEnd:" << rtEnd;
				}
			}
			if(dtIntegration)
			{
				if(m_curDriftTime < dtStart || m_curDriftTime > dtEnd)
				{
					// qDebug() << __FILE__ << __LINE__
					//<< "returning because m_curDriftTime "
					//"< dtStart || m_curDriftTime > "
					//"dtEnd";
					return 0;
				}
				else
				{
					// qDebug() << __FILE__ << __LINE__ << "Correct dt integration
					// value:"
					//<< "m_curDriftTime:" << m_curDriftTime
					//<< "dtStart:" << dtStart << "dtEnd:" << dtEnd;
				}
			}

			// At this point we know that we'll need the data:

			MassSpectrum *newSpectrum = MassSpectrum::createFromBase64EncByteArrays(
					&m_mzQByteArray, &m_iQByteArray, m_dataCompression);

			if(newSpectrum == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// qDebug() << __FILE__ << __LINE__ << "for m_curDriftTime:" <<
			// m_curDriftTime
			//<< "newSpectrum size:" << newSpectrum->size();

			double newTic = 0;

			// Now the more specific mz integration
			if(mzIntegration)
			{

				// We need to integrate the intensity of the various mass peaks of the
				// spectrum into the TIC, but only for the mz integration range.

				newTic = newSpectrum->tic(mzStart, mzEnd);
			}
			else
				newTic = newSpectrum->tic();

			// Now check if there was already a mass spectrum that was processed with
			// the same m_curDriftTime (which should occur if the mass spectra were
			// acquired for a mobility experiment and there was more than one mobility
			// cycle acquisition.

			double oldTic = m_keyValMap.value(m_curDriftTime, qSNaN());

			if(!qIsNaN(oldTic))
			{

				// There was a same retention time processed already. Simply make the
				// intensity increment.

				// qDebug() << __FILE__ << __LINE__
				//<< "There was already a tic for drift time: "
				//<< m_curDriftTime;

				newTic += oldTic;
			}

			// Now, either overwrite the previous pair (not a QMultiMap) with the
			// incremented tic or simply insert the new one.

			m_keyValMap.insert(m_curDriftTime, newTic);

			// qDebug() << __FILE__ << __LINE__
			//<< "After insertion new m_keyValMap pair:" << m_curDriftTime << "-"
			//<< newTic << "New size of the map:" << m_keyValMap.size();

			return 1;
		}


	void
		MassSpecDataFileLoaderMzml::parseMzmlFile()
		{

			// If this object is tasked with more than one parsing, then, make sure we
			// start fresh.

			resetMobilityData();
			resetSpectralData();

			m_parseProgression = 0;
			m_errorMessageList.clear();

			// qDebug() << __FILE__ << __LINE__
			// << "Entering function parseMzmlFile";

			// Parse the mzML file so as to inject the data into the database file
			// handled by msSqlFileHandler parameter.

			QFile mzmlFile(m_fileName);

			if(!mzmlFile.open(QFile::ReadOnly | QFile::Text))
			{
				qDebug() << __FILE__ << __LINE__ << "Failed to open file" << m_fileName;

				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}

			// Will set the value to m_spectrumCount only if it has not been set
			// already. The intialization value at instance creation is set to 
			// std::numeric_limits<uint>::max()

			if(m_spectrumCount == std::numeric_limits<uint>::max())
				readSpectrumCount();

			// readSpectrumCount() returns std::numeric_limits<uint>::max() on error.
			if(m_spectrumCount == std::numeric_limits<uint>::max())
				qFatal("Fatal error at %s@%d. Failed to read the number of spectra. Program aborted.",
						__FILE__, __LINE__);

			emit setupProgressFeedbackSignal("Parsing mzML file", 1, m_spectrumCount);

			m_xmlStreamReader.setDevice(&mzmlFile);

			// The mzML files may contain an index or not. Depending on that, the
			// first element varies. We need to take this into account.

			m_xmlStreamReader.readNextStartElement();

			if(m_xmlStreamReader.name() == "indexedmzML")
			{
				// qDebug() << __FILE__ << __LINE__
				// << "The mzML file indexed.";

				parse_indexedmzML_element();
			}
			else if(m_xmlStreamReader.name() == "mzML")
			{
				// qDebug() << __FILE__ << __LINE__
				// << "The mzML file is not indexed.";

				parse_mzML_element();
			}
			else
			{
				qDebug() << __FILE__ << __LINE__ << "Failed parsing the mzML file.";

				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}

			// At this point we should close the mzML file.
			mzmlFile.close();
		}


	int
		MassSpecDataFileLoaderMzml::loadData(MassSpecDataSet *massSpecDataSet)
		{
			mp_massSpecDataSet = massSpecDataSet;

			parseMzmlFile();

			// Reset the local copy of the pointer.
			mp_massSpecDataSet = nullptr;

			return massSpecDataSet->m_massSpectra.size();
		}


	int
		MassSpecDataFileLoaderMzml::loadData(const QString &dbFileName)
		{
			if(dbFileName.isEmpty())
				return -1;

			m_sqlite3Handler.m_dbFileName = dbFileName;

			m_sqlite3Handler.initializeDbForWrite();

			parseMzmlFile();

			return 1;
		}


	int
		MassSpecDataFileLoaderMzml::loadData(const History &  history,
				IntegrationType  integrationType,
				QVector<double> *keyVector,
				QVector<double> *valVector)
		{
			if(keyVector == nullptr || valVector == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			m_history         = history;
			m_integrationType = integrationType;

			// Remember to reset that bool value to false later.
			m_isLoadToVectors = true;

			// Will set the value to m_spectrumCount only if it has not been set
			// already. The intialization value at instance creation is set to 
			// std::numeric_limits<uint>::max().

			if(m_spectrumCount == std::numeric_limits<uint>::max())
				readSpectrumCount();

			// readSpectrumCount() returns std::numeric_limits<uint>::max() on error.
			if(m_spectrumCount == std::numeric_limits<uint>::max())
				qFatal("Fatal error at %s@%d. Failed to read the number of spectra. Program aborted.",
						__FILE__, __LINE__);

			parseMzmlFile();


			// Reset this variable for another loading use later.
			m_isLoadToVectors = false;


			// Depending on the integration type, the data are now in the form of a
			// m_massSpectrum that is the result of the combination of all the parsed
			// spectra (to MZ integration) or in the form of a m_keyValMap where the
			// relevant data were stored (to DT integration).

			if(m_integrationType == IntegrationType::ANY_TO_MZ)
			{

				// At this point all the data have been integrated into m_massSpectrum.
				// So we need to convert that to vectors:

				m_massSpectrum.toVectors(keyVector, valVector);
			}
			else if(m_integrationType == IntegrationType::ANY_TO_DT)
			{

				// The data were stored in m_keyValMap in the form of (dt, tic) pairs.

				QList<double> keyList = m_keyValMap.keys();

				for(int iter = 0; iter < keyList.size(); ++iter)
				{
					double key = keyList.at(iter);
					keyVector->append(key);
					valVector->append(m_keyValMap.value(key));
				}
			}
			else if(m_integrationType == IntegrationType::ANY_TO_RT)
			{

				// The data were stored in m_keyValMap in the form of (rt, tic) pairs.

				QList<double> keyList = m_keyValMap.keys();

				for(int iter = 0; iter < keyList.size(); ++iter)
				{
					double key = keyList.at(iter);
					keyVector->append(key);
					valVector->append(m_keyValMap.value(key));
				}
			}

			if(!keyVector->size())
				qDebug() << __FILE__ << __LINE__
					<< "The key vector is empty. Is not that worrisome?";

			// Sanity check:
			if(keyVector->size() != valVector->size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			m_history.freeList();
			m_integrationType = IntegrationType::NOT_SET;

			return keyVector->size();
		}


	int
		MassSpecDataFileLoaderMzml::streamedIntegration(
				const History &history, IntegrationType integrationType,
				QVector<double> *keyVector, QVector<double> *valVector)
		{
			if(!history.isValid())
			{
				qDebug() << __FILE__ << __LINE__
					<< "The History is not valid. Returning.";

				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

				return -1;
			}

			if(integrationType != IntegrationType::ANY_TO_MZ &&
					integrationType != IntegrationType::ANY_TO_DT &&
					integrationType != IntegrationType::ANY_TO_RT)

			{
				qDebug() << __FILE__ << __LINE__ << "No valid integration type is "
					"defined for the data to be "
					"produced. Returning.";
				return -1;
			}

			int result = loadData(history, integrationType, keyVector, valVector);

			// loadData manages to clean up the member history and integration type and
			// also the key/val member vectors.

			return result;
		}


	void
		MassSpecDataFileLoaderMzml::parse_indexedmzML_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "indexedmzML")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We have nothing to do here, we want to go to the <mzML> element.
			m_xmlStreamReader.readNextStartElement();

			if(m_xmlStreamReader.name() != "mzML")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			parse_mzML_element();

			// Let's check if we are not loosing cpu on spurious spectra:
			if(m_wasFunction2Spectrum)
				m_errorMessageList.append(
						"Encountered Function 2 spectra, parsing of the file stopped.\n");

			QString msg = QString("Total number of parsed spectra: %1\n"
					"Total number of stored spectra : %2\n"
					"Total number of skipped spectra: %3\n"
					"Total number of parsed chromatograms: %4\n"
					"The following errors were encountered: %5\n")
				.arg(m_spectrumIndex + 1)
				.arg(m_storedSpectraCount)
				.arg(m_totalSkippedSpectra)
				.arg(m_storedChromatogramCount)
				.arg(m_errorMessageList.join('\n'));

			emit dataLoadMessageSignal(msg, LOG_TO_CONSOLE);

			// We may also use this function with a terminal-only program, so it
			// might
			// be asked that we send the feedback to the terminal Stdout.
			if(m_feedbackToStdout)
				printf("%s\n", msg.toLatin1().data());

			// qDebug() << __FILE__ << __LINE__ << "\n"
			// << "Total number of parsed spectra : " << m_spectrumIndex + 1 <<
			// "\n"
			// << "Total number of stored spectra : " << m_storedSpectraCount <<
			// "\n"
			// << "Total number of skipped spectra: " << m_totalSkippedSpectra <<
			// "\n"
			// << "Total number of parsed chromatograms: " <<
			// m_storedChromatogramCount
			// << "\n"
			// << "The following errors were encountered:\n" << m_errorMessageList
			// <<
			// "\n";
		}


	void
		MassSpecDataFileLoaderMzml::parse_mzML_element()
		{
			// Now starts the real interesting parsing: we want to rush to the
			// <run>
			// element.
			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() == "run")
					parse_run_element();
				else
				{
					// qDebug() << __FILE__ << __LINE__
					// << "Parsing element: " <<
					// m_xmlStreamReader.name().toString();

					m_xmlStreamReader.skipCurrentElement();
				}

				// Let's check if we are not loosing cpu on spurious spectra:
				if(m_wasFunction2Spectrum)
				{
					return;
				}
			}
		}


	void
		MassSpecDataFileLoaderMzml::parse_run_element()
		{
			if(!m_xmlStreamReader.isStartElement() || m_xmlStreamReader.name() != "run")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// Get the start time stamp that is useful for us.
			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

			// This acquisition date:time datum is essential to record data later (if
			// required) to a SQLite3-formatted file.
			QDateTime dateTime;

			QString startTimeStamp =
				attributes.value(QString(), "startTimeStamp").toString();

			if(startTimeStamp.isEmpty())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// Format of the startTimeStamp: "2016-02-22T10:29:15Z"
			dateTime = QDateTime::fromString(startTimeStamp, Qt::ISODate);

			if(!dateTime.isValid())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			m_msFileMetaData.setStringItem("acquisitionDate",
					dateTime.toString("yyyy.MM.dd"));


			// Now that we got the date, go to the <spectrumList> element.

			while(m_xmlStreamReader.readNextStartElement())
			{
				// qDebug() << __FILE__ << __LINE__ << "in the run loop:"
				//<< "now in element: " << m_xmlStreamReader.name()
				//<< "now in token: " << m_xmlStreamReader.tokenString();

				if(m_xmlStreamReader.name() == "spectrumList")
					parse_spectrumList_element();
				else if(m_xmlStreamReader.name() == "chromatogramList")
					parse_chromatogramList_element();
				else
				{
					// qDebug() << __FILE__ << __LINE__ << "Skippping element "
					//<< m_xmlStreamReader.name().toString();

					m_xmlStreamReader.skipCurrentElement();
				}
			}
		}


	void
		MassSpecDataFileLoaderMzml::parse_spectrumList_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "spectrumList")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

#if 0
			// We do not care of this value anymore, because we know that when the user
			// asks that the mobility data be saved in the raw file, upon conversion of
			// that file into mzML by Proteowizard, there are two sets of spectra store in
			// the mzML file: the Function 1 spectra, correct ones, real ones ; the
			// Function 2 spectra that are totally useless, with empty <binary></binary>
			// elements and 'nan' drift time values. Thus, we decided to store in the
			// database the number of actual spectra stored in it.
			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

			QString value = attributes.value(QString(), "count").toString();
			if(value.isEmpty())
				qFatal("Fatal error at %s@%d. Program aborted.",
						__FILE__, __LINE__);

			int count = value.toInt(&ok, 10 /*base*/);
			if(!ok)
				qFatal("Fatal error at %s@%d. Program aborted.",
						__FILE__, __LINE__);
#endif

			// Now go to the innumerous <spectrum> elements...

			// Store some previous round variables, that we'll need in case we are
			// parsing mobility drift data.

			mp_newSpectrum = nullptr;

			int count = 0;

			while(m_xmlStreamReader.readNextStartElement())
			{
				// qDebug() << __FILE__ << __LINE__ << "in the run loop:"
				//<< "now in element: " << m_xmlStreamReader.name()
				//<< "now in token: " << m_xmlStreamReader.tokenString();

				// If the user asks for abortion of the process, then do that.
				if(m_isOperationCancelled)
				{
					qDebug() << __FILE__ << __LINE__ << "Breaking the while loop "
						"due to abort requested "
						"by the user.";

					break;
				}

				if(m_xmlStreamReader.name() == "spectrum")
				{
					// qDebug() << __FILE__ << __LINE__ << "Entering the "
					//"m_xmlStreamReader.name()"
					//" == spectrum if block";

					// A new spectrum necessarily corresponds to a new drift bin.
					m_curDriftBin++;

					// Actually parse this "spectrum" element.

					parse_spectrum_element();

					// Let's check if we are not loosing cpu on spurious spectra:
					if(m_wasFunction2Spectrum)
					{

						// We are parsing a spectrum of Funtion 2, we do not want
						// to
						// store these data. And we also want to skip all
						// remaining
						// fuspectra, which we know belong also to Function 2.
						// However,
						// we cannot stop the processing of the file here because
						// we
						// want to go to the <chromatogramList> element.

						// At this precise moment we are in <spectrum>. If we skip
						// the
						// current element, we should be back to the
						// <spectrumList>
						// element.  When we return from this function, the caller
						// function will call next start element and we'll be in
						// <chromatogramList>

						// qDebug() << __FILE__ << __LINE__
						// << "Detected function2 spectrum, "
						// << "now in element: " << m_xmlStreamReader.name()
						// << "now in token: " << m_xmlStreamReader.tokenString()
						// << "Finally, now skip current element.";

						QString msg = QString("Skipping Function2 elements, running to "
								"the chromatogram element.\n");
						emit dataLoadMessageSignal(msg, LOG_TO_BOTH);

						m_xmlStreamReader.skipCurrentElement();

						// qDebug() << __FILE__ << __LINE__
						// << "right after skipping current element"
						// << "now in element: " << m_xmlStreamReader.name()
						// << "now in token: " << m_xmlStreamReader.tokenString()
						// << "Finally, now breaking the loop.";

						// An now we can break.
						break;
					}

					// Maintain a count of the parsed spectra, so that we can give
					// the
					// user a feedback each 100 spectra.

					m_parseProgression++;

					// qDebug() << __FILE__ << __LINE__ << "m_parseProgression "
					//<< m_parseProgression << "and "
					//<< "m_parseProgressionNumber" << m_parseProgressionNumber;

					if(m_parseProgression > m_parseProgressionNumber)
					{
						m_parseProgression = 1;

						QString msg = QString("Parsing spectrum number %1 -- "
								"skipped empty spectra : %2\n")
							.arg(m_spectrumIndex + 1)
							.arg(m_totalSkippedSpectra);

						emit dataLoadMessageSignal(msg, LOG_TO_BOTH);
						emit updateProgressFeedbackSignal(msg, m_spectrumIndex + 1);

						// We may also use this function with a terminal-only program,
						// so it might be asked that we send the feedback to the
						// terminal Stdout.

						if(m_feedbackToStdout)
							printf("%s@%d: %s\n", __FILE__, __LINE__,
									msg.toLatin1().data());

						// qDebug() << __FILE__ << __LINE__
						//<< msg;
					}

					// Increment the count, and then emit a signal to inform this
					// function user that we have parsed one more spectrum. This is used
					// by the data loader users to update a progress bar dialog.

					++count;

					emit dataLoadProgressSignal(count);

					// Now that we have parsed the element, we should know its index and
					// thus we can compute some stuff related to mobility drift times.

					if(!m_isMobilityExperiment)
						seedMobilityData();
					else
					{
						if(m_curDriftBin > m_driftBinCount)
						{
							m_curDriftBin = 1;
						}

						// qDebug() << __FILE__ << __LINE__
						//<< "spectrum index: " << m_spectrumIndex
						//<< "current drift bin: " << m_curDriftBin;
					}

					// At this point, we need to check if we are loading data to
					// transfer them to a SQLite3 database:

					if(!m_sqlite3Handler.m_dbFileName.isEmpty())
					{

						// qDebug() << __FILE__ << __LINE__ << "The data load is for "
						//"injecting data to the SQLite3 database.";

						// We only store the driftData to the database if this file is
						// actually about a mobility experiment.

						if(m_curDriftTime != -1)
						{

							// Now that we have settled down with proper values for
							// drift stuff, let's try to send these to the database (or
							// get a row number if the data were already in, that is if
							// we are iterating in a spectrum that belong not to the
							// first drift cycle run).

							bool sqlResult = m_sqlite3Handler.writeDbDriftData(
									m_curDriftTime, m_curDriftBin, &m_driftDataId);

							if(!sqlResult || m_driftDataId == -1)
							{
								qDebug()
									<< __FILE__ << __LINE__
									<< "Failed to insert precursor ion data in the "
									"database. Exiting.\n";

								qFatal("Fatal error at %s@%d. Program aborted.",
										__FILE__, __LINE__);
							}
						}
					}

					int mzSize = m_mzQByteArray.size();
					int iSize  = m_iQByteArray.size();

					if(mzSize || iSize)
					{
						// qDebug() << __FILE__ << __LINE__ << "At least one of the "
						//"mz/i byte arrays "
						//"contains data.";

						// Any of the two mz or i byte arrays contain something.
						// Let's
						// check that a bit more in detail:

						if(mzSize != iSize)
						{
							if(m_dataCompression != DATA_COMPRESSION_ZLIB)
							{
								qDebug() << __FILE__ << __LINE__
									<< "It is not possible that the list of m/z "
									"values has a number of "
									<< "items different from that of the list of "
									"i values.";

								qFatal("Fatal error at %s@%d. Program aborted.",
										__FILE__, __LINE__);
							}
							else
							{
								// qDebug() << __FILE__ << __LINE__
								// << "spectrum index " << m_spectrumIndex
								// << "has mzList size and iList size differ, "
								// << "but the data are compressed, so this is not
								// an
								// error.";
							}
						}

						// Whatever the processing is going to be below, we create a
						// mass spectrum with the data:

						// qDebug() << __FILE__ << __LINE__
						// << "Now creating the mass spectrum starting from the
						// byte
						// arrays";

						mp_newSpectrum = MassSpectrum::createFromBase64EncByteArrays(
								&m_mzQByteArray, &m_iQByteArray, m_dataCompression);

						if(mp_newSpectrum == nullptr)
							qFatal("Fatal error at %s@%d. Program aborted.", __FILE__,
									__LINE__);

						qDebug() << __FILE__ << __LINE__
							<< "Retention time for the spectrum at index : "
							<< m_spectrumIndex << "is: " << m_scanStartTime
							<< "and number of data points:" << mp_newSpectrum->size()
							<< "with current drift bin: " << m_curDriftBin;

						// At this point, depending on the task at hand we have
						// the
						// following possibilities:
						// 1. Read the data to the mp_massSpecDataSet
						// 2. Read the data to the mp_key/valVector.

						if(mp_massSpecDataSet != nullptr)
						{
							// qDebug() << __FILE__ << __LINE__
							//<< "We are parsing to the MassSpecDataSet.";

							loadSpectrumToMassSpecDataSet();

							qDebug() << __FILE__ << __LINE__
								<< "Done loading spectrum to mass spec data set:"
								<< "m_spectrumIndex:" << m_spectrumIndex
								<< "m_tic:" << m_tic
								<< "m_scanStartTime:" << m_scanStartTime;
						}
						else if(m_isLoadToVectors)
						{
							// qDebug() << __FILE__ << __LINE__
							//<< "We are parsing to the vectors.";

							loadSpectrumToVectors();
						}
						else if(!m_sqlite3Handler.m_dbFileName.isEmpty())
						{

							//qDebug() << __FILE__ << __LINE__ 
							//<< "We are transferring the data to the SQLite3 database.";

							// This call fails with qFatal, so no need to test for error
							// here.

							m_sqlite3Handler.writeDbSpectralData(
									m_spectrumIndex, m_spectrumTitle, m_specPeakCount,
									m_msLevel, m_tic, m_collisionEnergy, m_precursorIonId,
									m_scanStartTime, m_driftDataId, m_scanWindowLowerLimit,
									m_scanWindowUpperLimit, m_mzQByteArray, m_iQByteArray);

							// The spectrum was effectively stored.
							m_storedSpectraCount++;
						}
					}
					// End of
					// if(mzSize || iSize)
					else
					{

						// The data were totally empty, neither mz nor i contained data.
						//
						// Just skip the spectrum that has one of its component lists
						// empty.

						// qDebug() << __FILE__ << __LINE__
						//<< "None of the mz or i byte arrays did contain data.";

						++m_totalSkippedSpectra;
					}
				}
				// End of
				// if(m_xmlStreamReader.name() == "spectrum")
				else
				{
					// qDebug() << __FILE__ << __LINE__ << "Skipping element: "
					//<< m_xmlStreamReader.name().toString();

					m_xmlStreamReader.skipCurrentElement();
				}
			}
			// End of
			// while(m_xmlStreamReader.readNextStartElement())

			// We might have interrupted the loop above because the operation was
			// cancelled from the caller function in a given window (cancel buttons
			// over there). Thus reset the value to false:

			m_isOperationCancelled = false;

			// At this point we have finished loading all the spectra in the file.

			if(!m_sqlite3Handler.m_dbFileName.isEmpty())
			{

				// If we were loading data to send them to the SQLite3 database, then we
				// can finish sending the metadata to the corresponding database table.

				// First, we need to see if the data were compressed, because in that
				// case we do not need to change anything to the metaData, but if the
				// data were not compressed, then we have to document it.

				int compressionType;

				bool ok = m_msFileMetaData.intItem("compressionType", compressionType);
				if(!ok)
					m_msFileMetaData.setIntItem("compressionType",
							static_cast<int>(m_dataCompression));

				// Also, set the number of actual spectra stored in the database.
				m_msFileMetaData.setIntItem("spectrumListCount", m_storedSpectraCount);

				if(!m_sqlite3Handler.writeDbMassSpecFileMetaData(m_msFileMetaData))
				{
					qDebug() << __FILE__ << __LINE__
						<< "Failed to insert the metaData in the database.";
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__,
							__LINE__);
				}
			}
			// End of
			// if(!MassSpecSqlite3Handler::m_dbFileName.isEmpty())
			else if(mp_massSpecDataSet)
			{
				int rtHashSize = mp_massSpecDataSet->m_rtHash.size();
				int dtHashSize = mp_massSpecDataSet->m_dtHash.size();

				// Sanity check, that is only correct if we are not in streamed
				// mode and
				// we load a drift data file.
				if(m_curDriftTime != -1 && !mp_massSpecDataSet->m_isStreamed)
				{
					if(rtHashSize != dtHashSize)
						qFatal("Fatal error at %s@%d. Program aborted.", __FILE__,
								__LINE__);
				}

				QString msg = QString("Number of spectra in the set: %1\n"
						"Number of retention time values: %2\n"
						"Number of drift time values: %3\n")
					.arg(mp_massSpecDataSet->m_massSpectra.size())
					.arg(rtHashSize)
					.arg(dtHashSize);

				// qDebug() << __FILE__ << __LINE__ << msg.toLatin1();

				emit dataLoadMessageSignal(msg, LOG_TO_BOTH);

				// for(int iter = 0; iter < keyList.size(); ++iter)
				//{
				// MassSpectrum *spectrum =
				// mp_massSpecDataSet->m_rtHash.value(keyList.at(iter));

				// qDebug() << __FILE__ << __LINE__
				// << "For retention time " << keyList.at(iter) << "the tic is: "
				// << spectrum->tic();
				//}
			}
		}


	void
		MassSpecDataFileLoaderMzml::parse_spectrum_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "spectrum")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// Start with a general reset of all the class members that will store the
			// data obtained from parsing this <spectrum> element. The only data that
			// are not reset are the mobility data because we need them to be "static"
			// over multiple spectra iterations.

			resetSpectralData();

			bool ok = false;

			QString spectrumTitle;

			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

			// this is the format of the element.
			// <spectrum index="100001" id="function=2 process=0 scan=2"
			// defaultArrayLength="0">

			QString value = attributes.value(QString(), "index").toString();
			if(value.isEmpty())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			m_spectrumIndex = value.toInt(&ok, 10 /*base*/);
			if(!ok)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			value = attributes.value(QString(), "defaultArrayLength").toString();
			m_specPeakCount = value.toInt(&ok, 10 /*base*/);

			value = attributes.value(QString(), "id").toString();
			if(value.isEmpty())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// When the user asks for addition of raw data from the MassLynx software,
			// then the raw file contains two set of mobility data. Function1 contains
			// the proper set and Function2 contains the second set which is not
			// correctly converted by ProteoWizard into the mzML file. There are a huge
			// number of spectra that are containing 'nan' values for drift time. We
			// want to skip these.

			if(value.contains("function=2"))
			{
				qDebug() << __FILE__ << __LINE__ << "Spectrum at index "
					<< m_spectrumIndex
					<< "is a 'Function 2' spectrum, skipping it.";

				// The caller needs to know that we have hit the part of the file
				// where
				// all the spectra are spurious Function 2 spectra.

				m_wasFunction2Spectrum = true;

				m_xmlStreamReader.skipCurrentElement();
				return;
			}

			while(m_xmlStreamReader.readNextStartElement())
			{
				// qDebug() << __FILE__ << __LINE__ << "in the run loop:"
				//<< "now in element: " << m_xmlStreamReader.name()
				//<< "now in token: " << m_xmlStreamReader.tokenString();

				if(m_xmlStreamReader.name() == "cvParam")
					parse_spectrum_cvParam_element();
				else if(m_xmlStreamReader.name() == "scanList")
					parse_scanList_element();
				else if(m_xmlStreamReader.name() == "precursorList")
				{
					parse_precursorList_element();

					// We have finished parsing all the selected ions, make a sanity
					// check.

					if(m_selectedIonListCount != m_precursorMzList.size())
						qFatal("Fatal error at %s@%d. Program aborted.", __FILE__,
								__LINE__);

					// At this point, we should insert these data into the precursorIons
					// table, or if the data are there already, get the proper row id for
					// the table record.

					QString precursorMzString;

					// Note that there might be some intensity information, but as of now,
					// we do not see any in the mzML file. But the database expects two
					// lists: mz and i. So craft  a fake i list.

					QString precursorIString;

					for(int iter = 0; iter < m_precursorMzList.size(); ++iter)
					{
						if(!iter)
						{
							precursorMzString +=
								QString("%1").arg(m_precursorMzList.at(iter), 0, 'f', 12);
							// qDebug() << __FILE__ << __LINE__ << __FUNCTION__
							// << "precursorMzString:" << precursorMzString;

							precursorIString += "0.0";
						}
						else
						{
							precursorMzString += QString(",%1").arg(
									m_precursorMzList.at(iter), 0, 'f', 12);
							// qDebug() << __FILE__ << __LINE__ << __FUNCTION__
							// << "precursorMzString:" << precursorMzString;

							precursorIString += ",0.0";
						}
					}

					if(!m_sqlite3Handler.m_dbFileName.isEmpty())
					{
						bool sqlResult = m_sqlite3Handler.writeDbPrecursorIons(
								precursorMzString, precursorIString, &m_precursorIonId);

						if(!sqlResult || m_precursorIonId == -1)
						{
							qDebug() << __FILE__ << __LINE__
								<< "Failed to insert "
								"precursor ion data in the "
								"database. Exiting.\n";

							qFatal("Fatal error at %s@%d. Program aborted.", __FILE__,
									__LINE__);
						}
					}
				}
				// End of
				// else if(m_xmlStreamReader.name() == "precursorList")
				else if(m_xmlStreamReader.name() == "binaryDataArrayList")
					parse_spectrum_binaryDataArrayList_element();
				else
				{
					// qDebug() << __FILE__ << __LINE__ << "Skipping element "
					//<< m_xmlStreamReader.name().toString();

					m_xmlStreamReader.skipCurrentElement();
				}
			}
		}


	void
		MassSpecDataFileLoaderMzml::parse_spectrum_cvParam_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "cvParam")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// Check if there is something of interest.

			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

			QString cvRef     = attributes.value(QString(), "cvRef").toString();
			QString accession = attributes.value(QString(), "accession").toString();
			QString name      = attributes.value(QString(), "name").toString();

			if(cvRef == "MS" && accession == "MS:1000511" && name == "ms level")
			{
				QString value = attributes.value(QString(), "value").toString();

				bool ok;
				m_msLevel = value.toInt(&ok, 10 /*base*/);
				if(!ok)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else if(cvRef == "MS" && accession == "MS:1000796" &&
					name == "spectrum title")
			{
				m_spectrumTitle = attributes.value(QString(), "value").toString();
			}
			else
			{
				// qDebug() << __FILE__ << __LINE__
				// << "Parsing element " << m_xmlStreamReader.name().toString();
			}

			m_xmlStreamReader.skipCurrentElement();
		}


	void
		MassSpecDataFileLoaderMzml::parse_scanList_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "scanList")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

			QString value = attributes.value(QString(), "count").toString();

			bool ok;
			m_scanCount = value.toInt(&ok, 10 /*base*/);
			if(!ok)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// Now iterate in the contained elements:
			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() == "cvParam")
					parse_scanList_cvParam_element();
				else if(m_xmlStreamReader.name() == "scan")
					parse_scanList_scan_element();
				else
					m_xmlStreamReader.skipCurrentElement();
			}

			// We are at the end of <scanList>. Good.
		}


	void
		MassSpecDataFileLoaderMzml::parse_scanList_cvParam_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "cvParam")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			m_xmlStreamReader.skipCurrentElement();
		}


	void
		MassSpecDataFileLoaderMzml::parse_scanList_scan_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "scan")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() == "cvParam")
					parse_scan_cvParam_element();
				else if(m_xmlStreamReader.name() == "scanWindowList")
					parse_scanWindowList_element();
				else
					m_xmlStreamReader.skipCurrentElement();
			}

			// We are at the end of <scan>. Good.
		}

	void
		MassSpecDataFileLoaderMzml::parse_scan_cvParam_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "cvParam")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// <cvParam cvRef="MS" accession="MS:1000016" name="scan start time"
			// value="0.0339000002" unitCvRef="UO" unitAccession="UO:0000031"
			// unitName="minute"/>
			// <cvParam cvRef="MS" accession="MS:1002476" name="ion mobility drift
			// time"
			// value="0.0" unitCvRef="UO" unitAccession="UO:0000028"
			// unitName="millisecond"/>

			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

			QString cvRef     = attributes.value(QString(), "cvRef").toString();
			QString accession = attributes.value(QString(), "accession").toString();
			QString name      = attributes.value(QString(), "name").toString();
			QString unitName  = attributes.value(QString(), "unitName").toString();

			bool ok;

			if(cvRef == "MS" && accession == "MS:1000016" && name == "scan start time")
			{
				QString value = attributes.value(QString(), "value").toString();

				m_scanStartTime = value.toDouble(&ok);
				if(!ok)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

				// qDebug() << __FILE__ << __LINE__
				//<< "m_scanStartTime: " << m_scanStartTime;

				// We need to set all the values to min, even if they are in seconds,
				// for example:
				if(unitName == "second")
					m_scanStartTime = m_scanStartTime / 60;
				else if(unitName == "millisecond")
					m_scanStartTime = (m_scanStartTime / 1000) / 60;
				// else that must be min, which is fine.
			}
			else if(cvRef == "MS" && accession == "MS:1002476" &&
					name == "ion mobility drift time")
			{
				QString value = attributes.value(QString(), "value").toString();

				m_curDriftTime = value.toDouble(&ok);
				if(!ok)
				{
					QString msg =
						QString("Spectrum at index %1 had a bad drift time value: '%2'\n")
						.arg(m_spectrumIndex)
						.arg(value);

					m_errorMessageList.append(msg);

					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__,
							__LINE__);
				}


				// qDebug() << __FILE__ << __LINE__
				//<< "m_curDriftTime: " << m_curDriftTime;


				// We previously stored the drift data in the corresponding
				// database
				// table, but this was wrong as the drift bin value had not had
				// time
				// enough to be corrected. So we move the db stuff in the
				// parse_spectrum_element().
			}

			m_xmlStreamReader.skipCurrentElement();
		}


	void
		MassSpecDataFileLoaderMzml::parse_scanWindowList_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "scanWindowList")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			//<scanWindowList count="1">
			//<scanWindow>
			//<cvParam cvRef="MS" accession="MS:1000501" name="scan window lower
			// limit"
			// value="600.0" unitCvRef="MS" unitAccession="MS:1000040"
			// unitName="m/z"/>
			//<cvParam cvRef="MS" accession="MS:1000500" name="scan window upper
			// limit"
			// value="2500.0" unitCvRef="MS" unitAccession="MS:1000040"
			// unitName="m/z"/>
			//</scanWindow>
			//</scanWindowList>

			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();
			QString value = attributes.value(QString(), "count").toString();

			bool ok;
			m_scanWindowListCount = value.toInt(&ok, 10 /*base*/);
			if(!ok)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() == "scanWindow")
					parse_scanWindow_element();
				else
					m_xmlStreamReader.skipCurrentElement();
			}

			// We are at the end of <scanWindowList>. Good.
		}


	void
		MassSpecDataFileLoaderMzml::parse_scanWindow_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "scanWindow")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// Debugging purposes only
			// int lineNumber = m_xmlStreamReader.lineNumber();
			// qDebug() << __FILE__ << __LINE__
			// << "lineNumber: " << lineNumber;

			//<cvParam cvRef="MS" accession="MS:1000501" name="scan window lower
			// limit"
			// value="600.0" unitCvRef="MS" unitAccession="MS:1000040"
			// unitName="m/z"/>
			//<cvParam cvRef="MS" accession="MS:1000500" name="scan window upper
			// limit"
			// value="2500.0" unitCvRef="MS" unitAccession="MS:1000040"
			// unitName="m/z"/>


			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() == "cvParam")
				{
					QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

					QString cvRef = attributes.value(QString(), "cvRef").toString();
					QString accession =
						attributes.value(QString(), "accession").toString();
					QString name = attributes.value(QString(), "name").toString();

					bool ok;

					if(cvRef == "MS" && accession == "MS:1000501" &&
							name == "scan window lower limit")
					{
						QString value = attributes.value(QString(), "value").toString();

						m_scanWindowLowerLimit = value.toDouble(&ok);
						if(!ok)
							qFatal("Fatal error at %s@%d. Program aborted.", __FILE__,
									__LINE__);
					}
					else if(cvRef == "MS" && accession == "MS:1000500" &&
							name == "scan window upper limit")
					{
						QString value = attributes.value(QString(), "value").toString();

						m_scanWindowUpperLimit = value.toDouble(&ok);
						if(!ok)
							qFatal("Fatal error at %s@%d. Program aborted.", __FILE__,
									__LINE__);
					}
				}
				// Whatever the parsed element, we skip it now.
				m_xmlStreamReader.skipCurrentElement();
			}

			// We are at the end of <scanWindow>. Good.
		}


	void
		MassSpecDataFileLoaderMzml::parse_precursorList_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "precursorList")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();
			QString value = attributes.value(QString(), "count").toString();

			bool ok;
			m_precursorListCount = value.toInt(&ok, 10 /*base*/);

			if(!ok)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() == "precursor")
					parse_precursor_element();
				else
					m_xmlStreamReader.skipCurrentElement();
			}

			// We are at the end of <precursorList>. Good.
		}

	void
		MassSpecDataFileLoaderMzml::parse_precursor_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "precursor")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() == "isolationWindow")
					parse_isolationWindow_element();
				else if(m_xmlStreamReader.name() == "selectedIonList")
					parse_selectedIonList_element();
				else if(m_xmlStreamReader.name() == "activation")
					parse_activation_element();
			}

			// We are at the end of <precursor>. Good.
		}

	void
		MassSpecDataFileLoaderMzml::parse_isolationWindow_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "isolationWindow")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			//<isolationWindow>
			//<cvParam cvRef="MS" accession="MS:1000827" name="isolation window
			// target
			// m/z" value="1246" unitCvRef="MS" unitAccession="MS:1000040"
			// unitName="m/z"/>
			//</isolationWindow>

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() == "cvParam")
				{
					QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

					QString cvRef = attributes.value(QString(), "cvRef").toString();
					QString accession =
						attributes.value(QString(), "accession").toString();
					QString name = attributes.value(QString(), "name").toString();

					bool ok;

					if(cvRef == "MS" && accession == "MS:1000827" &&
							name == "isolation window target m/z")
					{
						QString value = attributes.value(QString(), "value").toString();

						m_precursorMz = value.toDouble(&ok);
						if(!ok)
							qFatal("Fatal error at %s@%d. Program aborted.", __FILE__,
									__LINE__);
					}
				}
				// Whatever the parsed element, we skip it now.
				m_xmlStreamReader.skipCurrentElement();
			}
			// We are at the end of <isolationWindow>. Good.
		}


	void
		MassSpecDataFileLoaderMzml::parse_selectedIonList_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "selectedIonList")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			//<selectedIonList count="1">
			//<selectedIon>
			//<cvParam cvRef="MS" accession="MS:1000744" name="selected ion m/z"
			// value="1246.0" unitCvRef="MS" unitAccession="MS:1000040"
			// unitName="m/z"/>
			//</selectedIon>
			//</selectedIonList>

			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();
			QString value = attributes.value(QString(), "count").toString();

			bool ok;
			m_selectedIonListCount = value.toInt(&ok, 10 /*base*/);

			if(!ok)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() == "selectedIon")
					parse_selectedIon_element();
				else
					m_xmlStreamReader.skipCurrentElement();
			}

			// We are at the end of <selectedIonList>. Good.
		}


	void
		MassSpecDataFileLoaderMzml::parse_selectedIon_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "selectedIon")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			//<selectedIon>
			//<cvParam cvRef="MS" accession="MS:1000744" name="selected ion m/z"
			// value="1246.0" unitCvRef="MS" unitAccession="MS:1000040"
			// unitName="m/z"/>
			//</selectedIon>

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() == "cvParam")
				{
					QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

					QString cvRef = attributes.value(QString(), "cvRef").toString();
					QString accession =
						attributes.value(QString(), "accession").toString();
					QString name = attributes.value(QString(), "name").toString();

					bool ok;

					if(cvRef == "MS" && accession == "MS:1000744" &&
							name == "selected ion m/z")
					{
						QString value = attributes.value(QString(), "value").toString();

						double mz = value.toDouble(&ok);

						if(!ok)
							qFatal("Fatal error at %s@%d. Program aborted.", __FILE__,
									__LINE__);

						m_precursorMzList.append(mz);
					}
				}
				// Whatever the parsed element, we skip it now.
				m_xmlStreamReader.skipCurrentElement();
			}
			// We are at the end of <selectedIon>. Good.
		}


	void
		MassSpecDataFileLoaderMzml::parse_activation_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "activation")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			//<activation>
			//<cvParam cvRef="MS" accession="MS:1000422" name="beam-type
			// collision-induced dissociation" value=""/>
			//<cvParam cvRef="MS" accession="MS:1000045" name="collision energy"
			// value="4" unitCvRef="UO" unitAccession="UO:0000266"
			// unitName="electronvolt"/>
			//</activation>

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() == "cvParam")
				{
					QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

					QString cvRef = attributes.value(QString(), "cvRef").toString();
					QString accession =
						attributes.value(QString(), "accession").toString();
					QString name = attributes.value(QString(), "name").toString();

					bool ok;

					if(cvRef == "MS" && accession == "MS:1000045" &&
							name == "collision energy")
					{
						QString value = attributes.value(QString(), "value").toString();

						m_collisionEnergy = value.toDouble(&ok);
						if(!ok)
							qFatal("Fatal error at %s@%d. Program aborted.", __FILE__,
									__LINE__);
					}
				}
				m_xmlStreamReader.skipCurrentElement();
			}

			// We are at the end of <activation>. Good.
		}


	void
		MassSpecDataFileLoaderMzml::parse_spectrum_binaryDataArrayList_element()
		{
			//<binaryDataArrayList count="2">
			//<binaryDataArray encodedLength="0">
			//<cvParam cvRef="MS" accession="MS:1000523" name="64-bit float"
			// value=""/>
			//<cvParam cvRef="MS" accession="MS:1000576" name="no compression"
			// value=""/>
			//<cvParam cvRef="MS" accession="MS:1000514" name="m/z array" value=""
			// unitCvRef="MS" unitAccession="MS:1000040" unitName="m/z"/>
			//<binary></binary>
			//</binaryDataArray>
			//<binaryDataArray encodedLength="0">
			//<cvParam cvRef="MS" accession="MS:1000523" name="64-bit float"
			// value=""/>
			//<cvParam cvRef="MS" accession="MS:1000576" name="no compression"
			// value=""/>
			//<cvParam cvRef="MS" accession="MS:1000515" name="intensity array"
			// value=""
			// unitCvRef="MS" unitAccession="MS:1000131" unitName="number of
			// detector
			// counts"/>
			//<binary></binary>
			//</binaryDataArray>
			//</binaryDataArrayList>

			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "binaryDataArrayList")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();
			QString value = attributes.value(QString(), "count").toString();

			bool ok;
			m_spectrumBinaryDataArrayListCount = value.toInt(&ok, 10 /*base*/);

			if(!ok)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			while(m_xmlStreamReader.readNextStartElement())
			{
				// qDebug() << __FILE__ << __LINE__ << "in the run loop:"
				//<< "now in element: " << m_xmlStreamReader.name()
				//<< "now in token: " << m_xmlStreamReader.tokenString();

				if(m_xmlStreamReader.name() == "binaryDataArray")
					parse_spectrum_binaryDataArray_element();

				else
					m_xmlStreamReader.skipCurrentElement();
			}

			// We are at the end of <binaryDataArrayList>. Good.
		}


	void
		MassSpecDataFileLoaderMzml::parse_spectrum_binaryDataArray_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "binaryDataArray")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We will need to know what was the kind of data that were parsed,
			// and
			// we'll store that bit of info in this variable.

			m_dataKind = DataKind::DATA_KIND_UNSET;

			// qDebug() << __FILE__ << __LINE__
			//<< "Resetting m_dataKind to DataKind::DATA_KIND_UNSET: "
			//<< m_dataKind;

			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();
			QString value = attributes.value(QString(), "encodedLength").toString();

			bool ok;
			m_encodedLength = value.toDouble(&ok);

			if(!ok)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			//<binaryDataArray encodedLength="0">
			//<cvParam cvRef="MS" accession="MS:1000523" name="64-bit float"
			// value=""/>
			//<cvParam cvRef="MS" accession="MS:1000576" name="no compression"
			// value=""/>
			//<cvParam cvRef="MS" accession="MS:1000514" name="m/z array" value=""
			// unitCvRef="MS" unitAccession="MS:1000040" unitName="m/z"/>
			//<binary></binary>
			//</binaryDataArray>
			//

			// But it might also be that there is compression:
			//
			// <cvParam cvRef="MS" accession="MS:1000574" name="zlib compression"
			// value=""/>

			QString binaryTagDataString;

			while(m_xmlStreamReader.readNextStartElement())
			{
				// qDebug() << __FILE__ << __LINE__
				//<< "now in element: " << m_xmlStreamReader.name()
				//<< "now in token: " << m_xmlStreamReader.tokenString();

				if(m_xmlStreamReader.name() == "cvParam")
					parse_spectrum_binaryDataArray_cvParam_element();
				else if(m_xmlStreamReader.name() == "binary")
				{
					// qDebug() << __FILE__ << __LINE__
					//<< "Entering the m_xmlStreamReader.name() == binary block."
					//<< "with m_dataKind:" << m_dataKind;

					// At this point, get the data. We need to know if these data
					// are
					// for mz or i.

					if(m_dataKind == DataKind::DATA_KIND_MZ)
					{
						binaryTagDataString = m_xmlStreamReader.readElementText();

						if(binaryTagDataString.size() < 3)
						{
							// qDebug() << __FILE__ << __LINE__
							//<< "mz data too short to be meaningful. Skipping.";
						}
						else
						{
							// We have some data to crunch.

							// The binaryTagDataString is a base64-encoded string.
							// We
							// need to fill-in a QByteArray with these data. This
							// is
							// what is going to be stored in the database.

							m_mzQByteArray.clear();
							m_mzQByteArray.append(binaryTagDataString.toLatin1());

							// qDebug() << __FILE__ << __LINE__
							//<< "Appended data to m_mzQByteArray. Size:"
							//<< m_mzQByteArray.size();
						}
					}
					// End of
					// if(m_dataKind == DataKind::DATA_KIND_MZ)
					else if(m_dataKind == DataKind::DATA_KIND_I)
					{
						binaryTagDataString = m_xmlStreamReader.readElementText();

						if(binaryTagDataString.size() < 3)
						{
							// qDebug() << __FILE__ << __LINE__
							//<< "i data too short to be meaningful. Skipping.";
						}
						else
						{
							// We have some data to crunch.

							// The binaryTagDataString is a base64-encoded string.
							// We
							// need to fill-in a QByteArray with these data that
							// will be
							// copied to the database.

							m_iQByteArray.clear();
							m_iQByteArray.append(binaryTagDataString.toLatin1());

							// qDebug() << __FILE__ << __LINE__
							//<< "Appended data to m_iQByteArray. Size:"
							//<< m_iQByteArray.size();
						}

						// Do not skip the element, because we have read all its
						// text,
						// and thus we are already at the </binary> end token.
					}
					// End of
					// else if(m_dataKind == DataKind::DATA_KIND_I)
					else
					{
						qDebug() << __FILE__ << __LINE__ << "DataKind not tested.";
					}
				}
				// End of
				// else if(m_xmlStreamReader.name() == "binary")

				// We are at the end of <binaryDataArray>. Good.
			}
			// End of
			// while(m_xmlStreamReader.readNextStartElement())
		}


	void
		MassSpecDataFileLoaderMzml::parse_spectrum_binaryDataArray_cvParam_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "cvParam")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			//<binaryDataArray encodedLength="0">
			//<cvParam cvRef="MS" accession="MS:1000523" name="64-bit float"
			// value=""/>
			//<cvParam cvRef="MS" accession="MS:1000576" name="no compression"
			// value=""/>
			//<cvParam cvRef="MS" accession="MS:1000514" name="m/z array" value=""
			// unitCvRef="MS" unitAccession="MS:1000040" unitName="m/z"/>
			//<binary></binary>
			//</binaryDataArray>
			//
			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

			QString cvRef     = attributes.value(QString(), "cvRef").toString();
			QString accession = attributes.value(QString(), "accession").toString();
			QString name      = attributes.value(QString(), "name").toString();

			if(cvRef == "MS" && accession == "MS:1000514" && name == "m/z array")
				m_dataKind = DataKind::DATA_KIND_MZ;
			else if(cvRef == "MS" && accession == "MS:1000515" &&
					name == "intensity array")
				m_dataKind = DataKind::DATA_KIND_I;
			else if(cvRef == "MS" && accession == "MS:1000574" &&
					name == "zlib compression")
			{
				m_dataCompression = DATA_COMPRESSION_ZLIB;

				// Store the compression for the metaData table of the database:
				m_msFileMetaData.setIntItem("compressionType",
						static_cast<int>(m_dataCompression));
			}

			// qDebug() << __FILE__ << __LINE__ << "m_dataKind:" << m_dataKind
			//<< "m_dataCompression:" << m_dataCompression;

			m_xmlStreamReader.skipCurrentElement();
		}


	void
		MassSpecDataFileLoaderMzml::parse_chromatogramList_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "chromatogramList")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			bool ok = false;

			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

			// this is the format of the element.
			// <chromatogramList count="1"
			// defaultDataProcessingRef="pwiz_Reader_Waters_conversion">

			QString value = attributes.value(QString(), "count").toString();
			if(value.isEmpty())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			m_chromatogramListCount = value.toInt(&ok, 10 /*base*/);
			if(!ok)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() == "chromatogram")
				{
					parse_chromatogram_element();

					// At this point, make sure that both QByteArray instances
					// have the
					// same
					// number of items.

					int timeSize = m_chromTimeQByteArray.size();
					int iSize    = m_chromIQByteArray.size();

					if(timeSize && iSize)
					{
						if(timeSize != iSize)
						{
							if(m_dataCompression != DATA_COMPRESSION_ZLIB)
							{
								qDebug() << __FILE__ << __LINE__
									<< "It is not possible that the list of time "
									"values has a number of "
									<< "items different from that of the list of "
									"i values.";

								qFatal("Fatal error at %s@%d. Program aborted.",
										__FILE__, __LINE__);
							}
							else
							{
								// qDebug() << __FILE__ << __LINE__
								// << "chromatogram index " << m_chromatogramIndex
								// << "has timeList size and iList size differ, "
								//"but the data are compressed, so this is not an
								// error.";
							}
						}

						// Yes, both data arrays were parsed. Their size is not 0
						// and
						// both sizes
						// are identical (same number of double instances in each
						// QByteArray).
						// Let's make the database call to store that to the
						// database.

						if(!m_sqlite3Handler.m_dbFileName.isEmpty())
						{
							// This call fails with qFatal, so no need to test for
							// error
							// here.
							m_sqlite3Handler.writeDbChromatographicData(
									m_chromatogramIndex, m_chromatogramTitle,
									m_chromTimeQByteArray, m_chromIQByteArray);

							// The chromatogram was effectively stored.
							m_storedChromatogramCount++;
						}
					}
					else
					{
						// The chromatogram has empty binary data arrays, we skip
						// it.
						// qDebug() << __FILE__ << __LINE__
						// << "The chromatogram at index " << m_chromatogramIndex
						// << "has empty binary time and i data. We skip it.";
					}
				}
				else
					// ! if(m_xmlStreamReader.name() == "chromatogram")
				{
					// qDebug() << __FILE__ << __LINE__ << "Parsing element "
					//<< m_xmlStreamReader.name().toString();

					m_xmlStreamReader.skipCurrentElement();
				}
			}
		}


	void
		MassSpecDataFileLoaderMzml::parse_chromatogram_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "chromatogram")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);


			//	<chromatogram index="0" id="TIC" defaultArrayLength="6">
			//		<cvParam cvRef="MS" accession="MS:1000235" name="total ion
			// current
			// chromatogram" value=""/>
			//		<binaryDataArrayList count="2">
			//		<binaryDataArray encodedLength="64">
			//		<cvParam cvRef="MS" accession="MS:1000523" name="64-bit float"
			// value=""/>
			//		<cvParam cvRef="MS" accession="MS:1000576" name="no compression"
			// value=""/>
			//		<cvParam cvRef="MS" accession="MS:1000595" name="time array"
			// value=""
			// unitCvRef="UO" unitAccession="UO:0000031" unitName="minute"/>
			//		<binary>AAAAgIZdoT8AAABgpASqPwAAAMD4VrE/AAAAwIeqtT8AAACgFv65PwAAAEC9Ur4/</binary>
			//		</binaryDataArray>
			//		<binaryDataArray encodedLength="64">
			//		<cvParam cvRef="MS" accession="MS:1000523" name="64-bit float"
			// value=""/>
			//		<cvParam cvRef="MS" accession="MS:1000576" name="no compression"
			// value=""/>
			//		<cvParam cvRef="MS" accession="MS:1000515" name="intensity
			// array"
			// value="" unitCvRef="MS" unitAccession="MS:1000131" unitName="number
			// of
			// detector counts"/>
			//		<binary>AAAAACgwD0EAAAAAWCQLQQAAAABICQxBAAAAAODWCkEAAAAACPwKQQAAAABInglB</binary>
			//		</binaryDataArray>
			//		</binaryDataArrayList>
			//		</chromatogram>

			bool ok = false;

			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

			QString value = attributes.value(QString(), "index").toString();
			if(value.isEmpty())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			m_chromatogramIndex = value.toInt(&ok, 10 /*base*/);
			if(!ok)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() == "cvParam")
					parse_chromatogram_cvParam_element();
				else if(m_xmlStreamReader.name() == "binaryDataArrayList")
					parse_chromatogram_binaryDataArrayList_element();
				else
				{
					m_xmlStreamReader.skipCurrentElement();

					// qDebug() << __FILE__ << __LINE__
					//<< "Name is: " << m_xmlStreamReader.name().toString();
				}
			}
		}


	void
		MassSpecDataFileLoaderMzml::parse_chromatogram_cvParam_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "cvParam")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// <cvParam cvRef="MS" accession="MS:1000235" name="total ion current
			// chromatogram" value=""/>

			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

			QString cvRef     = attributes.value(QString(), "cvRef").toString();
			QString accession = attributes.value(QString(), "accession").toString();
			QString name      = attributes.value(QString(), "name").toString();

			if(cvRef == "MS" && accession == "MS:1000235" &&
					name == "total ion current chromatogram")
			{
				// The string below is empty, we let it here for note but we
				// cannot use
				// it.
				// m_chromatogramTitle = attributes.value(QString(),
				// "value").toString();
				m_chromatogramTitle = name;
			}

			m_xmlStreamReader.skipCurrentElement();
		}

	void
		MassSpecDataFileLoaderMzml::parse_chromatogram_binaryDataArrayList_element()
		{
			//	<chromatogramList count="1"
			// defaultDataProcessingRef="pwiz_Reader_Waters_conversion">
			//		<chromatogram index="0" id="TIC" defaultArrayLength="6">
			//		<cvParam cvRef="MS" accession="MS:1000235" name="total ion
			// current
			// chromatogram" value=""/>
			//		<binaryDataArrayList count="2">
			//		<binaryDataArray encodedLength="64">
			//		<cvParam cvRef="MS" accession="MS:1000523" name="64-bit float"
			// value=""/>
			//		<cvParam cvRef="MS" accession="MS:1000576" name="no compression"
			// value=""/>
			//		<cvParam cvRef="MS" accession="MS:1000595" name="time array"
			// value=""
			// unitCvRef="UO" unitAccession="UO:0000031" unitName="minute"/>
			//		<binary>AAAAgIZdoT8AAABgpASqPwAAAMD4VrE/AAAAwIeqtT8AAACgFv65PwAAAEC9Ur4/</binary>
			//		</binaryDataArray>
			//		<binaryDataArray encodedLength="64">
			//		<cvParam cvRef="MS" accession="MS:1000523" name="64-bit float"
			// value=""/>
			//		<cvParam cvRef="MS" accession="MS:1000576" name="no compression"
			// value=""/>
			//		<cvParam cvRef="MS" accession="MS:1000515" name="intensity
			// array"
			// value="" unitCvRef="MS" unitAccession="MS:1000131" unitName="number
			// of
			// detector counts"/>
			//		<binary>AAAAACgwD0EAAAAAWCQLQQAAAABICQxBAAAAAODWCkEAAAAACPwKQQAAAABInglB</binary>
			//		</binaryDataArray>
			//		</binaryDataArrayList>
			//		</chromatogram>
			//		</chromatogramList>

			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "binaryDataArrayList")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();
			QString value = attributes.value(QString(), "count").toString();

			bool ok;
			m_chromatogramBinaryDataArrayListCount = value.toInt(&ok, 10 /*base*/);

			if(!ok)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			while(m_xmlStreamReader.readNextStartElement())
			{
				if(m_xmlStreamReader.name() == "binaryDataArray")
					parse_chromatogram_binaryDataArray_element();
				else
					m_xmlStreamReader.skipCurrentElement();
			}

			// We are at the end of <binaryDataArrayList>. Good.
		}


	void
		MassSpecDataFileLoaderMzml::
		parse_chromatogram_binaryDataArray_cvParam_element()
		{
			m_xmlStreamReader.skipCurrentElement();
		}


	void
		MassSpecDataFileLoaderMzml::parse_chromatogram_binaryDataArray_element()
		{
			if(!m_xmlStreamReader.isStartElement() ||
					m_xmlStreamReader.name() != "binaryDataArray")
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);


			QXmlStreamAttributes attributes = m_xmlStreamReader.attributes();

			QString cvRef     = attributes.value(QString(), "cvRef").toString();
			QString accession = attributes.value(QString(), "accession").toString();
			QString name      = attributes.value(QString(), "name").toString();

			// <cvParam cvRef="MS" accession="MS:1000595" name="time array"
			// value=""
			// unitCvRef="UO" unitAccession="UO:0000031" unitName="minute"/>
			// or
			// <cvParam cvRef="MS" accession="MS:1000515" name="intensity array"
			// value="" unitCvRef="MS" unitAccession="MS:1000131" unitName="number
			// of
			// detector counts"/>

			if(cvRef == "MS" && accession == "MS:1000595" && name == "time array")
				m_dataKind = DATA_KIND_TIME;
			else if(cvRef == "MS" && accession == "MS:1000515" &&
					name == "intensity array")
				m_dataKind = DATA_KIND_I;
			else if(cvRef == "MS" && accession == "MS:1000574" &&
					name == "zlib compression")
			{
				m_dataCompression = DATA_COMPRESSION_ZLIB;

				// Store the compression for the metaData table of the database:
				m_msFileMetaData.setIntItem("compressionType",
						static_cast<int>(m_dataCompression));
			}

			m_xmlStreamReader.skipCurrentElement();
		}


} // namespace msXpertSuite
