/*   msXpertSuite - mass spectrometry software suite
 *   -----------------------------------------------
 *   Copyright(C) 2016 Filippo Rusconi
 *
 *   http://www.{site}.org
 *
 *   This file is part of the msXpertSuite project.
 *
 *   The msXpertSuite project is the successor of the massXpert project. This
 *   project now includes various independent modules:
 *   - massXpert, the original mass spectrometry environment software (greatly
 *     refactored code);
 *   - mobXpert, a module to process and mine mobility mass spectrometry data;
 *   - viewXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


///////////////////////////// Qt include


///////////////////////////// Local includes
#include "ColorMapPlotWidget.hpp"
#include "ColorMapWnd.hpp"
#include "AnalysisPreferences.hpp"
#include "MainWindow.hpp"
#include "History.hpp"


namespace msXpertSuite
{


	ColorMapPlotWidget::ColorMapPlotWidget(QWidget * parent,
			AbstractMultiPlotWnd *parentWnd,
			const QString & name,
			const QString & desc,
			const QString & fileName,
			MassSpecDataSet * massSpecDataSet)
		: AbstractPlotWidget{parent, parentWnd, name,
			desc, massSpecDataSet, fileName}
	{
		registerQtKey(Qt::Key_R); // Retention time (TIC chromatogram) calculation
		registerQtKey(Qt::Key_S); // Mass spectrum calculation
		registerQtKey(Qt::Key_D); // Drift spectrum calculation
		registerQtKey(Qt::Key_I); // TIC intensity calculation

		// Give the axes some labels:
		xAxis->setLabel("drift time (ms)");
		yAxis->setLabel("m/z");

		clearPlottables();

		setupWidget();

		mp_colorMap = new QCPColorMap(xAxis, yAxis);
		addPlottable(mp_colorMap);
	}


	ColorMapPlotWidget::~ColorMapPlotWidget()
	{
	}


	IntegrationType
		ColorMapPlotWidget::arbitraryIntegrationType()
		{
			// Fixme, set correctly defining value.
			return IntegrationType::ANY_TO_DT;
		}


	bool
		ColorMapPlotWidget::setupWidget()
		{
			QPen pen(m_pen);

			pen.setColor("white");
			m_zoomRect->setPen(pen);
			m_descText->setColor("white");
			m_hStartTracer->setPen(pen);
			m_selectLine->setPen(m_pen);
			m_xDeltaText->setColor("white");

			return true;
		}

	void
		ColorMapPlotWidget::mouseReleaseHandledEvent(QMouseEvent * event,
				AbstractPlotWidget *emitter)
		{
			if(m_pressedKeyCode == Qt::Key_R && event->button() == Qt::RightButton)
			{
				integrateToRt(emitter->m_xRangeMin, emitter->m_xRangeMax,
						emitter->m_yRangeMin, emitter->m_yRangeMax);
			}

			if(m_pressedKeyCode == Qt::Key_S && event->button() == Qt::RightButton)
			{
				integrateToMz(emitter->m_xRangeMin, emitter->m_xRangeMax,
						emitter->m_yRangeMin, emitter->m_yRangeMax);
			}

			if(m_pressedKeyCode == Qt::Key_D && event->button() == Qt::RightButton)
			{
				integrateToDt(emitter->m_xRangeMin, emitter->m_xRangeMax,
						emitter->m_yRangeMin, emitter->m_yRangeMax);
			}

			if(m_pressedKeyCode == Qt::Key_I && event->button() == Qt::RightButton)
			{
				integrateToTicIntensity(emitter->m_xRangeMin, emitter->m_xRangeMax,
						emitter->m_yRangeMin, emitter->m_yRangeMax);
			}
		}


	void 
		ColorMapPlotWidget::integrateToRt(double xAxisLower, double xAxisUpper,
				double yAxisLower, double yAxisUpper)
		{
			//FIXME this version maybe does not required to get to the emitter to know
			//the color map, the emitter is 'this' ?
			//QCPAxis *xAxis =
			//static_cast<ColorMapPlotWidget *>(emitter)->mp_colorMap->keyAxis();

			QCPAxis *xAxis = mp_colorMap->keyAxis();
			QPen basePen = xAxis->basePen();
			QColor color = basePen.color();

			History localHistory = m_history;
			HistoryItem *histItem = new HistoryItem;

			histItem->newIntegrationRange(IntegrationType::DATA_TO_DT_MZ_DT_RANGE, 
					xAxisLower, xAxisUpper);

			histItem->newIntegrationRange(IntegrationType::DATA_TO_DT_MZ_MZ_RANGE, 
					yAxisLower, yAxisUpper);

			// Aggreate the new item to the local history copy we made.
			localHistory.appendHistoryItem(histItem);

			emit newTicChromatogramByDtAndMz(mp_massSpecDataSet,
					IntegrationType::ANY_TO_RT, localHistory, color);
		}


	void 
		ColorMapPlotWidget::integrateToMz(double xAxisLower, double xAxisUpper,
				double yAxisLower, double yAxisUpper)
		{
			//FIXME this version maybe does not required to get to the emitter to know
			//the color map, the emitter is 'this' ?
			//QCPAxis *xAxis =
			//static_cast<ColorMapPlotWidget *>(emitter)->mp_colorMap->keyAxis();

			QCPAxis *xAxis = mp_colorMap->keyAxis();
			QPen basePen = xAxis->basePen();
			QColor color = basePen.color();

			History localHistory = m_history;
			HistoryItem *histItem = new HistoryItem;

			histItem->newIntegrationRange(IntegrationType::DATA_TO_DT_MZ_DT_RANGE, 
					xAxisLower, xAxisUpper);

			histItem->newIntegrationRange(IntegrationType::DATA_TO_DT_MZ_MZ_RANGE, 
					yAxisLower, yAxisUpper);

			// Aggreate the new item to the local history copy we made.
			localHistory.appendHistoryItem(histItem);

			emit newMassSpectrumByDtAndMz(mp_massSpecDataSet,
					IntegrationType::ANY_TO_MZ, localHistory, color);
		}


	void 
		ColorMapPlotWidget::integrateToDt(double xAxisLower, double xAxisUpper,
				double yAxisLower, double yAxisUpper)
		{
			//FIXME this version maybe does not required to get to the emitter to know
			//the color map, the emitter is 'this' ?
			//QCPAxis *xAxis =
			//static_cast<ColorMapPlotWidget *>(emitter)->mp_colorMap->keyAxis();

			QCPAxis *xAxis = mp_colorMap->keyAxis();
			QPen basePen = xAxis->basePen();
			QColor color = basePen.color();

			History localHistory = m_history;
			HistoryItem *histItem = new HistoryItem;

			histItem->newIntegrationRange(IntegrationType::DATA_TO_DT_MZ_DT_RANGE, 
					xAxisLower, xAxisUpper);

			histItem->newIntegrationRange(IntegrationType::DATA_TO_DT_MZ_MZ_RANGE, 
					yAxisLower, yAxisUpper);

			// Aggreate the new item to the local history copy we made.
			localHistory.appendHistoryItem(histItem);

			emit newDriftSpectrumByDtAndMz(mp_massSpecDataSet,
					IntegrationType::ANY_TO_DT, localHistory, color);
		}


	// This is not a conventional integration, because it does not produce any
	// graph, it only produces a single TIC intensity value that is then shown in
	// the status bar. 
	double 
		ColorMapPlotWidget::integrateToTicIntensity(double xAxisLower, double xAxisUpper,
				double yAxisLower, double yAxisUpper)
		{
			//FIXME this version maybe does not required to get to the emitter to know
			//the color map, the emitter is 'this' ?
			//QCPAxis *xAxis =
			//static_cast<ColorMapPlotWidget *>(emitter)->mp_colorMap->keyAxis();

			History localHistory = m_history;
			HistoryItem *histItem = new HistoryItem;

			histItem->newIntegrationRange(IntegrationType::DATA_TO_DT_MZ_DT_RANGE, 
					xAxisLower, xAxisUpper);

			histItem->newIntegrationRange(IntegrationType::DATA_TO_DT_MZ_MZ_RANGE, 
					yAxisLower, yAxisUpper);

			// Aggreate the new item to the local history copy we made.
			localHistory.appendHistoryItem(histItem);

			ColorMapWnd *parentWnd = static_cast<ColorMapWnd *>(mp_parentWnd);

			// Capture immediately the text that is located in the status bar
			// because if the computation is long that text will have disappeared
			// and we won't be able to show it in the TIC int text.

			QString currentMessage = parentWnd->statusBar()->currentMessage();

			QApplication::setOverrideCursor(QCursor(Qt::BusyCursor));

			parentWnd->ticIntensity(mp_massSpecDataSet,
					&m_lastI, IntegrationType::DATA_TO_DT_MZ_TO_TIC_INT, localHistory);

			QApplication::restoreOverrideCursor();

			parentWnd->statusBar()->showMessage(currentMessage +
					QString(" ; (Tic I = %1)").arg(m_lastI, 0, 'g', 3));

			return m_lastI;
		}

	void
		ColorMapPlotWidget::keyReleaseEvent(QKeyEvent *event)
		{
			ColorMapWnd *parentWnd = static_cast<ColorMapWnd *>(mp_parentWnd);
			MainWindow * mainWindow = static_cast<MainWindow *>(parentWnd->parent());

			if(event->key() == Qt::Key_L)
			{
				// The user wants to copy the current cursor location to the console
				// window, such that it remains available after having moved the cursor.

				QCPAxis *xAxis = mp_colorMap->keyAxis();
				QPen basePen = xAxis->basePen();
				QColor color = basePen.color();

				QString label;

				label = QString("(%1,%2)\n")
					.arg(m_lastMousedPlotPoint.x())
					.arg(m_lastMousedPlotPoint.y());

				if(label.isEmpty())
				{
					parentWnd->statusBar()->showMessage(
							"Failed to craft a text label.");

					event->accept();
					return;
				}
				else
				{

					// Send the label to the console window, so that we can make a
					// copy paste.

					mainWindow->logConsoleMessage(label, color);
				}

				event->accept();

				return;
			}

			// Now let the base class do its work.
			AbstractPlotWidget::keyReleaseEvent(event);
		}


} // namespace msXpertSuite
