/*   msXpertSuite - mass spectrometry software suite
 *   -----------------------------------------------
 *   Copyright(C) 2016 Filippo Rusconi
 *
 *   http://www.{site}.org
 *
 *   This file is part of the msXpertSuite project.
 *
 *   The msXpertSuite project is the successor of the massXpert project. This
 *   project now includes various independent modules:
 *   - massXpert, the original mass spectrometry environment software (greatly
 *     refactored code);
 *   - mineXpert, a powerful TIC chromatogram/mass/drift spectrum viewer/miner;
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <qmath.h>
#include <omp.h>


/////////////////////// Qt includes
#include <QApplication>
#include <QClipboard>

/////////////////////// Local includes
#include "DriftSpecWnd.hpp"
#include "MainWindow.hpp"
#include "MassSpecDataFileLoaderSqlite3.hpp"
#include "MassSpecDataFileLoaderPwiz.hpp"


namespace msXpertSuite
{

	int DriftSpecWnd::widgetCounter = 0;

	DriftSpecWnd::DriftSpecWnd(QWidget *parent)
		: AbstractMultiPlotWnd{parent, "mineXpert" /*m_name*/,
			"Drift-spectrum" /*m_desc*/}
	{
		initialize();

		show();
	}

	DriftSpecWnd::~DriftSpecWnd()
	{
	}


	void
		DriftSpecWnd::initialize()
		{
			initializePlotRegion();
		}

	void
		DriftSpecWnd::initializePlotRegion()
		{
			// There are two regions in the window. The upper part of the mp_splitter
			// is gonna receive the multi-graph plotwidget. The lower part of the
			// splitter will receive all the individual plot widgets.

			mp_multiGraphPlotWidget =
				new DriftSpecPlotWidget(this, this, m_name, m_desc);
			mp_multiGraphPlotWidget->setObjectName("multiGraphPlot");

			// We will need later to know that this plot widget is actually for
			// displaying many different graphs, contrary to the normal situation.
			mp_multiGraphPlotWidget->m_isMultiGraph = true;

			// We need to remove the single graph that was created upon creation of
			// the QCustomPlot widget because we'll later add graphs, and we want to
			// have a proper correlation between the index of the graphs and their
			// pointer value. If we did not remove that graph "by default", then there
			// would be a gap of 1 between the index of the first added mass graph and
			// its expected value.
			mp_multiGraphPlotWidget->clearGraphs();

			// Ensure that there is a connection between this plot widget and this
			// containing window, such that if there is a requirement to replot, the
			// widget gets the signal.
			connect(this, &AbstractMultiPlotWnd::replotWithAxisRange,
					mp_multiGraphPlotWidget, &AbstractPlotWidget::replotWithAxisRange);

			// This window is responsible for the writing of the analysis stanzas to
			// the file. We need to catch the signals.
			connect(mp_multiGraphPlotWidget, &DriftSpecPlotWidget::recordAnalysisStanza,
					this, &DriftSpecWnd::recordAnalysisStanza);

			// This window is responsible for displaying the key/value pairs of the
			// graphs onto which the mouse is moved. We need to catch the signal.
			connect(mp_multiGraphPlotWidget, &AbstractPlotWidget::updateStatusBarSignal, this,
					&AbstractMultiPlotWnd::updateStatusBar);


			// We want to create a scroll area in which the multigraph plot widget
			// will be sitting.
			QScrollArea *scrollArea = new QScrollArea();
			scrollArea->setObjectName(QStringLiteral("scrollArea"));
			scrollArea->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
			scrollArea->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
			// We accept that the plots are resized, but we set a minimum size when we
			// create them.
			scrollArea->setWidgetResizable(true);
			scrollArea->setAlignment(Qt::AlignCenter);
			scrollArea->setWidget(mp_multiGraphPlotWidget);
			// And now prepend to the mp_splitter that new scroll area. Remember that
			// the mp_splitter had already a widget set to it, that is, the scroll
			// area that is now located below the mp_multiGraphPlotWidget, that is
			// responsible for the hosting of all the single-graph plot widgets.
			mp_splitter->insertWidget(0, scrollArea);
		}



	DriftSpecPlotWidget *
		DriftSpecWnd::addPlotWidget(QVector<double> keyVector,
				QVector<double> valVector, const QString &desc,
				MassSpecDataSet *massSpecDataSet)
		{
			MainWindow *mainWindow = static_cast<MainWindow *>(parent());

			// Beware that massSpecDataSet might be nullptr.
			QString fileName;

			if(massSpecDataSet != nullptr)
				fileName = massSpecDataSet->m_fileName;
			else
				fileName = "NOT_SET";

			DriftSpecPlotWidget *widget = new DriftSpecPlotWidget(
					this, this, m_name, desc, fileName, massSpecDataSet);
			widget->setObjectName("monoGraphPlot");

			///////////////// BEGIN Script availability of the new widget //////////////////

			// Craft a name that differentiates this widget from the previous and next
			// ones. We use the static counter number to affix to the string
			// "driftSpecPlot".

			QString propName =
				QString("driftSpecPlot%1").arg(DriftSpecWnd::widgetCounter);
			QString alias = "lastDriftSpectrumPlot";

			QString explanation = QString("Make available a drift spectrum plot under name \"%1\" "
					"with alias \"%2\".\n").arg(propName).arg(alias);

			// Make the new plot widget globally available under the crafted property
			// name and alias. Also, put out the explanation text. Note that by
			// providing 'this', we make the new object appear in the scripting objet
			// tree widget as a child item of this.

			QScriptValue scriptValue;
			mainWindow->mp_scriptingWnd->publishQObject(widget, this /* parent object */,
					propName, alias, explanation, QString() /* help string */, scriptValue);

#if 0
			// Old way of doing, we now handle the alias in one go above

			// Now make the new plot available under the name "lastDriftSpecPlot"
			// for the user to catch it with own variable, because that object name
			// will be overwritten later.

			mainWindow->mp_scriptingWnd->publishQObject(widget, this /* parent object */,
					"lastDriftSpecPlot", QString() /* alias */,
					"The last (in time) drift spectrum plot created. Use this handle at "
					"creation time to store the object under a name of your choice",
					"Make available the last mass spectrum plot under name \"lastDriftSpectrumPlot\"",
					scriptValue);
#endif

			++DriftSpecWnd::widgetCounter;

			///////////////// END Script availability of the new widget //////////////////



			widget->setMinimumSize(400, 150);
			mp_scrollAreaVBoxLayout->addWidget(widget);

			widget->graph()->setData(keyVector, valVector);

			// Let's see if one color was set as the desired color, otherwise use
			// black.
			QPen currentPen = widget->graph()->pen();

			if(!m_wishedColor.isValid())
				currentPen.setColor(Qt::black);
			else
				currentPen.setColor(m_wishedColor);

			widget->graph()->setPen(currentPen);

			widget->rescaleAxes(true /*only enlarge*/);
			widget->replot();

			// Store in the history the new ranges of both x and y axes.
			widget->updateAxisRangeHistory();

			// At this point, make sure that the same data are represented on the
			// multi-graph plot widget.

			QCPGraph *newGraph = mp_multiGraphPlotWidget->addGraph();
			newGraph->setData(keyVector, valVector);
			newGraph->setPen(currentPen);
			mp_multiGraphPlotWidget->rescaleAxes(true /*only enlarge*/);
			mp_multiGraphPlotWidget->replot();

			// Maintain a hash connection between any given plot widget and the
			// corresponding graph in the multigraph widget.
			m_plotWidgetGraphMap.insert(widget, newGraph);

			// Finally append the new widget to the list of widgets that belong to
			// this window.
			m_plotWidgetList.append(widget);

			// Signal/slot connections.

			connect(widget, &DriftSpecPlotWidget::recordAnalysisStanza, this,
					&DriftSpecWnd::recordAnalysisStanza);

			connect(widget, &AbstractPlotWidget::updateStatusBarSignal, this,
					&AbstractMultiPlotWnd::updateStatusBar);

			return widget;
		}


	void
		DriftSpecWnd::setAnalysisPreferences(AnalysisPreferences *preferences)
		{
			if(preferences == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			mp_analysisPreferences = preferences;
		}


	void
		DriftSpecWnd::setAnalysisFile(QFile *analysisFile)
		{
			if(analysisFile == nullptr)
				qFatal(
						"Fatal error at %s@%d. The file pointer cannot be 0. Program aborted.",
						__FILE__, __LINE__);

			mp_analysisFile = analysisFile;
		}

	void
		DriftSpecWnd::newDriftSpectrumByRt(MassSpecDataSet *massSpecDataSet,
				IntegrationType integrationType,
				History history, QColor color)
		{
			if(massSpecDataSet == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			m_wishedColor = color;

			// We get a History reference that informs us about the ranges in the
			// sense that the innermost range for integrationType is by necessity the
			// integration range we work on.

			double rangeStart;
			double rangeEnd;

			if(!history.innermostRange(integrationType, &rangeStart, &rangeEnd))
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// qDebug() << __FILE__ << __LINE__
			//<< "rangeStart:" << rangeStart << "rangeEnd:" << rangeEnd;

			MainWindow *mainWindow = static_cast<MainWindow *>(parent());
			QString msgSkeleton =
				QString("Computing drift spectrum between rt %1 and rt %2")
				.arg(rangeStart, 0, 'f', 6)
				.arg(rangeEnd, 0, 'f', 6);
			mainWindow->logConsoleMessage(msgSkeleton);

#if 0
			// No more of use here at this is performed in the calling function.

			History localHistory = history;
			HistoryItem *histItem = new HistoryItem;
			histItem->newIntegrationRange(integrationType, rangeStart, rangeEnd);
			histItem->setDateTime();

			// Aggreate the new item to the history we got as a parameter and that
			// we copied into localHistory.
			localHistory.appendHistoryItem(histItem);
#endif

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);

			double start = omp_get_wtime();

			QVector<double> keyVector;
			QVector<double> valVector;

			if(massSpecDataSet->m_isStreamed)
			{
				newStreamedDriftSpectrum(massSpecDataSet, integrationType, &keyVector,
						&valVector, history, color);
			}
			else
			{ // that is, !if(massSpecDataSet->m_isStreamed)
				// At this point, we already know that we have to perform the work on
				// the basis
				// of data that are in the RAM already.

				// We are doing an integration RT_TO-DT, that is that we are selecting
				// retention times as defined by a retention time range :
				// [rtStart-rtEnd]. This is fine, but where are these data from ? Did
				// these data come from a previous integration or did they come right
				// over the initial RT ? Further, was there a single previous
				// integration or were there multiple ones ? And what integrations were
				// these ? MZ-based ? Or what ?

				// Let's see an example : start with the initial RT. Make a first
				// integration to MZ with a RT range. Then, in the MZ spectrum, make an
				// integration to DT with a MZ range. Finally, in the DT spectrum, make
				// an integration to MZ with a DT range. We need to always document what
				// we do and check each time all the previous integrations so as to
				// provide the expected result.

				// RT integration.
				// ===============
				// Already set at the top of this function.

				// MZ integration.
				// ===============
				bool mzIntegration = false;
				double mzStart = -1;
				double mzEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of m/z value ranges, then we should make sure we do take into
				// account the more internal m/z value range.
				mzIntegration = history.innermostMzRange(&mzStart, &mzEnd);

				if(mzIntegration)
				{
					mainWindow->logConsoleMessage(
							"m/z integration accounted for: \n" +
							QString("[%1-%2]\n").arg(mzStart).arg(mzEnd));
				}
				else
					qDebug() << __FILE__ << __LINE__ << "mzIntegration is true";


				// DT integration.
				// ===============
				bool dtIntegration = false;
				double dtStart = -1;
				double dtEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of dt value ranges, then we should make sure we do take into
				// account the more internal dt value range.
				dtIntegration = history.innermostDtRange(&dtStart, &dtEnd);

				if(dtIntegration)
				{
					mainWindow->logConsoleMessage(
							"dt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(dtStart).arg(dtEnd));
				}
				else
					qDebug() << __FILE__ << __LINE__ << "dtIntegration is true";

				// Get handles to the original data in the calling entity.
				MsMultiHash *rtHash = &massSpecDataSet->m_rtHash;

				// Get a list of all the retention times that are stored in the spec
				// data set and that are contained in the rt range passed as parameters.

				QList<double> rtList;

				rtHash->rangedKeys(rtList, rangeStart, rangeEnd);
				qSort(rtList.begin(), rtList.end());

				int rtListSize = rtList.size();

				qDebug() << __FILE__ << __LINE__
					<< "Number of unique retention time keys: " << rtListSize
					<< "from scan start time " << rtList.first()
					<< "to scan start time" << rtList.last();

				QString msgSkeleton =
					QString("Computing drift spectrum between rt %1 and rt %2")
					.arg(rangeStart, 0, 'f', 6)
					.arg(rangeEnd, 0, 'f', 6);

				setupProgressFeedback(msgSkeleton, 1, rtListSize);

				// For each drift time, we'll want to store the combined total ion
				// current. A map is a good option to increment previously existing
				// values or to just add new values. Then it will be easy to ask for a
				// list of keys (ordered) and of values (ordered according to the order
				// of the keys).
				QMap<double, double> dtTicMap;

				int iterationCount = 0;

#pragma omp parallel for

				for(int iter = 0; iter < rtListSize; ++iter)
				{

					if(m_isOperationCancelled)
						continue;

					double rt = rtList.at(iter);

					MassSpectrumList rtMassSpectra(rtHash->values(rt));

					if(dtIntegration)
					{
						// We have to remove from all the rtMassSpectra, all the spectra
						// that
						// are not in the drift time range.

						for(int jter = 0; jter < rtMassSpectra.size(); ++jter)
						{
							MassSpectrum *spectrum = rtMassSpectra.at(jter);

							// Old version when MassSpectrum did not have m_rt/m_dt
							// member
							// variables.
							// double dt = dtHash->key(spectrum);

							if(spectrum->m_dt < dtStart || spectrum->m_dt > dtEnd)
							{
								// Remove the mass spectrum from the rtMassSpectra
								// because it does
								// not have the proper dt value.
								rtMassSpectra.removeAt(jter);
								--jter;
							}
						}
					}

					// At this point, the rtMassSpectra only contain spectra that match
					// both the retention time range and the drift time range.

					for(int kter = 0; kter < rtMassSpectra.size(); ++kter)
					{
						MassSpectrum *spectrum = rtMassSpectra.at(kter);

						double newTic = 0.0;

						if(mzIntegration)
							newTic = spectrum->tic(mzStart, mzEnd);
						else
							newTic = spectrum->tic();

						// Had another spectrum by the same dt already been iterated in?
						double oldTic = dtTicMap.value(spectrum->m_dt, -1000);

						if(oldTic != -1000)
						{
							// A dt by the same value alreay existed, update the tic
							// value and
							// insert it, that will erase the old value, because the
							// hash is
							// NOT a multihash.
							newTic += oldTic;
						}

#pragma omp critical
						dtTicMap.insert(spectrum->m_dt, newTic);
					}

					QString msg = msgSkeleton +
						QString("-- rt %1 with %2 spectra")
						.arg(rt, 0, 'f', 6)
						.arg(rtMassSpectra.size());

#pragma omp critical
					{
						++iterationCount;

						// The label and progress bar widgets were setup by the master
						// thread,
						// so we can only modify them (actually emitting signals) from
						// that
						// same master thread:
						if(omp_get_thread_num() == 0)
							updateProgressFeedback(msg, iterationCount);
					}
				}

				// End of for(int iter = 0; iter < rtListSize; ++iter)

				keyVector = QVector<double>::fromList(dtTicMap.keys());
				valVector = QVector<double>::fromList(dtTicMap.values());
			}

			// End of
			// else // that is, !if(massSpecDataSet->m_isStreamed)

			double end = omp_get_wtime();

			QString msg = QString("Computed drift spectrum in %1 s").arg(end - start);

			// Show 100% completion of the work.
			setupProgressFeedback(msg, 0, 100, 100);

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			mainWindow->logConsoleMessage(msg);

			// qDebug() << __FILE__ << __LINE__ << msg;

			// In some seconds, hide the progress bar and set the label to an empty
			// text.

			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			// Sanity check

			if(keyVector.size() != valVector.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We may have gotten here because the operation was cancelled. In any case,
			// reset the boolean value.
			m_isOperationCancelled = false;

			// We need to plot that stuff into a new widget.

			// Create a text item to indicate the rt range.

			QString text = QString("%1\n%2 - retention time range : [%3-%4] ")
				.arg(massSpecDataSet->m_fileName)
				.arg(m_desc)
				.arg(rangeStart, 0, 'f', 6)
				.arg(rangeEnd, 0, 'f', 6);

			// Allocate the new plot widget in which the data are to be plot.

			DriftSpecPlotWidget *widget = static_cast<DriftSpecPlotWidget *>(
					addPlotWidget(keyVector, valVector, text, massSpecDataSet));

			// The history that we get as a parameter needs to be appended to the new
			// widget's history so that we have a full account of the history.

			widget->m_history.appendHistory(history, true /*remove duplicates*/);
			mainWindow->logConsoleMessage(widget->m_history.asText(
						"Plot widget history: \n", "End plot widget history:\n"));
			// And now only the innermost ranges of the whole history.
			mainWindow->logConsoleMessage(widget->m_history.innermostRangesAsText());

			// The plot widget reacts to the Qt::Key_Delete key by calling a function
			// that emits a set of signals, of which the destroyPlotWidget signal
			// actually indirectly triggers the destruction of the plot widget. This
			// signal is now connected to the main window where a function will call
			// for the destruction of all the target plot widgets, that is all the
			// plot widgets that were created as the result of an integration
			// computation of the plot widget currently being destroyed.

			connect(widget, &AbstractPlotWidget::destroyPlotWidget, mainWindow,
					&MainWindow::destroyAlsoAllTargetPlotWidgets);

			connect(widget, &AbstractPlotWidget::toggleMultiGraphGraphSignal,
					mainWindow, &MainWindow::toggleMultiGraphGraph);

			connect(widget, &DriftSpecPlotWidget::newMassSpectrumByDt,
					mainWindow->mp_massSpecWnd, &MassSpecWnd::newMassSpectrumByDt);

			connect(widget, &DriftSpecPlotWidget::newTicChromatogramByDt,
					mainWindow->mp_ticChromWnd, &TicChromWnd::newTicChromatogramByDt);


			// Finally make the mapping between the tic chrom window plot widget and
			// this new mass spec window plot widget.
			AbstractPlotWidget *senderPlotWidget =
				static_cast<AbstractPlotWidget *>(QObject::sender());

			if(senderPlotWidget == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			mainWindow->m_dataPlotWidgetRelationer.newRelation(
					massSpecDataSet, senderPlotWidget, widget,
					nullptr /* open spectra dialog list widget item */);

			// Ensure that there is a connection between this plot widget and this
			// containing window, such that if there is a requirement to replot, the
			// widget gets the signal.
			connect(this, &AbstractMultiPlotWnd::replotWithAxisRange, widget,
					&AbstractPlotWidget::replotWithAxisRange);

			// Make sure we get the signal requiring to redraw the background of the
			// widget, depending on it being focused or not.
			connect(this, &AbstractMultiPlotWnd::redrawPlotBackground, widget,
					&AbstractPlotWidget::redrawPlotBackground);
		}


	void
		DriftSpecWnd::newDriftSpectrumByMz(MassSpecDataSet *massSpecDataSet,
				IntegrationType integrationType,
				History history, QColor color)
		{
			if(massSpecDataSet == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// Check if the data are for a drift experiment, otherwise do not do
			// anything. Indeed, we are not in streamed mode, so the data should
			// already be there and if they were for a mobility experiment, then we
			// should be able to test it.

			if(massSpecDataSet->m_dtHash.isEmpty())
				return;

			m_wishedColor = color;

			// We get a History reference that informs us about the ranges in the
			// sense that the innermost range for integrationType is by necessity the
			// integration range we work on.

			double rangeStart;
			double rangeEnd;

			if(!history.innermostRange(integrationType, &rangeStart, &rangeEnd))
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			MainWindow *mainWindow = static_cast<MainWindow *>(parent());
			QString msgSkeleton =
				QString("Computing drift spectrum between m/z %1 and m/z %2")
				.arg(rangeStart, 0, 'f', 6)
				.arg(rangeEnd, 0, 'f', 6);
			mainWindow->logConsoleMessage(msgSkeleton);

#if 0
			// No more of use here at this is performed in the calling function.

			History localHistory = history;
			HistoryItem *histItem = new HistoryItem;
			histItem->newIntegrationRange(IntegrationType::MZ_TO_DT, rangeStart,
					rangeEnd);
			histItem->setDateTime();

			// Aggreate the new item to the history we got as a parameter and that
			// we copied into localHistory.
			localHistory.appendHistoryItem(histItem);
#endif

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);

			double start = omp_get_wtime();

			QVector<double> keyVector;
			QVector<double> valVector;

			if(massSpecDataSet->m_isStreamed)
			{
				newStreamedDriftSpectrum(massSpecDataSet, integrationType, &keyVector,
						&valVector, history, color);
			}
			else
			{ // that is, !if(massSpecDataSet->m_isStreamed)
				// At this point, we already know that we have to perform the work on
				// the basis
				// of data that are in the RAM already.

				// We are doing an integration MZ_TO_DT, that is that we are selecting
				// m/z values as defined by a m/z range : [rangeStart-rangeEnd]. This
				// is fine, but where are these data from ? Did these data come from a
				// previous integration or did they come right over the initial RT ?
				// Further, was there a single previous integration or were there
				// multiple ones ? And what integrations were these ? DT-based ? Or
				// what ?

				// Let's see an example : start with the initial RT. Make a first
				// integration to MZ with a RT range. Then, in the MZ spectrum, make an
				// integration to DT with a MZ range. Finally, in the DT spectrum, make
				// an integration to MZ with a DT range. We need to always document what
				// we do and check each time all the previous integrations so as to
				// provide
				// the expected result.

				// MZ integration.
				// ===============
				// Already set at the top of this function.

				// RT integration.
				// ===============
				bool rtIntegration = false;
				double rtStart = -1;
				double rtEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of retention time value ranges, then we should make sure we do
				// take into
				// account the more internal retention time value range.
				rtIntegration = history.innermostRtRange(&rtStart, &rtEnd);

				if(rtIntegration)
				{
					mainWindow->logConsoleMessage(
							"rt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(rtStart).arg(rtEnd));
				}

				// DT integration.
				// ===============
				bool dtIntegration = false;
				double dtStart = -1;
				double dtEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of dt value ranges, then we should make sure we do take into
				// account the more internal dt value range.
				dtIntegration = history.innermostDtRange(&dtStart, &dtEnd);

				if(dtIntegration)
				{
					mainWindow->logConsoleMessage(
							"dt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(dtStart).arg(dtEnd));
				}

				// Get handles to the original data in the calling entity.
				MsMultiHash *dtHash = &massSpecDataSet->m_dtHash;

				// Get the list of unique values of the drift times, which might exist
				// more than once if the acquisition was of more than a single mobility
				// cycle.
				QList<double> dtList;

				// If there had been a dt integration in the history, take that into
				// account by only selecting the dtHash keys that match that integration
				// range, otherwise get all the unique keys.
				if(dtIntegration)
					dtHash->rangedKeys(dtList, dtStart, dtEnd);
				else
					dtList = dtHash->uniqueKeys();

				// And now make sure all the keys are sorted fine, because we will bear
				// on
				// this sortedness laters.
				qSort(dtList.begin(), dtList.end());

				int dtListSize = dtList.size();

				// qDebug() << __FILE__ << __LINE__
				// << "Number of unique drigt time keys: " << dtListSize;

				setupProgressFeedback(msgSkeleton, 1, dtListSize);

				// Set aside some space to store all the total ion current values that
				// we
				// are going to compute for each drift time of the mass spectra's
				// MassPeak
				// entities that are in the [rangeStart--rangeEnd] range.
				QMap<double, double> dtTicMap;

				int iterationCount = 0;

#pragma omp parallel for

				for(int iter = 0; iter < dtListSize; ++iter)
				{

					if(m_isOperationCancelled)
						continue;

					double dt = dtList.at(iter);

					// Get the list of all the spectra that have been acquired for
					// currently iterated drift time dt.
					MassSpectrumList dtMassSpectra = dtHash->values(dt);

					if(rtIntegration)
					{
						// In the history that we got as parameter, there was a previous
						// retention time integration. We need to check that.
						// We have to remove from all the rtMassSpectra, all the spectra
						// that
						// are not in the drift time range.

						for(int jter = 0; jter < dtMassSpectra.size(); ++jter)
						{
							MassSpectrum *spectrum = dtMassSpectra.at(jter);

							// Old version when MassSpectrum did not have m_rt/m_dt
							// member
							// variables.
							// double rt = rtHash->key(spectrum);

							if(spectrum->m_rt < rtStart || spectrum->m_rt > rtEnd)
							{
								// Remove the mass spectrum from the dtMassSpectra
								// because it does
								// not have the proper rt value.

								// qDebug() << __FILE__ << __LINE__
								// << "Removing one spectrum on the basis of retention
								// time range criterium.";

								dtMassSpectra.removeAt(jter);
								--jter;
							}
						}
					}

					// At this point the dtMassSpectra only contain spectra that both
					// comply with the RT and DT restrictions. We'll need to ensure we now
					// comply with the MZ restrictions.

					// qDebug() << __FILE__ << __LINE__
					// << "At the end of the rtIntegration, remain" <<
					// dtMassSpectra.size() << "spectra";

					// Allocate a new mass spectrum into which to combine the mass spectra
					// obtained for a given dt.

					MassSpectrum combinedMassSpectrum;

					// We need to use a mz-specific combine function because each MassPeak
					// must be checked for its mz value that must be in between rangeStart
					// and rangeEnd to be combined.

					combinedMassSpectrum.combine(dtMassSpectra, rangeStart, rangeEnd);

					// At this point, we want to compute the tic of the combined spectrum.

					double tic = combinedMassSpectrum.tic();

					QString msg =
						msgSkeleton +
						QString("-- dt %1, combined mass spectrum has TIC value %2")
						.arg(dt, 0, 'f', 6)
						.arg(tic, 0, 'f', 0);

#pragma omp critical
					{
						dtTicMap.insert(dt, tic);

						++iterationCount;

						// The label and progress bar widgets were setup by the master
						// thread,
						// so we can only modify them (actually emitting signals) from
						// that
						// same master thread:
						if(omp_get_thread_num() == 0)
							updateProgressFeedback(msg, iterationCount);
					}
					// End of
					// #pragma omp critical
				}

				// End of
				// for(int iter = 0; iter < dtListSize; ++iter)

				keyVector = QVector<double>::fromList(dtTicMap.keys());
				valVector = QVector<double>::fromList(dtTicMap.values());
			}

			// End of
			// else // that is, !if(massSpecDataSet->m_isStreamed)

			double end = omp_get_wtime();

			QString msg = QString("Computed drift spectrum in %1 s").arg(end - start);

			// Show 100% completion of the work.
			setupProgressFeedback(msg, 0, 100, 100);

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			mainWindow->logConsoleMessage(msg);

			// qDebug() << __FILE__ << __LINE__ << msg;

			// In some seconds, hide the progress bar and set the label to an empty
			// text.

			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			// Sanity check

			if(keyVector.size() != valVector.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We may have gotten here because the operation was cancelled. In any case,
			// reset the boolean value.
			m_isOperationCancelled = false;

			// We need to plot that stuff into a new widget.

			// Create a text item to indicate the rt range.

			QString text = QString("%1\n%2 - m/z range : [%3-%4] ")
				.arg(massSpecDataSet->m_fileName)
				.arg(m_desc)
				.arg(rangeStart, 0, 'f', 6)
				.arg(rangeEnd, 0, 'f', 6);

			// Allocate the new plot widget in which the data are to be plot.
			DriftSpecPlotWidget *widget = static_cast<DriftSpecPlotWidget *>(
					addPlotWidget(keyVector, valVector, text, massSpecDataSet));

			// The history that we get as a parameter needs to be appended to the new
			// widget's history so that we have a full account of the history.

			widget->m_history.appendHistory(history, true /*remove duplicates*/);

			mainWindow->logConsoleMessage(widget->m_history.asText(
						"Plot widget history: \n", "End plot widget history:\n"));

			// And now only the innermost ranges of the whole history.

			mainWindow->logConsoleMessage(widget->m_history.innermostRangesAsText());

			// The plot widget reacts to the Qt::Key_Delete key by calling a function
			// that emits a set of signals, of which the destroyPlotWidget signal
			// actually indirectly triggers the destruction of the plot widget. This
			// signal is now connected to the main window where a function will call for
			// the destruction of all the target plot widgets, that is all the plot
			// widgets that were created as the result of an integration computation of
			// the plot widget currently being destroyed.

			connect(widget, &AbstractPlotWidget::destroyPlotWidget, mainWindow,
					&MainWindow::destroyAlsoAllTargetPlotWidgets);

			connect(widget, &AbstractPlotWidget::toggleMultiGraphGraphSignal,
					mainWindow, &MainWindow::toggleMultiGraphGraph);

			connect(widget, &DriftSpecPlotWidget::newMassSpectrumByDt,
					mainWindow->mp_massSpecWnd, &MassSpecWnd::newMassSpectrumByDt);

			connect(widget, &DriftSpecPlotWidget::newTicChromatogramByDt,
					mainWindow->mp_ticChromWnd, &TicChromWnd::newTicChromatogramByDt);


			// Finally make the mapping between the tic chrom window plot widget and
			// this new mass spec window plot widget.
			AbstractPlotWidget *senderPlotWidget =
				static_cast<AbstractPlotWidget *>(QObject::sender());

			if(senderPlotWidget == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			mainWindow->m_dataPlotWidgetRelationer.newRelation(
					massSpecDataSet, senderPlotWidget, widget,
					nullptr /* open spectra dialog list widget item */);

			// Ensure that there is a connection between this plot widget and this
			// containing window, such that if there is a requirement to replot, the
			// widget gets the signal.
			connect(this, &AbstractMultiPlotWnd::replotWithAxisRange, widget,
					&AbstractPlotWidget::replotWithAxisRange);

			// Make sure we get the signal requiring to redraw the background of the
			// widget, depending on it being focused or not.
			connect(this, &AbstractMultiPlotWnd::redrawPlotBackground, widget,
					&AbstractPlotWidget::redrawPlotBackground);
		}


	void
		DriftSpecWnd::newDriftSpectrumByDtAndMz(MassSpecDataSet *massSpecDataSet,
				IntegrationType integrationType,
				History history, QColor color)
		{
			if(massSpecDataSet == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			m_wishedColor = color;

			// We get a History reference that informs us about the ranges in the
			// sense that the innermost range for integrationType is by necessity the
			// integration range we work on. Note this is a particular case because we
			// need to get two pairs of values since we are called by signal emission
			// from the ColorMapPlotWidget, where two dimensions are simulatenously
			// taken into account: the DT and the MZ.

			double dtRangeStart;
			double dtRangeEnd;

			double mzRangeStart;
			double mzRangeEnd;

			if(!history.innermostRange(IntegrationType::DATA_TO_DT_MZ_DT_RANGE,
						&dtRangeStart, &dtRangeEnd))
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			if(!history.innermostRange(IntegrationType::DATA_TO_DT_MZ_MZ_RANGE,
						&mzRangeStart, &mzRangeEnd))
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			MainWindow *mainWindow = static_cast<MainWindow *>(parent());
			QString msgSkeleton =
				QString("Computing drift spectrum between m/z %1 and m/z %2")
				.arg(mzRangeStart, 0, 'f', 6)
				.arg(mzRangeEnd, 0, 'f', 6);
			mainWindow->logConsoleMessage(msgSkeleton);

#if 0
			// No more of use here at this is performed in the calling function.

			// Make a local copy of the history that was passed as parameter, so we
			// can modify it.

			History localHistory = history;

			HistoryItem *histItem = new HistoryItem;

			// Because we have integrated using both DT and MZ, we'll have to create
			// two integration ranges for the same history item.

			// First the DT integration range:
			histItem->newIntegrationRange(IntegrationType::DATA_TO_DT_MZ_DT_RANGE,
					dtStart, dtEnd);

			// And now the MZ integration range:
			histItem->newIntegrationRange(IntegrationType::DATA_TO_DT_MZ_MZ_RANGE,
					mzStart, mzEnd);

			// Aggreate the new item to the history we got as a parameter and that
			// we copied into localHistory.
			localHistory.appendHistoryItem(histItem);
#endif

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);

			double start = omp_get_wtime();

			QVector<double> keyVector;
			QVector<double> valVector;

			if(massSpecDataSet->m_isStreamed)
			{
				newStreamedDriftSpectrum(massSpecDataSet, integrationType, &keyVector,
						&valVector, history, color);
			}
			else
			{

				// that is, !if(massSpecDataSet->m_isStreamed) At this point, we already
				// know that we have to perform the work on the basis of data that are
				// in the RAM already.

				// We are doing an integration MZ_TO_DT, that is that we are selecting
				// m/z values as defined by a m/z range : [mzStart-mzEnd]. This is
				// fine, but where are these data from ? Did these data come from a
				// previous integration or did they come right over the initial RT ?
				// Further, was there a single previous integration or were there
				// multiple ones ? And what integrations were these ? DT-based ? Or
				// what ?

				// Let's see an example : start with the initial RT. Make a first
				// integration to MZ with a RT range. Then, in the MZ spectrum, make an
				// integration to DT with a MZ range. Finally, in the DT spectrum, make
				// an integration to MZ with a DT range. We need to always document what
				// we do and check each time all the previous integrations so as to
				// provide the expected result.

				// MZ integration.
				// ===============
				// Already set at the top of this function.

				// DT integration.
				// ===============
				// Already set at the top of this function.

				// RT integration.
				// ===============
				bool rtIntegration = false;
				double rtStart = -1;
				double rtEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of retention time value ranges, then we should make sure we do
				// take into
				// account the more internal retention time value range.
				rtIntegration = history.innermostRtRange(&rtStart, &rtEnd);

				if(rtIntegration)
				{
					mainWindow->logConsoleMessage(
							"rt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(rtStart).arg(rtEnd));
				}

				// Get handles to the original data in the calling entity.
				MsMultiHash *dtHash = &massSpecDataSet->m_dtHash;

				// Get the list of unique values of the drift times, which might exist
				// more than once if the acquisition was of more than a single mobility
				// cycle.
				QList<double> dtList;

				dtHash->rangedKeys(dtList, dtRangeStart, dtRangeEnd);

				// And now make sure all the keys are sorted fine, because we will bear
				// on
				// this sortedness laters.
				qSort(dtList.begin(), dtList.end());

				int dtListSize = dtList.size();

				// qDebug() << __FILE__ << __LINE__
				// << "Number of unique drigt time keys: " << dtListSize;

				setupProgressFeedback(msgSkeleton, 1, dtListSize);

				// Set aside some space to store all the total ion current values that
				// we
				// are going to compute for each drift time of the mass spectra's
				// MassPeak
				// entities that are in the [mzStart--mzEnd] range.
				QMap<double, double> dtTicMap;

				int iterationCount = 0;

#pragma omp parallel for

				for(int iter = 0; iter < dtListSize; ++iter)
				{

					if(m_isOperationCancelled)
						continue;

					double dt = dtList.at(iter);

					// Get the list of all the spectra that have been acquired for
					// currently iterated drift time dt.
					MassSpectrumList dtMassSpectra = dtHash->values(dt);

					if(rtIntegration)
					{
						// In the history that we got as parameter, there was a previous
						// retention time integration. We need to check that.
						// We have to remove from all the rtMassSpectra, all the spectra
						// that
						// are not in the drift time range.

						for(int jter = 0; jter < dtMassSpectra.size(); ++jter)
						{
							MassSpectrum *spectrum = dtMassSpectra.at(jter);

							// Old version when MassSpectrum did not have m_rt/m_dt
							// member
							// variables.

							// double rt = rtHash->key(spectrum);

							if(spectrum->m_rt < rtStart || spectrum->m_rt > rtEnd)
							{
								// Remove the mass spectrum from the dtMassSpectra
								// because it does
								// not have the proper rt value.

								// qDebug() << __FILE__ << __LINE__
								// << "Removing one spectrum on the basis of retention
								// time range criterium.";

								dtMassSpectra.removeAt(jter);
								--jter;
							}
						}
					}

					// At this point the dtMassSpectra only contain spectra that both
					// comply
					// with the RT and DT restrictions. We'll need to ensure we now
					// comply
					// with the MZ restrictions.

					// qDebug() << __FILE__ << __LINE__
					// << "At the end of the rtIntegration, remain" <<
					// dtMassSpectra.size() << "spectra";

					// Allocate a new mass spectrum into which to combine the mass
					// spectra
					// obtained for a given dt.
					MassSpectrum combinedMassSpectrum;

					// We need to use a mz-specific combine function because each
					// MassPeak
					// must be checked for its mz value that must be in between mzStart
					// and
					// mzEnd to be combined.
					combinedMassSpectrum.combine(dtMassSpectra, mzRangeStart,
							mzRangeEnd);

					// At this point, we want to compute the tic of the combined
					// spectrum.
					double tic = combinedMassSpectrum.tic();

					QString msg =
						msgSkeleton +
						QString("-- dt %1, combined mass spectrum has TIC value %2")
						.arg(dt, 0, 'f', 6)
						.arg(tic, 0, 'f', 0);

#pragma omp critical
					{
						dtTicMap.insert(dt, tic);

						++iterationCount;

						// The label and progress bar widgets were setup by the master
						// thread,
						// so we can only modify them (actually emitting signals) from
						// that
						// same master thread:
						if(omp_get_thread_num() == 0)
							updateProgressFeedback(msg, iterationCount);
					}
				}

				// End of
				// for(int iter = 0; iter < dtListSize; ++iter)

				keyVector = QVector<double>::fromList(dtTicMap.keys());
				valVector = QVector<double>::fromList(dtTicMap.values());
			}
			// End of
			// else // that is, !if(massSpecDataSet->m_isStreamed)

			double end = omp_get_wtime();

			QString msg = QString("Computed drift spectrum in %1 s").arg(end - start);

			// Show 100% completion of the work.
			setupProgressFeedback(msg, 0, 100, 100);

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			mainWindow->logConsoleMessage(msg);

			// qDebug() << __FILE__ << __LINE__ << msg;

			// In some seconds, hide the progress bar and set the label to an empty
			// text.

			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			// Sanity check
			if(keyVector.size() != valVector.size())
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We may have gotten here because the operation was cancelled. In any case,
			// reset the boolean value.
			m_isOperationCancelled = false;

			// We need to plot that stuff into a new widget.

			// Create a text item to indicate the rt range.

			QString text = QString("%1\n%2 - m/z range : [%3-%4] ")
				.arg(massSpecDataSet->m_fileName)
				.arg(m_desc)
				.arg(mzRangeStart, 0, 'f', 6)
				.arg(mzRangeEnd, 0, 'f', 6);

			// Allocate the new plot widget in which the data are to be plot.

			DriftSpecPlotWidget *widget = static_cast<DriftSpecPlotWidget *>(
					addPlotWidget(keyVector, valVector, text, massSpecDataSet));

			// The history that we get as a parameter needs to be appended to the new
			// widget's history so that we have a full account of the history.

			widget->m_history.appendHistory(history, true /*remove duplicates*/);

			mainWindow->logConsoleMessage(widget->m_history.asText(
						"Plot widget history: \n", "End plot widget history:\n"));

			// And now only the innermost ranges of the whole history.

			mainWindow->logConsoleMessage(widget->m_history.innermostRangesAsText());

			// The plot widget reacts to the Qt::Key_Delete key by calling a function
			// that emits a set of signals, of which the destroyPlotWidget signal
			// actually indirectly triggers the destruction of the plot widget. This
			// signal is now connected to the main window where a function will call for
			// the destruction of all the target plot widgets, that is all the plot
			// widgets that were created as the result of an integration computation of
			// the plot widget currently being destroyed.

			connect(widget, &AbstractPlotWidget::destroyPlotWidget, mainWindow,
					&MainWindow::destroyAlsoAllTargetPlotWidgets);

			connect(widget, &AbstractPlotWidget::toggleMultiGraphGraphSignal,
					mainWindow, &MainWindow::toggleMultiGraphGraph);

			connect(widget, &DriftSpecPlotWidget::newMassSpectrumByDt,
					mainWindow->mp_massSpecWnd, &MassSpecWnd::newMassSpectrumByDt);

			connect(widget, &DriftSpecPlotWidget::newTicChromatogramByDt,
					mainWindow->mp_ticChromWnd, &TicChromWnd::newTicChromatogramByDt);

			// Finally make the mapping between the tic chrom window plot widget and
			// this new mass spec window plot widget.

			AbstractPlotWidget *senderPlotWidget =
				static_cast<AbstractPlotWidget *>(QObject::sender());

			if(senderPlotWidget == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			mainWindow->m_dataPlotWidgetRelationer.newRelation(
					massSpecDataSet, senderPlotWidget, widget,
					nullptr /* open spectra dialog list widget item */);

			// Ensure that there is a connection between this plot widget and this
			// containing window, such that if there is a requirement to replot, the
			// widget gets the signal.
			connect(this, &AbstractMultiPlotWnd::replotWithAxisRange, widget,
					&AbstractPlotWidget::replotWithAxisRange);

			// Make sure we get the signal requiring to redraw the background of the
			// widget, depending on it being focused or not.
			connect(this, &AbstractMultiPlotWnd::redrawPlotBackground, widget,
					&AbstractPlotWidget::redrawPlotBackground);
		}


	// We are not going to produce any graph, we are integrating the data to a
	// single TIC intensity value.
	void 
		DriftSpecWnd::ticIntensity(MassSpecDataSet *massSpecDataSet,
				double *intensity, IntegrationType integrationType, History history)
		{
			if(massSpecDataSet == Q_NULLPTR)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			if(intensity == Q_NULLPTR)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We get a History reference that informs us about the ranges in the
			// sense that the innermost range for integrationType is by necessity the
			// integration range we work on.

			double rangeStart;
			double rangeEnd;

			if(!history.innermostRange(integrationType, &rangeStart, &rangeEnd))
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			MainWindow *mainWindow = static_cast<MainWindow *>(parent());
			QString msgSkeleton =
				QString("Computing TIC intensity between dt %1 and dt %2")
				.arg(rangeStart, 0, 'f', 6)
				.arg(rangeEnd, 0, 'f', 6);
			mainWindow->logConsoleMessage(msgSkeleton);

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);

			double start = omp_get_wtime();

			if(massSpecDataSet->m_isStreamed)
			{
				streamedTicIntensity(massSpecDataSet, 
						intensity, integrationType, history);
			}
			else // that is, !if(massSpecDataSet->m_isStreamed)
			{

				// RT integration.
				// ===============
				bool rtIntegration = false;
				double rtStart = -1;
				double rtEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of rt value ranges, then we should make sure we do take into
				// account the more internal rt value range.
				rtIntegration = history.innermostRtRange(&rtStart, &rtEnd);
				if(rtIntegration)
				{
					mainWindow->logConsoleMessage(
							"rt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(rtStart).arg(rtEnd));
				}

				// MZ integration.
				// ===============
				bool mzIntegration = false;
				double mzStart = -1;
				double mzEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of mz value ranges, then we should make sure we do take into
				// account the more internal mz value range.
				mzIntegration = history.innermostMzRange(&mzStart, &mzEnd);
				if(mzIntegration)
				{
					mainWindow->logConsoleMessage(
							"mz integration accounted for: \n" +
							QString("[%1-%2]\n").arg(mzStart).arg(mzEnd));
				}				

				// DT integration.
				// ===============
				bool dtIntegration = false;
				double dtStart = -1;
				double dtEnd = -1;

				// If the history is comprised of many successive integrations on the
				// basis of dt value ranges, then we should make sure we do take into
				// account the more internal dt value range.
				dtIntegration = history.innermostDtRange(&dtStart, &dtEnd);
				if(dtIntegration)
				{
					mainWindow->logConsoleMessage(
							"dt integration accounted for: \n" +
							QString("[%1-%2]\n").arg(dtStart).arg(dtEnd));
				}

				// Get handles to the original data in the calling entity.
				MsMultiHash *rtHash = &massSpecDataSet->m_rtHash;

				// One way to start the computation is simply to define which retention
				// times should go in the Tic intensity calculation.

				QList<double> rtList;
				if(rtIntegration)
					rtHash->rangedKeys(rtList, rtStart, rtEnd);
				else
					rtList = rtHash->uniqueKeys();

				qSort(rtList.begin(), rtList.end());

				int rtListSize = rtList.size();

				// qDebug() << __FILE__ << __LINE__
				// << "Number of unique retention time keys: " << rtListSize
				// << "from scan start time "<< rtList.first()
				// << "to scan start time" << rtList.last();

				setupProgressFeedback(msgSkeleton, 1, rtListSize);

				// For each retention time, we'll want to
				// 1. get the corresponding mass spectra
				// 2. only retain the spectra that match the drift time (if applicable)
				// 3. compute a TIC value for the retained spectra that is summed 
				// to all the preceding ones, to finally yield a total Tic intensity
				// value (that actually is simply an integration).

				double localIntensity = 0.0;
				int iterationCount = 0;

#pragma omp parallel for
				for(int iter = 0; iter < rtListSize; ++iter)
				{

					if(m_isOperationCancelled)
						continue;

					double rt = rtList.at(iter);

					MassSpectrumList rtMassSpectra(rtHash->values(rt));

					if(dtIntegration)
					{

						// We have to remove from all the rtMassSpectra, all the spectra
						// that are not in the drift time range.

						for(int jter = 0; jter < rtMassSpectra.size(); ++jter)
						{
							MassSpectrum *spectrum = rtMassSpectra.at(jter);

							// Old version when MassSpectrum did not have m_rt/m_dt
							// member
							// variables.
							// double dt = dtHash->key(spectrum);

							if(spectrum->m_dt < dtStart || spectrum->m_dt > dtEnd)
							{
								// Remove the mass spectrum from the rtMassSpectra
								// because it does
								// not have the proper dt value.
								rtMassSpectra.removeAt(jter);
								--jter;
								continue;
							}
						}
					}

					// At this point, the rtMassSpectra only contain spectra that match
					// both the retention time range and the drift time range.

					for(int kter = 0; kter < rtMassSpectra.size(); ++kter)
					{
						double newTic = 0.0;

						if(mzIntegration)
							newTic = rtMassSpectra.at(kter)->tic(mzStart, mzEnd);
						else
							newTic = rtMassSpectra.at(kter)->tic();

#pragma omp critical
						localIntensity += newTic;
					}
					// End of for(int kter = 0; kter < massSpectra.size(); ++kter)

					QString msg = msgSkeleton +
						QString("-- rt %1 with %2 spectra")
						.arg(rt, 0, 'f', 6)
						.arg(rtMassSpectra.size());

#pragma omp critical
					{
						++iterationCount;

						// The label and progress bar widgets were setup by the master
						// thread, so we can only modify them (actually emitting signals)
						// from that same master thread:

						if(omp_get_thread_num() == 0)
							updateProgressFeedback(msg, iterationCount);
					}
				}
				// End of
				// #pragma omp parallel for
				// for(int iter = 0; iter < rtList.size(); ++iter)

				// Finally, update the value passed to the function:
				*intensity = localIntensity;
			}
			// End of
			// else // that is, !if(massSpecDataSet->m_isStreamed)

			double end = omp_get_wtime();

			double duration = end - start;

			QString msg = QString("Took %1 s to compute TIC intensity: %2\n")
				.arg(duration)
				.arg(*intensity, 0, 'g', 3);

			// Show 100% completion of the work.
			setupProgressFeedback(msg, 0, 100, 100);

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			mainWindow->logConsoleMessage(msg);

			// qDebug() << __FILE__ << __LINE__ << msg;

			// Pretty quickly hide the progress bar and set the label to an empty
			// text so that the user can see the TIC I value on the statusbar.

			// The idea is that if the computation is very quick, then the message
			// should vanish almost immediately. Otherwise let it there longer to show
			// to the user that the computation is computing-intensive.

			// For every second of integration, show the message 20 ms. But do not
			// exceed 3000 ms.
			int timeOut = duration * 20; 
			if(timeOut > 3000)
				timeOut = 3000;

			QTimer::singleShot(timeOut, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			// We may have gotten here because the operation was cancelled. In any
			// case, reset the boolean value.

			m_isOperationCancelled = false;
		}

	void
		DriftSpecWnd::newStreamedDriftSpectrum(MassSpecDataSet *massSpecDataSet,
				IntegrationType integrationType,
				QVector<double> *keyVector,
				QVector<double> *valVector,
				History history, QColor color)
		{
			// Depending on the format of the data file from which the data were
			// loaded in stream mode, it will be possible or not to perform the
			// required integration.

			if(massSpecDataSet->m_fileFormat ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_SQLITE3)
			{
				MassSpecDataFileLoaderSqlite3 loader(massSpecDataSet->m_fileName);

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderSqlite3::setupProgressFeedbackSignal,
						this, &AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderSqlite3::updateProgressFeedbackSignal,
						this, &AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal, &loader,
						&MassSpecDataFileLoaderSqlite3::cancelOperation);

				int res = loader.streamedIntegration(
						history, IntegrationType::ANY_TO_DT, keyVector, valVector);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else if(massSpecDataSet->m_fileFormat ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_MZML)
			{
				MassSpecDataFileLoaderPwiz loader(massSpecDataSet->m_fileName);

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderPwiz::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderPwiz::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal, &loader,
						&MassSpecDataFileLoaderPwiz::cancelOperation);


				int res = loader.streamedIntegration(
						history, IntegrationType::ANY_TO_DT, keyVector, valVector);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else
			{
				QMessageBox msgBox;
				msgBox.setText(
						"The operation cannot be carried over because the "
						"data were loaded in stream and not from a compatible file "
						"(PWiz-loaded or SQLite3.");

				msgBox.setStandardButtons(QMessageBox::Ok);
				msgBox.exec();

				// Hide the cancel button.
				mp_statusBarCancelPushButton->setVisible(false);

				return;
			}
		}

	void
		DriftSpecWnd::streamedTicIntensity(MassSpecDataSet *massSpecDataSet,
				double *intensity,IntegrationType integrationType, History history)
		{
			if(intensity == Q_NULLPTR)
				qFatal("Fatal error at %s@%d. Program aborted.",
						__FILE__, __LINE__);

			// Depending on the format of the data file from which the data were
			// loaded in stream mode, it will be possible or not to perform the
			// required integration.

			if(massSpecDataSet->m_fileFormat ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_SQLITE3)
			{
				MassSpecDataFileLoaderSqlite3 loader(massSpecDataSet->m_fileName);

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderSqlite3::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderSqlite3::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal, &loader,
						&MassSpecDataFileLoaderSqlite3::cancelOperation);

				int res = loader.streamedTicIntensity(history, intensity);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else if(massSpecDataSet->m_fileFormat ==
					MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_MZML)
			{
				MassSpecDataFileLoaderPwiz loader(massSpecDataSet->m_fileName);

				// Make sure we connect the feedback signal/slot
				connect(&loader,
						&MassSpecDataFileLoaderPwiz::setupProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::setupProgressFeedback);

				connect(&loader,
						&MassSpecDataFileLoaderPwiz::updateProgressFeedbackSignal, this,
						&AbstractMultiPlotWnd::updateProgressFeedback);

				connect(this, &AbstractMultiPlotWnd::cancelOperationSignal,
						&loader, &MassSpecDataFileLoaderPwiz::cancelOperation);

				int res = loader.streamedTicIntensity(history, intensity);

				if(res == -1)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}
			else
			{
				QMessageBox msgBox;
				msgBox.setText(
						"The operation cannot be carried over because the "
						"data were loaded in stream and not from a usable file.");

				msgBox.setStandardButtons(QMessageBox::Ok);
				msgBox.exec();

				// Hide the cancel button.
				mp_statusBarCancelPushButton->setVisible(false);

				return;
			}
		}


	void
		DriftSpecWnd::recordAnalysisStanza(const QString &stanza, const QColor &color)
		{
			// The record might go to a file or to the console or both.

			if((mp_analysisPreferences->m_recordTarget &
						RecordTarget::RECORD_TO_FILE) == RecordTarget::RECORD_TO_FILE)
			{
				if(mp_analysisFile != nullptr)
				{

					// Write the stanza, that was crafted by the calling plot widget to
					// the file.

					if(!mp_analysisFile->open(QIODevice::Append))
					{
						statusBar()->showMessage(
								QString("Could not record the step because "
									"the file could not be opened."),
								4000);
					}
					else
					{
						mp_analysisFile->write(stanza.toLatin1());
						// Force writing because we may want to have tail -f work fine
						// on the
						// file, and see modifications live to change fiels in a text
						// editor.
						mp_analysisFile->flush();
						mp_analysisFile->close();
					}
				}
				else
				{
					qDebug() << __FILE__ << __LINE__
						<< "The mp_analysisFile pointer is nullptr.";

					statusBar()->showMessage(
							QString("Could not record the analysis step to file. "
								"Please define a file to write the data to."),
							4000);
				}
			}

			// Also, if recording to the console is asked for, then do that also.
			if((mp_analysisPreferences->m_recordTarget &
						RecordTarget::RECORD_TO_CONSOLE) == RecordTarget::RECORD_TO_CONSOLE)
			{
				MainWindow *mainWindow = static_cast<MainWindow *>(parent());
				mainWindow->logConsoleMessage(stanza, color);
			}

			// Also, if recording to the clipboard is asked for, then do that also.
			if((mp_analysisPreferences->m_recordTarget &
						RecordTarget::RECORD_TO_CLIPBOARD) == RecordTarget::RECORD_TO_CLIPBOARD)
			{
				QClipboard *clipboard = QApplication::clipboard();
				clipboard->setText(clipboard->text() + stanza, QClipboard::Clipboard);
			}

			return;
		}


	bool
		DriftSpecWnd::focusNextPrevChild(bool next)
		{
			// In this window, we do not want that the TAB key serves to move from
			// widget
			// to widget.
			return false;
		}


} // namespace msxpertsuite
