/*   msXpertSuite - mass spectrometry software suite
 *   -----------------------------------------------
 *   Copyright(C) 2016 Filippo Rusconi
 *
 *   http://www.{site}.org
 *
 *   This file is part of the msXpertSuite project.
 *
 *   The msXpertSuite project is the successor of the massXpert project. This
 *   project now includes various independent modules:
 *   - massXpert, the original mass spectrometry environment software (greatly
 *     refactored code);
 *   - mineXpert, a powerful TIC chromatogram/mass spectrum viewer/miner;
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <omp.h>
#include <cstring>
#include <iostream>

////////////////////// Qt includes
#include <QtGui>
#include <QtGlobal>
#include <QMessageBox>
#include <QFileDialog>
#include <QMenuBar>
#include <QAction>
#include <QColorDialog>
#include <QInputDialog>
#include <QThread>
#include <QClipboard>


//#include "qcustomplot-fr.hpp"
#include "qcustomplot.h"

/////////////////////// Local includes
#include "MainWindow.hpp"
#include "globals.hpp"
#include "AboutMineXpertDlg.hpp"
#include "Application.hpp"
#include "MassSpectrum.hpp"

#include "AbstractMultiPlotWnd.hpp"

#include "MassSpecDataFileFormatAnalyzer.hpp"
#include "MassSpecDataFileLoaderSqlite3.hpp"
#include "MassSpecDataFileLoaderPwiz.hpp"
//#include "MassSpecDataFileLoaderMzml.hpp"
#include "MassSpecDataFileLoaderPwiz.hpp"
#include "MassSpecDataFileLoaderMs1.hpp"
#include "MassSpecDataFileLoaderXy.hpp"
#include "MassSpecDataFileLoaderBrukerXy.hpp"
#include "MassSpecDataFileLoaderDx.hpp"
#include "ScriptingWnd.hpp"



namespace msXpertSuite
{

	MainWindow::MainWindow(const QString &moduleName) 
		:
			m_moduleName{moduleName}
	{
		if(!setupWindow())
			qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
	}

	MainWindow::~MainWindow()
	{
		writeSettings();

		// Close the analysis file.
		m_analysisFile.close();

		// The AnalysisPreferences were allocated on the heap.
		delete mpa_analysisPreferences;

		while(!m_colorList.isEmpty())
			delete m_colorList.takeFirst();

		while(!m_usedColorList.isEmpty())
			delete m_usedColorList.takeFirst();
	}

	void
		MainWindow::closeEvent(QCloseEvent *event)
		{
			writeSettings();
			event->accept();
		}

	void 
		MainWindow::quit()
		{
			close();
		}

	bool
		MainWindow::setupWindow()
		{

			// Create the actions and menus of the main window.
			createMenusAndActions();

			QPixmap pixmap(":/images/msxpertsuite-icon-32.png");
			QIcon icon(pixmap);
			setWindowIcon(icon);

			setAttribute(Qt::WA_DeleteOnClose);
			statusBar()->setSizeGripEnabled(true);

			mp_statusBarLabel = new QLabel(this);
			statusBar()->addPermanentWidget(mp_statusBarLabel);
			mp_statusBarLabel->setVisible(false);

			mp_statusBarProgressBar = new QProgressBar(this);
			mp_statusBarProgressBar->setAlignment(Qt::AlignLeft);

			statusBar()->addPermanentWidget(mp_statusBarProgressBar);
			mp_statusBarProgressBar->setVisible(false);

			// We want to have a cancel push button available.
			const QIcon cancelIcon = QIcon(":/images/red-cross-cancel.png");
			mp_statusBarCancelPushButton = new QPushButton(cancelIcon, "", this);
			connect(mp_statusBarCancelPushButton, &QPushButton::clicked, this,
					&MainWindow::cancelPushButtonClicked);
			statusBar()->addPermanentWidget(mp_statusBarCancelPushButton);
			mp_statusBarCancelPushButton->setVisible(false);

			readSettings();

			initialize();

			// Cannot initialize scripting before initialize() because all the windows
			// (including the scripting one) must have been created at the time we
			// initialize scripting (we do export to the scripting environment all the
			// window objects).

			initializeScripting();

			return true;
		}


	bool
		MainWindow::initialize()
		{
			// Allocate all the main windows

			mp_consoleWnd = new ConsoleWnd(dynamic_cast<QWidget *>(this), "mineXpert");

			mp_scriptingWnd = new ScriptingWnd(dynamic_cast<QWidget *>(this), "mineXpert");

			mp_ticChromWnd =
				new TicChromWnd(dynamic_cast<QWidget *>(this));

			// When a plot is destroyed, let this window know it is asking for the
			// plot color to be recycled.
			connect(mp_ticChromWnd, &AbstractMultiPlotWnd::recyclePlottingColor, this,
					&MainWindow::recyclePlottingColor);

			mp_massSpecWnd = new MassSpecWnd(dynamic_cast<QWidget *>(this));

			// When a plot is destroyed, let this window know it is asking for the
			// plot color to be recycled.
			connect(mp_massSpecWnd, &AbstractMultiPlotWnd::recyclePlottingColor, this,
					&MainWindow::recyclePlottingColor);

			mp_driftSpecWnd = new DriftSpecWnd(dynamic_cast<QWidget *>(this));

			// When a plot is destroyed, let this window know it is asking for the
			// plot color to be recycled.
			connect(mp_driftSpecWnd, &AbstractMultiPlotWnd::recyclePlottingColor, this,
					&MainWindow::recyclePlottingColor);

			mp_colorMapWnd = new ColorMapWnd(dynamic_cast<QWidget *>(this));

			// When a plot is destroyed, let this window know it is asking for the
			// plot color to be recycled.
			connect(mp_colorMapWnd, &AbstractMultiPlotWnd::recyclePlottingColor, this,
					&MainWindow::recyclePlottingColor);

			// The dialog window that lists all the open spectra.
			mp_openSpectraDlg = new OpenSpectraDlg(dynamic_cast<QWidget *>(this));
			mp_openSpectraDlg->show();
			// mp_openSpectraDlg->raise();
			mp_openSpectraDlg->activateWindow();


			m_colorList.append(new QColor(QColor().fromRgb(0xff0000)));
			m_colorList.append(new QColor(QColor().fromRgb(0x185e00)));
			m_colorList.append(new QColor(QColor().fromRgb(0x0000ff)));
			m_colorList.append(new QColor(QColor().fromRgb(0xffa400)));
			m_colorList.append(new QColor(QColor().fromRgb(0x00ffd7)));
			m_colorList.append(new QColor(QColor().fromRgb(0x7a00ff)));
			m_colorList.append(new QColor(QColor().fromRgb(0xff009c)));
			m_colorList.append(new QColor(QColor().fromRgb(0x0072ff)));
			m_colorList.append(new QColor(QColor().fromRgb(0x19ff00)));
			m_colorList.append(new QColor(QColor().fromRgb(0xf08080)));
			m_colorList.append(new QColor(QColor().fromRgb(0x4682b4)));
			m_colorList.append(new QColor(QColor().fromRgb(0x556b2f)));
			m_colorList.append(new QColor(QColor().fromRgb(0xff00ff)));
			m_colorList.append(new QColor(QColor().fromRgb(0x7dbee7)));

			return true;
		}


	void
		MainWindow::initializeScripting()
		{

			// Make available to the scripting environment all the entities that will
			// be useful to script mineXpert.

			// Placeholder, although at the moment we do not use it.
			QScriptValue scriptValue;

			// mainWindow/mainWnd

			mp_scriptingWnd->publishQObject(this, Q_NULLPTR /* parent QObject */,
					"mainWindow", 
					"mainWnd", 
					"Make available the main window under name "
					"\"mainWindow\" with alias \"mainWnd\".\n",
					"The main window is the program main window from which all ther other "
					"windows depend (the TIC chromatogram, mass spectrum, drift spectrum, color map, "
					"console, scripting console windows, for example).",
					scriptValue,
					QScriptEngine::QtOwnership, 
					QScriptEngine::ExcludeSuperClassContents |
					QScriptEngine::ExcludeDeleteLater |
					QScriptEngine::ExcludeSlots);

			// mainWindow.quit()

			mp_scriptingWnd->runScript("var quit = function() {mainWindow.quit()};",
					"Make available \"mainWindow.quit() under alias quit().\n");


			// consoleWnd 

			mp_scriptingWnd->publishQObject(mp_consoleWnd, Q_NULLPTR /* parent QObject */,
					"consoleWnd", 
					QString(), 
					"Make available the console window under name \"consoleWnd\".\n",
					"The console window displays all the useful messages about the program "
					"performing the various actions required by the user, thus providing "
					"useful feedback to the user",
					scriptValue,
					QScriptEngine::QtOwnership, 
					QScriptEngine::ExcludeSuperClassContents |
					QScriptEngine::ExcludeDeleteLater |
					QScriptEngine::ExcludeSlots);

			// ticChromWnd 

			mp_scriptingWnd->publishQObject(mp_ticChromWnd, Q_NULLPTR /* parent QObject */, 
					"ticChromWnd", 
					QString(), 
					"Make available the TIC chromatogram window under name \"ticChromWnd\".\n",
					"The TIC chromatogram window displays the total ion current chromatogram for "
					"the opened mass spectrometry files. It can also show TIC chromatograms obtained "
					"after an integration operation; in which case it is a XIC (extracted ion current) "
					"chromatogram",
					scriptValue,
					QScriptEngine::QtOwnership, 
					QScriptEngine::ExcludeSuperClassContents |
					QScriptEngine::ExcludeDeleteLater |
					QScriptEngine::ExcludeSlots);

			// massSpecWnd 

			mp_scriptingWnd->publishQObject(mp_massSpecWnd, Q_NULLPTR /* parent QObject */, 
					"massSpecWnd", 
					QString(), 
					"Make available the mass spectrum window under name \"massSpecWnd\".\n",
					"The mass spectrum window displays the mass spectra obtained as a result "
					"of an integration either from the TIC chromatogram, the drift spectrum or "
					"the color map.",
					scriptValue,
					QScriptEngine::QtOwnership, 
					QScriptEngine::ExcludeSuperClassContents |
					QScriptEngine::ExcludeDeleteLater |
					QScriptEngine::ExcludeSlots);

			// driftSpecWnd 

			mp_scriptingWnd->publishQObject(mp_driftSpecWnd, Q_NULLPTR /* parent QObject */, 
					"driftSpecWnd", 
					QString(), 
					"Make available the drift spectrum window under name \"driftSpecWnd\".\n",
					"The drift spectrum displays the drift spectra obtained as a result "
					"of an integration either from the TIC chromatogram, the mass spectrum or "
					"the color map.",
					scriptValue,
					QScriptEngine::QtOwnership, 
					QScriptEngine::ExcludeSuperClassContents |
					QScriptEngine::ExcludeDeleteLater |
					QScriptEngine::ExcludeSlots);

			// colorMapWnd 

			mp_scriptingWnd->publishQObject(mp_colorMapWnd, Q_NULLPTR /* parent QObject */, 
					"colorMapWnd", 
					QString(), 
					"Make available the color map window under name \"colorMapWnd\".\n",
					"The color map window displays the mass/drift spectra in the form of a map "
					"with the drift time on the abscissa and the m/z data on the ordinates. The "
					"intensity of the ion current is color-coded.",
					scriptValue,
					QScriptEngine::QtOwnership, 
					QScriptEngine::ExcludeSuperClassContents |
					QScriptEngine::ExcludeDeleteLater |
					QScriptEngine::ExcludeSlots);

			// openSpectraDlg 

			mp_scriptingWnd->publishQObject(mp_openSpectraDlg, Q_NULLPTR /* parent QObject */, 
					"openSpectraDlg", 
					QString(), 
					"Make available the open spectra dialog under name \"openSpectraDlg\".\n",
					"The open spectra dialog window displays all the mass spectra that are "
					"currently opened in the program. It is possible to perform actions on the "
					"selected items, like showing/hiding/closing...",
					scriptValue,
					QScriptEngine::QtOwnership, 
					QScriptEngine::ExcludeSuperClassContents |
					QScriptEngine::ExcludeDeleteLater |
					QScriptEngine::ExcludeSlots);

			// data plot relationer

			mp_scriptingWnd->publishQObject(&m_dataPlotWidgetRelationer, this,
					"plotRelationer", 
					QString(), 
					"Make available the plot relationer under name \"dataPlotRelation\" with alias \"relationer\".\n",
					"The data plot widget relationer is a faciity that allows one to trace the "
					"parentship between plot widgets.",
					scriptValue,
					QScriptEngine::QtOwnership, 
					QScriptEngine::ExcludeSuperClassContents |
					QScriptEngine::ExcludeDeleteLater |
					QScriptEngine::ExcludeSlots);
		}


	void
		MainWindow::recyclePlottingColor(QColor color)
		{
			if(!color.isValid())
				return;

			for(int iter = 0; iter < m_usedColorList.size(); ++iter)
			{
				// qDebug() << __FILE__ << __LINE__
				// << "MainWindow::recycleColor with size: " << m_usedColorList.size()
				// << "and iter: " << iter;

				QColor *iterColor = m_usedColorList.at(iter);

				if(*iterColor == color)
				{
					m_colorList.append(iterColor);
					m_usedColorList.takeAt(iter);

					// We must return immediately because we changed the used color list
					// and we cannot go on iteratring in it!!!
					return;
				}
			}
		}


	bool	
		MainWindow::isPlottingColorUsed(const QColor &color) const
		{
			if(!color.isValid())
				return false;

			for(int iter = 0; iter < m_usedColorList.size(); ++iter)
			{
				// qDebug() << __FILE__ << __LINE__
				// << "MainWindow::recycleColor with size: " << m_usedColorList.size()
				// << "and iter: " << iter;

				QColor *iterColor = m_usedColorList.at(iter);

				if(*iterColor == color)
					return true;
			}

			return false;
		}


	void
		MainWindow::clearPlots()
		{
			mp_openSpectraDlg->actionComboBoxActivated("Close all");
		}


	void
		MainWindow::openFullMassSpectrometryFileDlg()
		{
			openMassSpectrometryFile(QString(), false /* isStreamed */);
		}

	void
		MainWindow::openStreamedMassSpectrometryFileDlg()
		{
			openMassSpectrometryFile(QString(), true /* isStreamed */);
		}


	void
		MainWindow::createMassSpectrumClipboard()
		{
			QClipboard *clipboard = QGuiApplication::clipboard();
			QString msData = clipboard->text();

			logConsoleMessage("Starting the parsing of the mass data text from clipboard.\n");

			MassSpecDataFileLoaderXy loader;


			// This data set will be reparented when we provide it to the handling
			// widget (window or plot widget or whatever QObject).
			MassSpecDataSet *massSpecDataSet = new MassSpecDataSet(this, "NOT_SET"); 

			massSpecDataSet->m_isStreamed = false;

			if(loader.loadDataFromString(msData, massSpecDataSet) < 1)
			{
				statusBarMessage("Failed to load any mass spectrum.");
				delete massSpecDataSet;
				return;
			}

			// We can set some more data to the massSpecDataSet:
			massSpecDataSet->m_fileFormat =
				MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_XY;

			// When loading a new file, we want to first generate a TIC chromatogram
			// for the user to start the investigation on its data.
			//
			// This is actually the entry point to the handling of the data because
			// ownership of the data will be transferred (using the QObject
			// parentship mechanis) to the widget that will display the TIC.

			// Each new plot must have a different color than all the others already
			// used, such that we can distinguish the files throughout all the
			// program. We have two lists of colors, the list of available colors
			// (not used currently) and the list of used colors. When a new plot is
			// created, one color is selected from the available colors list and
			// that color is moved to the used colors list so that it cannot be
			// reused. When a graph is closed, its color is migrated from the list
			// of used colors to the list of available colors.
			QColor color;
			QColor *useColor = nullptr;

			// Let's see if there is still a color available:

			if(m_colorList.isEmpty())
			{
				// Get a new color, since we do not have any more available.

				color = QColorDialog::getColor(Qt::white, this,
						"Please select a color for the plot");

				if(!color.isValid())
				{
					useColor = new QColor(Qt::black);
				}
				else
				{
					useColor = new QColor(color);
					m_usedColorList.append(useColor);
				}
			}
			else
			{
				useColor = m_colorList.first();
				m_usedColorList.append(useColor);
				m_colorList.takeFirst();
			}

			// Before computing the TIC and plotting it, let the window know what
			// color we wish for the plot.
			mp_ticChromWnd->m_wishedColor = *useColor;

			qDebug() << __FILE__ << __LINE__ << "Calculating the TIC chromatogram...";

			statusBarMessage("Calculating the TIC chromatogram...");
			logConsoleMessage("Calculating the TIC chromatogram...\n");

			// Now compute the TIC and plot it. The function returns the plot widget
			// that displays the chromatogram for this mass spec data set.

			AbstractPlotWidget *widget = mp_ticChromWnd->initialTic(massSpecDataSet);

			if(widget == nullptr)
			{
				QMessageBox msgBox;
				msgBox.setText(
						"The program failed to compute an initial TIC chromatogram. Please,"
						" ensure that the file being loaded is correct.");

				msgBox.setStandardButtons(QMessageBox::Ok);
				msgBox.exec();

				delete massSpecDataSet;

				return;
			}

			// At this point, the mainwindow should show a message telling that the
			// file was successfully loaded, since loading a file is actually two
			// things in sequence:
			// 1. Opening the file and loading the data
			// 2. Calculating a TIC chromatogram on the basis of the loaded data.

			statusBarMessage("Done loading file.");
			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			logConsoleMessage("Done loading file.\n");

			// Add a new item to the open spectrum dialog window widget list such
			// that we'll know which files are open. Note that the function below
			// will create a mapping between the list widget item and the tic chrom plot
			// widget that is set to be the parent of massSpecDataSet.
			QListWidgetItem *item =
				mp_openSpectraDlg->addOpenSpectrum(massSpecDataSet, *useColor);

			m_dataPlotWidgetRelationer.newRelation(
					massSpecDataSet, widget /*origin plot widget */,
					nullptr /*target plot widget */, item);
		}

	void 
		MainWindow::openMassSpectrometryFile(const QString &fileName, bool isStreamed)
		{
			if(fileName.isEmpty())
			{
				QFileDialog fileDialog(this);
				fileDialog.setFileMode(QFileDialog::ExistingFiles);
				fileDialog.setViewMode(QFileDialog::Detail);
				fileDialog.setAcceptMode(QFileDialog::AcceptOpen);

				fileDialog.setNameFilters(m_nameFilters);
				fileDialog.setDirectory(QDir::home());

				QStringList fileNames;

				if(fileDialog.exec())
					fileNames = fileDialog.selectedFiles();

				for(int iter = 0; iter < fileNames.size(); ++iter)
				{
					QString curFileName = fileNames.at(iter);

					openMassSpectrometryFile(curFileName, isStreamed);
				}
			}
			else
			{
				qDebug() << __FILE__ << __LINE__ << "Open file: " << fileName;

				MassSpecDataFileFormatAnalyzer analyzer;

				int format = analyzer.canLoadData(fileName);

				// The strategy here is to go from more detailed formats to more generic
				// ones, that'll speed the process.

				if(format == MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_SQLITE3)
				{
					openMassSpectrometryFileSqlite3(fileName, isStreamed);
				}
				else if(format == MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_MZML)
				{
					// openMassSpectrometryFileMzml(fileName, isStreamed);
					openMassSpectrometryFilePwiz(fileName, isStreamed);
				}
				else if(format == MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_DX)
				{
					openMassSpectrometryFileDx(fileName, isStreamed);
				}
				else if(format == MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_MS1)
				{
					openMassSpectrometryFileMs1(fileName, isStreamed);
				}
				else if(format == MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_XY)
				{
					openMassSpectrometryFileXy(fileName);
				}
				else
				{
					QMessageBox msgBox;
					msgBox.setText("Unknown file format");

					QString msg =
						QString("The requested file (%1) has not a valid format.")
						.arg(fileName);

					msgBox.setInformativeText(msg);
					msgBox.setStandardButtons(QMessageBox::Ok);
					msgBox.setDefaultButton(QMessageBox::Ok);

					msgBox.exec();
				}
			}
		}


	void
		MainWindow::openMassSpectrometryFiles(const QStringList &fileList,
				bool isStreamed)
		{
			for(int iter = 0; iter < fileList.size(); iter++)
			{
				QString fileName = fileList.at(iter);

				openMassSpectrometryFile(fileName, isStreamed);
			}
			// End of
			// for(int iter = 0; iter < fileList.size(); iter++)
		}


	void
		MainWindow::openMassSpectrometryFileSqlite3(QString fileName, bool isStreamed)
		{
			// Instantiate an object that will help interrogate the file:

			MassSpecDataFileLoaderSqlite3 loader(fileName);

			if(!loader.m_sqlite3Handler.initializeDbForRead(fileName))
			{
				qDebug() << __FILE__ << __LINE__
					<< "Failed to initialize the database for read. Returning.";

				return;
			}

			makeLoaderConnections(&loader);

			// Allocate a new mass spec data set and provide it to the loadData
			// function.

			// This data set will be reparented when we provide it to the handling
			// widget (window or plot widget or whatever QObject).

			MassSpecDataSet *massSpecDataSet =
				new MassSpecDataSet(this, "NOT_SET", fileName);

			massSpecDataSet->m_isStreamed = isStreamed;

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);

			statusBarMessage("Loading data for file" + fileName + "...");

			double start = omp_get_wtime();

			int readSpectra = loader.loadData(massSpecDataSet);

			if(!massSpecDataSet->m_isStreamed && readSpectra <= 0)
			{
				logConsoleMessage("Failed to load any mass spectrum.");

				return;
			}
			// else
			// We cannot test the returned value, because when the data are
			// streamed, we are not allocating mass spectra.

			statusBarMessage("Loaded data.");

			double end = omp_get_wtime();

			logConsoleMessage(QString("Used %1 s to load %2 spectra\n"
						"retention time hash has %3 items\n")
					.arg(end - start)
					.arg(massSpecDataSet->m_massSpectra.size())
					.arg(massSpecDataSet->m_rtHash.size()));

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			// We may have gotten here because the operation was cancelled. In any case,
			// reset the boolean value.
			m_isOperationCancelled = false;

			// We can set some more data to the massSpecDataSet:
			massSpecDataSet->m_fileFormat =
				MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_SQLITE3;

			// We do not do that anymore, we compute the TIC chromatogram real time.
			// Load the chromatogram that is stored in the database as it was stored
			// in the mzML file.
			// loadChromatogram();

			// When loading a new file, we want to first generate a TIC chromatogram
			// for the user to start the investigation on its data.
			//
			// This is actually the entry point to the handling of the data because
			// ownership of the data will be transferred (using the QObject
			// parentship mechanis) to the widget that will display the TIC.

			setupTicChromPlotWidget(massSpecDataSet);

			logConsoleMessage("Done loading file.\n");

			// At this point, the mainwindow should show a message telling that the
			// file was successfully loaded, since loading a file is actually two
			// things in sequence:
			// 1. Opening the file and loading the data
			// 2. Calculating a TIC chromatogram on the basis of the loaded data.

			statusBarMessage("Done loading file.");
			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});
		}


#if 0

	// This was the function that we used when we were making the parsing of the
	// mzML file with our own code. But now we use the libpwiz facilities.

	void
		MainWindow::openMassSpectrometryFileMzml(QString fileName, bool isStreamed)
		{
			// Instantiate an object that will help interrogate the file:

			MassSpecDataFileLoaderMzml loader(fileName);

			//m_mzmlFileLoader.setFileName(fileName);

			makeLoaderConnections(&loader);

			// We do not want to take into account the Function2 spectra that are empty
			// in the mzML file as converted by ProteoWizard from .raw Waters Synapt V2
			// files.

			int spectrumCount = loader.readSpectrumCount();
			QString msg;
			QMessageBox msgBox;

			if(!loader.containsMobilityData())
			{
				msg =
					QString("Info: this file contains %1 spectra without mobility data.")
					.arg(spectrumCount);
			}
			else
			{
				msg = QString("Info: the file contains %1 spectra with mobility data.")
					.arg(spectrumCount);
			}

			msgBox.setText(msg);
			msgBox.setInformativeText("Do you want to continue?");
			msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
			msgBox.setDefaultButton(QMessageBox::No);
			int ret = msgBox.exec();

			if(ret != QMessageBox::Yes)
				return;

			// At this point, the user wants to do the data load.

			// Set up the progress dialog, for this we need the number of real spectra
			// in the file:

			// qDebug() << __FILE__ << __LINE__
			// << "there are " << spectrumCount << "spectra";

			msg = QString("Loading %1 spectra from mzML file.\n").arg(spectrumCount);

			logConsoleMessage(msg);

			QApplication::processEvents();

			setupProgressFeedback(msg, 1 /* first value */,
					spectrumCount /*last value */);

			// This data set will be reparented when we provide it to the handling
			// widget (window or plot widget or whatever QObject).
			MassSpecDataSet *massSpecDataSet =
				new MassSpecDataSet(this, "NOT_SET", fileName);

			massSpecDataSet->m_isStreamed = isStreamed;

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);

			statusBarMessage("Loading data for file" + fileName + "...");

			double start = omp_get_wtime();

			//qDebug() << __FILE__ << __LINE__
			//<< "Starting the parsing of the mzML file.";

			int readSpectra = loader.loadData(massSpecDataSet);

			if(!massSpecDataSet->m_isStreamed && readSpectra <= 0)
			{
				logConsoleMessage("Failed to load any mass spectrum.");

				return;
			}
			// else
			// We cannot test the returned value, because when the data are
			// streamed, we are not allocating mass spectra.

			statusBarMessage("Loaded data.");

			double end = omp_get_wtime();

			logConsoleMessage(QString("Used %1 s to load %2 spectra\n"
						"retention time hash has %3 items\n")
					.arg(end - start)
					.arg(massSpecDataSet->m_massSpectra.size())
					.arg(massSpecDataSet->m_rtHash.size()));

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			// We may have gotten here because the operation was cancelled. In any case,
			// reset the boolean value.
			m_isOperationCancelled = false;

			// We can set some more data to the massSpecDataSet:
			massSpecDataSet->m_fileFormat =
				MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_MZML;

			// When loading a new file, we want to first generate a TIC chromatogram
			// for the user to start the investigation on its data.
			//
			// This is actually the entry point to the handling of the data because
			// ownership of the data will be transferred (using the QObject
			// parentship mechanis) to the widget that will display the TIC.

			setupTicChromPlotWidget(massSpecDataSet);

			logConsoleMessage("Done loading file.\n");

			// At this point, the mainwindow should show a message telling that the
			// file was successfully loaded, since loading a file is actually two
			// things in sequence:
			// 1. Opening the file and loading the data
			// 2. Calculating a TIC chromatogram on the basis of the loaded data.

			statusBarMessage("Done loading file.");
			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});
		}
#endif

	void
		MainWindow::openMassSpectrometryFilePwiz(QString fileName, bool isStreamed)
		{
			MassSpecDataFileLoaderPwiz loader(fileName);

			makeLoaderConnections(&loader);

			// This data set will be reparented when we provide it to the handling
			// widget (window or plot widget or whatever QObject).
			MassSpecDataSet *massSpecDataSet =
				new MassSpecDataSet(this, "NOT_SET", fileName);

			massSpecDataSet->m_isStreamed = isStreamed;

			// Make sure the cancel button is visible.
			mp_statusBarCancelPushButton->setVisible(true);

			statusBarMessage("Loading data for file" + fileName + "...");

			double start = omp_get_wtime();

			loader.loadData(massSpecDataSet);

			if(massSpecDataSet->m_massSpectra.size() == 0)
			{
				QMessageBox msgBox;
				msgBox.setText(
						"The program failed to load any mass data. Please,"
						" ensure that the file being loaded is correct.");

				msgBox.setStandardButtons(QMessageBox::Ok);
				msgBox.exec();

				delete massSpecDataSet;

				return;
			}

			statusBarMessage("Loaded data.");

			double end = omp_get_wtime();

			logConsoleMessage(QString("Used %1 s to load %2 spectra\n"
						"retention time hash has %3 items\n")
					.arg(end - start)
					.arg(massSpecDataSet->m_massSpectra.size())
					.arg(massSpecDataSet->m_rtHash.size()));

			// Hide the cancel button.
			mp_statusBarCancelPushButton->setVisible(false);

			// We may have gotten here because the operation was cancelled. In any case,
			// reset the boolean value.
			m_isOperationCancelled = false;

			// We can set some more data to the massSpecDataSet:
			massSpecDataSet->m_fileFormat =
				MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_MZML;

			// When loading a new file, we want to first generate a TIC chromatogram
			// for the user to start the investigation on its data.
			//
			// This is actually the entry point to the handling of the data because
			// ownership of the data will be transferred (using the QObject
			// parentship mechanis) to the widget that will display the TIC.

			setupTicChromPlotWidget(massSpecDataSet);

			logConsoleMessage("Done loading file.\n");

			// At this point, the mainwindow should show a message telling that the
			// file was successfully loaded, since loading a file is actually two
			// things in sequence:
			// 1. Opening the file and loading the data
			// 2. Calculating a TIC chromatogram on the basis of the loaded data.

			statusBarMessage("Done loading file.");
			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});
		}


	void
		MainWindow::openMassSpectrometryFileDx(QString fileName, bool isStreamed)
		{
			logConsoleMessage("Starting the parsing of the mass data dx file.\n");

			MassSpecDataFileLoaderDx loader(fileName);

			makeLoaderConnections(&loader);

			int spectrumCount = loader.readSpectrumCount();
			QString msg;
			QMessageBox msgBox;

			msg = QString("Info: this file contains %1 spectra.").arg(spectrumCount);

			msgBox.setText(msg);
			msgBox.setInformativeText("Do you want to continue?");
			msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
			msgBox.setDefaultButton(QMessageBox::No);
			int ret = msgBox.exec();

			if(ret != QMessageBox::Yes)
				return;

			// At this point, the user wants to do the data load.

			// Set up the progress dialog, for this we need the number of real spectra
			// in the file:

			// qDebug() << __FILE__ << __LINE__
			// << "there are " << spectrumCount << "spectra";

			msg = QString("Loading %1 spectra from dx file.\n").arg(spectrumCount);

			logConsoleMessage(msg);

			QApplication::processEvents();

			double start = omp_get_wtime();

			setupProgressFeedback(msg, 1 /* first value */,
					spectrumCount /*last value */);


			// This data set will be reparented when we provide it to the handling
			// widget (window or plot widget or whatever QObject).
			MassSpecDataSet *massSpecDataSet =
				new MassSpecDataSet(this, "NOT_SET", fileName);

			massSpecDataSet->m_isStreamed = isStreamed;

			loader.loadData(massSpecDataSet);

			if(massSpecDataSet->m_massSpectra.size() == 0)
			{
				QMessageBox msgBox;
				msgBox.setText(
						"The program failed to load any mass data. Please,"
						" ensure that the file being loaded is correct.");

				msgBox.setStandardButtons(QMessageBox::Ok);
				msgBox.exec();

				delete massSpecDataSet;

				return;
			}

			statusBarMessage("Loaded data.");

			double end = omp_get_wtime();

			logConsoleMessage(QString("Used %1 s to load %2 spectra\n"
						"retention time hash has %3 items\n")
					.arg(end - start)
					.arg(massSpecDataSet->m_massSpectra.size())
					.arg(massSpecDataSet->m_rtHash.size()));

			// When loading a new file, we want to first generate a TIC chromatogram
			// for the user to start the investigation on its data.
			//
			// This is actually the entry point to the handling of the data because
			// ownership of the data will be transferred (using the QObject
			// parentship mechanis) to the widget that will display the TIC.

			setupTicChromPlotWidget(massSpecDataSet);

			logConsoleMessage("Done loading file.\n");

			// At this point, the mainwindow should show a message telling that the
			// file was successfully loaded, since loading a file is actually two
			// things in sequence:
			// 1. Opening the file and loading the data
			// 2. Calculating a TIC chromatogram on the basis of the loaded data.

			statusBarMessage("Done loading file.");
			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});
		}


	void
		MainWindow::openMassSpectrometryFileMs1(QString fileName, bool isStreamed)
		{
			logConsoleMessage("Starting the parsing of the mass data dx file.\n");

			MassSpecDataFileLoaderMs1 loader(fileName);

			makeLoaderConnections(&loader);

			int spectrumCount = loader.readSpectrumCount();
			QString msg;
			QMessageBox msgBox;

			msg = QString("Info: this file contains %1 spectra.").arg(spectrumCount);

			msgBox.setText(msg);
			msgBox.setInformativeText("Do you want to continue?");
			msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
			msgBox.setDefaultButton(QMessageBox::No);
			int ret = msgBox.exec();

			if(ret != QMessageBox::Yes)
				return;

			// At this point, the user wants to do the data load.

			// Set up the progress dialog, for this we need the number of real spectra
			// in the file:

			// qDebug() << __FILE__ << __LINE__
			// << "there are " << spectrumCount << "spectra";

			msg = QString("Loading %1 spectra from dx file.\n").arg(spectrumCount);

			logConsoleMessage(msg);

			QApplication::processEvents();

			setupProgressFeedback(msg, 1 /* first value */,
					spectrumCount /*last value */);

			double start = omp_get_wtime();

			// This data set will be reparented when we provide it to the handling
			// widget (window or plot widget or whatever QObject).
			MassSpecDataSet *massSpecDataSet =
				new MassSpecDataSet(this, "NOT_SET", fileName);

			massSpecDataSet->m_isStreamed = false;

			loader.loadData(massSpecDataSet);

			if(massSpecDataSet->m_massSpectra.size() == 0)
			{
				QMessageBox msgBox;
				msgBox.setText(
						"The program failed to load any mass data. Please,"
						" ensure that the file being loaded is correct.");

				msgBox.setStandardButtons(QMessageBox::Ok);
				msgBox.exec();

				delete massSpecDataSet;

				return;
			}

			statusBarMessage("Loaded data.");

			double end = omp_get_wtime();

			logConsoleMessage(QString("Used %1 s to load %2 spectra\n"
						"retention time hash has %3 items\n")
					.arg(end - start)
					.arg(massSpecDataSet->m_massSpectra.size())
					.arg(massSpecDataSet->m_rtHash.size()));

			// We can set some more data to the massSpecDataSet:
			massSpecDataSet->m_fileFormat =
				MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_MS1;

			// We may have gotten here because the operation was cancelled. In any case,
			// reset the boolean value.
			m_isOperationCancelled = false;

			// When loading a new file, we want to first generate a TIC chromatogram
			// for the user to start the investigation on its data.
			//
			// This is actually the entry point to the handling of the data because
			// ownership of the data will be transferred (using the QObject
			// parentship mechanis) to the widget that will display the TIC.

			setupTicChromPlotWidget(massSpecDataSet);

			logConsoleMessage("Done loading file.\n");

			// At this point, the mainwindow should show a message telling that the
			// file was successfully loaded, since loading a file is actually two
			// things in sequence:
			// 1. Opening the file and loading the data
			// 2. Calculating a TIC chromatogram on the basis of the loaded data.

			statusBarMessage("Done loading file.");
			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});
		}


	void
		MainWindow::openMassSpectrometryFileBrukerXy(QString fileName, bool isStreamed)
		{
			logConsoleMessage("Starting the parsing of the mass data bruker xy file.\n");

			MassSpecDataFileLoaderBrukerXy loader(fileName);

			makeLoaderConnections(&loader);

			int spectrumCount = loader.readSpectrumCount();
			QString msg;
			QMessageBox msgBox;

			msg = QString("Info: this file contains %1 spectra.").arg(spectrumCount);

			msgBox.setText(msg);
			msgBox.setInformativeText("Do you want to continue?");
			msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
			msgBox.setDefaultButton(QMessageBox::No);
			int ret = msgBox.exec();

			if(ret != QMessageBox::Yes)
				return;

			// At this point, the user wants to do the data load.

			// Set up the progress dialog, for this we need the number of real spectra
			// in the file:

			// qDebug() << __FILE__ << __LINE__
			// << "there are " << spectrumCount << "spectra";

			msg = QString("Loading %1 spectra from dx file.\n").arg(spectrumCount);

			logConsoleMessage(msg);

			QApplication::processEvents();

			setupProgressFeedback(msg, 1 /* first value */,
					spectrumCount /*last value */);

			double start = omp_get_wtime();

			// This data set will be reparented when we provide it to the handling
			// widget (window or plot widget or whatever QObject).
			MassSpecDataSet *massSpecDataSet =
				new MassSpecDataSet(this, "NOT_SET", fileName);

			massSpecDataSet->m_isStreamed = false;

			loader.loadData(massSpecDataSet);

			if(massSpecDataSet->m_massSpectra.size() == 0)
			{
				QMessageBox msgBox;
				msgBox.setText(
						"The program failed to load any mass data. Please,"
						" ensure that the file being loaded is correct.");

				msgBox.setStandardButtons(QMessageBox::Ok);
				msgBox.exec();

				delete massSpecDataSet;

				return;
			}

			statusBarMessage("Loaded data.");

			double end = omp_get_wtime();

			logConsoleMessage(QString("Used %1 s to load %2 spectra\n"
						"retention time hash has %3 items\n")
					.arg(end - start)
					.arg(massSpecDataSet->m_massSpectra.size())
					.arg(massSpecDataSet->m_rtHash.size()));

			// We can set some more data to the massSpecDataSet:
			massSpecDataSet->m_fileFormat =
				MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_BRUKER_XY;

			// We may have gotten here because the operation was cancelled. In any case,
			// reset the boolean value.
			m_isOperationCancelled = false;

			// When loading a new file, we want to first generate a TIC chromatogram
			// for the user to start the investigation on its data.
			//
			// This is actually the entry point to the handling of the data because
			// ownership of the data will be transferred (using the QObject
			// parentship mechanis) to the widget that will display the TIC.

			setupTicChromPlotWidget(massSpecDataSet);

			logConsoleMessage("Done loading file.\n");

			// At this point, the mainwindow should show a message telling that the
			// file was successfully loaded, since loading a file is actually two
			// things in sequence:
			// 1. Opening the file and loading the data
			// 2. Calculating a TIC chromatogram on the basis of the loaded data.

			statusBarMessage("Done loading file.");
			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});
		}	


	void
		MainWindow::openMassSpectrometryFileXy(QString fileName)
		{
			// The file format is an easy one : two double values separated by either
			// ',' or '[space]' or '[tab]' or ';'.

			logConsoleMessage("Starting the parsing of the mass data text file.\n");

			MassSpecDataFileLoaderXy loader(fileName);

			// This data set will be reparented when we provide it to the handling
			// widget (window or plot widget or whatever QObject).
			MassSpecDataSet *massSpecDataSet =
				new MassSpecDataSet(this, "NOT_SET", fileName);

			massSpecDataSet->m_isStreamed = false;

			double start = omp_get_wtime();

			loader.loadData(massSpecDataSet);

			if(massSpecDataSet->m_massSpectra.size() == 0)
			{
				QMessageBox msgBox;
				msgBox.setText(
						"The program failed to load any mass data. Please,"
						" ensure that the file being loaded is correct.");

				msgBox.setStandardButtons(QMessageBox::Ok);
				msgBox.exec();

				delete massSpecDataSet;

				return;
			}

			statusBarMessage("Loaded data.");

			double end = omp_get_wtime();

			logConsoleMessage(QString("Used %1 s to load %2 spectra\n"
						"retention time hash has %3 items\n")
					.arg(end - start)
					.arg(massSpecDataSet->m_massSpectra.size())
					.arg(massSpecDataSet->m_rtHash.size()));

			// We can set some more data to the massSpecDataSet:
			massSpecDataSet->m_fileFormat =
				MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_XY;

			// When loading a new file, we want to first generate a TIC chromatogram
			// for the user to start the investigation on its data.
			//
			// This is actually the entry point to the handling of the data because
			// ownership of the data will be transferred (using the QObject
			// parentship mechanis) to the widget that will display the TIC.

			// Each new plot must have a different color than all the others already
			// used, such that we can distinguish the files throughout all the
			// program. We have two lists of colors, the list of available colors
			// (not used currently) and the list of used colors. When a new plot is
			// created, one color is selected from the available colors list and
			// that color is moved to the used colors list so that it cannot be
			// reused. When a graph is closed, its color is migrated from the list
			// of used colors to the list of available colors.
			QColor color;
			QColor *useColor = nullptr;

			// Let's see if there is still a color available:

			if(m_colorList.isEmpty())
			{
				// Get a new color, since we do not have any more available.

				color = QColorDialog::getColor(Qt::white, this,
						"Please select a color for the plot");

				if(!color.isValid())
				{
					useColor = new QColor(Qt::black);
				}
				else
				{
					useColor = new QColor(color);
					m_usedColorList.append(useColor);
				}
			}
			else
			{
				useColor = m_colorList.first();
				m_usedColorList.append(useColor);
				m_colorList.takeFirst();
			}

			// Before computing the TIC and plotting it, let the window know what
			// color we wish for the plot.
			mp_ticChromWnd->m_wishedColor = *useColor;

			qDebug() << __FILE__ << __LINE__ << "Calculating the TIC chromatogram...";

			statusBarMessage("Calculating the TIC chromatogram...");
			logConsoleMessage("Calculating the TIC chromatogram...\n");

			// Now compute the TIC and plot it. The function returns the plot widget
			// that displays the chromatogram for this mass spec data set.

			AbstractPlotWidget *widget = mp_ticChromWnd->initialTic(massSpecDataSet);

			if(widget == nullptr)
			{
				QMessageBox msgBox;
				msgBox.setText(
						"The program failed to compute an initial TIC chromatogram. Please,"
						" ensure that the file being loaded is correct.");

				msgBox.setStandardButtons(QMessageBox::Ok);
				msgBox.exec();

				delete massSpecDataSet;

				return;
			}


			// At this point, the mainwindow should show a message telling that the
			// file was successfully loaded, since loading a file is actually two
			// things in sequence:
			// 1. Opening the file and loading the data
			// 2. Calculating a TIC chromatogram on the basis of the loaded data.

			statusBarMessage("Done loading file.");
			QTimer::singleShot(3000, this, [=]() {
					this->updateProgressFeedback("", 0);
					this->mp_statusBarProgressBar->setVisible(false);
					});

			logConsoleMessage("Done loading file.\n");

			// Add a new item to the open spectrum dialog window widget list such
			// that we'll know which files are open. Note that the function below
			// will create a mapping between the list widget item and the tic chrom plot
			// widget that is set to be the parent of massSpecDataSet.
			QListWidgetItem *item =
				mp_openSpectraDlg->addOpenSpectrum(massSpecDataSet, *useColor);

			m_dataPlotWidgetRelationer.newRelation(
					massSpecDataSet, widget /*origin plot widget */,
					nullptr /*target plot widget */, item);
		}


	void
		MainWindow::makeLoaderConnections(MassSpecDataFileLoader *loader)
		{
			if(loader == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.",
						__FILE__, __LINE__);

			// Connect the loader to this window so that we can display the reader log
			// messages:

			connect(loader, &MassSpecDataFileLoader::setupProgressFeedbackSignal,
					this, &MainWindow::setupProgressFeedback);

			connect(loader, &MassSpecDataFileLoader::dataLoadMessageSignal,
					this, &MainWindow::dataLoadMessage);

			// Also, make sure we get the signal that informs us on the spectrum
			// reading progression:

			connect(loader, &MassSpecDataFileLoader::dataLoadProgressSignal,
					this, &MainWindow::statusBarProgress);

			connect(this, &MainWindow::cancelOperationSignal, loader,
					&MassSpecDataFileLoader::cancelOperation);
		}


	void
		MainWindow::analysisPrefDlg()
		{
			if(!mp_analysisPrefDlg)
			{
				mp_analysisPrefDlg = new AnalysisPreferencesDlg(
						dynamic_cast<QWidget *>(this), "mineXpert");
			}

			int res = mp_analysisPrefDlg->exec();

			if(res == QDialog::Accepted)
			{
				// Copy the preferences from the dialog window to *this.
				mpa_analysisPreferences =
					new AnalysisPreferences(mp_analysisPrefDlg->analysisPreferences());

				// Let's check if the recording should go a file, in which case we test
				// that we can open it properly.

				if((mpa_analysisPreferences->m_recordTarget &
							RecordTarget::RECORD_TO_FILE) == RecordTarget::RECORD_TO_FILE)
				{
					// Now check that the analysis file could be opened.
					m_analysisFile.setFileName(mpa_analysisPreferences->m_fileName);

					bool res = false;

					if(mpa_analysisPreferences->m_fileOpenMode == QIODevice::Truncate)
						// Interestingly, only truncate did not work. Since writeonly
						// implies
						// truncate, that's fine, and it works.
						res = m_analysisFile.open(QIODevice::WriteOnly);
					else
						res = m_analysisFile.open(QIODevice::Append);

					if(!res)
					{
						QMessageBox msgBox;
						msgBox.setText("Failed to open the analysis file, please fix "
								"the analysis preferences.");
						msgBox.exec();

						return;
					}
					else
					{
						// Close the file, which will be opened each time it is
						// necessary.
						m_analysisFile.close();

						// We can effectively start the work. For that we need to inform
						// the
						// mass spectrum window that the analysis file where to store
						// the
						// analysis data is available (the pointer won't be 0 anymore).
						mp_massSpecWnd->setAnalysisFile(&m_analysisFile);

						// Same for the drift time window:
						mp_driftSpecWnd->setAnalysisFile(&m_analysisFile);
					}
				}

				// Finally do not forget to set the gotten analyses to the mass spec
				// window.
				mp_massSpecWnd->setAnalysisPreferences(mpa_analysisPreferences);
				mp_driftSpecWnd->setAnalysisPreferences(mpa_analysisPreferences);
				mp_ticChromWnd->setAnalysisPreferences(mpa_analysisPreferences);
			}
			else // if(res == QDialog::Accepted)
			{
				// The dialog returned a rejected result. We cannot do anything.
			}
		}

	void
		MainWindow::createMenusAndActions()
		{
			// File menu
			m_fileMenu = menuBar()->addMenu("&File");

			m_openFullMsFileAct = new QAction(tr("Open &full mass spectrum file(s)"),
					dynamic_cast<QObject *>(this));
			m_openFullMsFileAct->setShortcut(QKeySequence("Ctrl+O, F"));
			m_openFullMsFileAct->setStatusTip(
					tr("Open file(s) and store data in memory"));
			connect(m_openFullMsFileAct, &QAction::triggered, this,
					&MainWindow::openFullMassSpectrometryFileDlg);
			m_fileMenu->addAction(m_openFullMsFileAct);

			m_openStreamedMsFileAct =
				new QAction(tr("Open &streamed mass spectrum file(s)"),
						dynamic_cast<QObject *>(this));
			m_openStreamedMsFileAct->setShortcut(QKeySequence("Ctrl+O, S"));
			m_openStreamedMsFileAct->setStatusTip(
					tr("Open files(s) leaving data on disk"));
			connect(m_openStreamedMsFileAct, &QAction::triggered, this,
					&MainWindow::openStreamedMassSpectrometryFileDlg);
			m_fileMenu->addAction(m_openStreamedMsFileAct);

			m_openMsClipboardAct = new QAction(tr("&Mass spectrum from clipboard"),
					dynamic_cast<QObject *>(this));
			m_openMsClipboardAct->setShortcut(QKeySequence("Ctrl+O, C"));
			m_openMsClipboardAct ->setStatusTip(
					tr("Create mass spectrum from clipboard data"));
			connect(m_openMsClipboardAct , &QAction::triggered, this,
					&MainWindow::createMassSpectrumClipboard);
			m_fileMenu->addAction(m_openMsClipboardAct );

			m_fileMenu->addSeparator();

			m_analysisPrefsAct =
				new QAction(tr("&Analysis preferences"), dynamic_cast<QObject *>(this));
			m_analysisPrefsAct->setShortcut(tr("Ctrl+P"));
			m_analysisPrefsAct->setStatusTip(tr("Set the analysis preferences"));
			connect(m_analysisPrefsAct, &QAction::triggered, this,
					&MainWindow::analysisPrefDlg);
			m_fileMenu->addAction(m_analysisPrefsAct);

			m_fileMenu->addSeparator();

			m_quitAct = new QAction(tr("&Quit"), dynamic_cast<QObject *>(this));
			m_quitAct->setShortcut(tr("Ctrl+Q"));
			m_quitAct->setStatusTip(tr("Exit mineXpert"));
			connect(m_quitAct, &QAction::triggered, this, &MainWindow::close);
			m_fileMenu->addAction(m_quitAct);


			// Plot menu
			m_plotMenu = menuBar()->addMenu("&Plot");

			m_clearPlotsAct =
				new QAction("&Clear plots", dynamic_cast<QObject *>(this));
			connect(m_clearPlotsAct, &QAction::triggered, this,
					&MainWindow::clearPlots);

			m_plotMenu->addAction(m_clearPlotsAct);


			// Windows menu
			m_windowsMenu = menuBar()->addMenu("&Windows");

			m_showTicChromWndAct =
				new QAction("Show &TIC window", dynamic_cast<QObject *>(this));
			m_showTicChromWndAct->setStatusTip(tr("Show the TIC chromatogram window"));
			connect(m_showTicChromWndAct, &QAction::triggered, this,
					&MainWindow::showTicChromWnd);
			m_windowsMenu->addAction(m_showTicChromWndAct);

			m_showMassSpecWndAct =
				new QAction("Show &Mass spectrum window", dynamic_cast<QObject *>(this));
			m_showMassSpecWndAct->setStatusTip(tr("Show the mass spectrum window"));
			connect(m_showMassSpecWndAct, &QAction::triggered, this,
					&MainWindow::showMassSpecWnd);
			m_windowsMenu->addAction(m_showMassSpecWndAct);

			m_showDriftSpecWndAct =
				new QAction("Show &Drift spectrum window", dynamic_cast<QObject *>(this));
			m_showDriftSpecWndAct->setStatusTip(tr("Show the drift spectrum window"));
			connect(m_showDriftSpecWndAct, &QAction::triggered, this,
					&MainWindow::showDriftSpecWnd);
			m_windowsMenu->addAction(m_showDriftSpecWndAct);

			m_showColorMapWndAct =
				new QAction("Show &Color map window", dynamic_cast<QObject *>(this));
			m_showColorMapWndAct->setStatusTip(tr("Show the color map window"));
			connect(m_showColorMapWndAct, &QAction::triggered, this,
					&MainWindow::showColorMapWnd);
			m_windowsMenu->addAction(m_showColorMapWndAct);

			m_windowsMenu->addSeparator();

			m_showOpenSpectraDlgAct = new QAction("Show &Open mass spectra dialog",
					dynamic_cast<QObject *>(this));
			m_showOpenSpectraDlgAct->setStatusTip(
					tr("Show the open mass spectra dialog"));
			connect(m_showOpenSpectraDlgAct, &QAction::triggered, this,
					&MainWindow::showOpenMassSpectraDlg);
			m_windowsMenu->addAction(m_showOpenSpectraDlgAct);

			m_showConsoleWndAct =
				new QAction("Show &Console window", dynamic_cast<QObject *>(this));
			m_showConsoleWndAct->setStatusTip(tr("Show the console window"));
			connect(m_showConsoleWndAct, &QAction::triggered, this,
					&MainWindow::showConsoleWnd);
			m_windowsMenu->addAction(m_showConsoleWndAct);

			m_showScriptingWndAct =
				new QAction("Show &Scripting window", dynamic_cast<QObject *>(this));
			m_showScriptingWndAct->setStatusTip(tr("Show the scripting window"));
			connect(m_showScriptingWndAct, &QAction::triggered, this,
					&MainWindow::showScriptingWnd);
			m_windowsMenu->addAction(m_showScriptingWndAct);

			m_windowsMenu->addSeparator();

			m_showAllWindowsAct =
				new QAction("Show all windows", dynamic_cast<QObject *>(this));
			m_showAllWindowsAct->setStatusTip(tr("Show all windows"));
			connect(m_showAllWindowsAct, &QAction::triggered, this,
					&MainWindow::showAllWindows);
			m_windowsMenu->addAction(m_showAllWindowsAct);

			m_saveWorkspaceAct =
				new QAction("Save workspace", dynamic_cast<QObject *>(this));
			m_saveWorkspaceAct->setStatusTip(tr("Save workspace"));
			connect(m_saveWorkspaceAct, &QAction::triggered, this,
					&MainWindow::saveWorkspace);
			m_windowsMenu->addAction(m_saveWorkspaceAct);


			// Help menu
			m_helpMenu = menuBar()->addMenu("&Help");

			m_aboutAct = new QAction(tr("&About"), dynamic_cast<QObject *>(this));
			m_aboutAct->setShortcut(tr("Ctrl+H"));
			m_aboutAct->setStatusTip(tr("Show the application's About box"));
			connect(m_aboutAct, &QAction::triggered, this, &MainWindow::about);
			m_helpMenu->addAction(m_aboutAct);

			m_aboutQtAct = new QAction(tr("About &Qt"), dynamic_cast<QObject *>(this));
			m_aboutQtAct->setStatusTip(tr("Show the Qt library's About box"));
			connect(m_aboutQtAct, &QAction::triggered, this, &Application::aboutQt);
			m_helpMenu->addAction(m_aboutQtAct);
		}

	void
		MainWindow::about()
		{
			// The application name will be set automatically by default parameter
			// value.
			AboutMineXpertDlg *dlg = new AboutMineXpertDlg(this);

			dlg->show();
		}

	void
		MainWindow::writeSettings()
		{
			QSettings settings;
			settings.setValue("MainWindow_geometry", saveGeometry());
			settings.setValue("MainWindow_windowState", saveState());
		}

	void
		MainWindow::readSettings()
		{
			QSettings settings;
			restoreGeometry(settings.value("MainWindow_geometry").toByteArray());
			restoreState(settings.value("MainWindow_windowState").toByteArray());
		}

#if 0 // fixme
	void
		MainWindow::loadSqlData(MassSpecDataSet *massSpecDataSet)
		{
			if(massSpecDataSet == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			double start = omp_get_wtime();

			int spectrumListCount = m_msSqlFileLoader.readSpectrumCount();

			// set the parentship to the tic chrom window, so that the dialog is
			// displayed at the center of that window were the data will be displayed.

			setupProgressFeedback("Loading mass spectra from the database file",
					1 /* first value */,
					spectrumListCount /*last value */);

			// When the loader sends a message about loading a new set of spectra,
			// we'll receive that message and set the new value to the progress
			// dialog.

			// We first try to know if the data in the database are for a
			// DRIFT_SPECTRA experiment or only for a MASS_SPECTRA experiment.
			int dataKind = MassDataKind::MASS_SPECTRA;

			bool isDriftExperiment = m_msSqlFileLoader.isDriftExperiment();

			if(isDriftExperiment)
				dataKind = MassDataKind::DRIFT_SPECTRA;
			else
				dataKind = MassDataKind::MASS_SPECTRA;

			int spectrumCount = m_msSqlFileLoader.loadData(massSpecDataSet, dataKind);

			// Whe the data are loaded as a stream, the result is not any MassSpectrum
			// whatsoever, so testing the number of mass spectra is not useful.
			if(!massSpecDataSet->m_isStreamed && spectrumCount <= 0)
			{
				logConsoleMessage("Failed to load any mass spectrum.");

				return;
			}

			// At this point we should have terminated loading all the data.


			double end = omp_get_wtime();

			logConsoleMessage(QString("Used %1 s to load %2 spectra\n"
						"retention time hash has %3 items\n")
					.arg(end - start)
					.arg(massSpecDataSet->m_massSpectra.size())
					.arg(massSpecDataSet->m_rtHash.size()));
		}
#endif


	void
		MainWindow::logConsoleMessage(QString msg, const QColor &color, bool overwrite)
		{
			mp_consoleWnd->logMessage(msg, color, overwrite);

			// Process the app events, to refresh the GUI.
			QApplication::processEvents();
		}

	void
		MainWindow::dataLoadMessage(QString message, int logType)
		{
			if(logType & LOG_TO_CONSOLE)
				logConsoleMessage(message);
			else if(logType & LOG_TO_CONSOLE_OVERWRITE)
				logConsoleMessage(message, QColor(), true);

			if(logType & LOG_TO_STATUS_BAR)
				statusBarMessage(message);

			// qDebug() << __FILE__ << __LINE__
			// << "dataLoadMessage:" << message;
		}

	void
		MainWindow::statusBarProgress(int value)
		{
			// Created invisible, thus make it visible first.
			mp_statusBarProgressBar->setVisible(true);
			mp_statusBarProgressBar->setValue(value);

			// Process the app events, to refresh the GUI.
			QApplication::processEvents();
		}

	void
		MainWindow::statusBarMessage(QString msg)
		{
			// Created invisible, thus make it visible first.
			mp_statusBarLabel->setVisible(true);
			mp_statusBarLabel->setText(msg);

			// Process the app events, to refresh the GUI.
			QApplication::processEvents();
		}

	void
		MainWindow::showTicChromWnd()
		{
			mp_ticChromWnd->activateWindow();
			mp_ticChromWnd->raise();
			mp_ticChromWnd->show();
		}

	void
		MainWindow::showMassSpecWnd()
		{
			mp_massSpecWnd->activateWindow();
			mp_massSpecWnd->raise();
			mp_massSpecWnd->show();
		}

	void
		MainWindow::showDriftSpecWnd()
		{
			mp_driftSpecWnd->activateWindow();
			mp_driftSpecWnd->raise();
			mp_driftSpecWnd->show();
		}

	void
		MainWindow::showColorMapWnd()
		{
			mp_colorMapWnd->activateWindow();
			mp_colorMapWnd->raise();
			mp_colorMapWnd->show();
		}

	void
		MainWindow::showOpenMassSpectraDlg()
		{
			mp_openSpectraDlg->activateWindow();
			mp_openSpectraDlg->raise();
			mp_openSpectraDlg->show();
		}

	void
		MainWindow::showConsoleWnd()
		{
			mp_consoleWnd->activateWindow();
			mp_consoleWnd->raise();
			mp_consoleWnd->show();
		}

	void
		MainWindow::showScriptingWnd()
		{
			mp_scriptingWnd->activateWindow();
			mp_scriptingWnd->raise();
			mp_scriptingWnd->show();
		}

	void
		MainWindow::showAllWindows()
		{
			mp_ticChromWnd->show();
			mp_massSpecWnd->show();
			mp_driftSpecWnd->show();
			mp_colorMapWnd->show();
			mp_openSpectraDlg->show();
			mp_consoleWnd->show();
			mp_scriptingWnd->show();
		}

	void
		MainWindow::saveWorkspace()
		{
			mp_ticChromWnd->writeSettings();
			mp_massSpecWnd->writeSettings();
			mp_driftSpecWnd->writeSettings();
			mp_colorMapWnd->writeSettings();
			mp_openSpectraDlg->writeSettings();
			mp_consoleWnd->writeSettings();
			mp_scriptingWnd->writeSettings();
		}

	void
		MainWindow::setupProgressFeedback(QString msg, int startValue, int endValue,
				int currentValue)
		{
			mp_statusBarProgressBar->setMinimum(startValue);
			mp_statusBarProgressBar->setMaximum(endValue);

			if(currentValue != -1)
				mp_statusBarProgressBar->setValue(currentValue);

			if(!msg.isEmpty())
				mp_statusBarLabel->setText(msg);

			QApplication::processEvents();
		}

	void
		MainWindow::updateProgressFeedback(QString msg, int currentValue)
		{
			statusBarMessage(msg);
			statusBarProgress(currentValue);
		}

	void
		MainWindow::destroyAlsoAllTargetPlotWidgets(
				AbstractPlotWidget *originPlotWidget)
		{
			if(originPlotWidget == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			QString originPlotWidgetString = QString("0x%1").arg(
					(quintptr)originPlotWidget, QT_POINTER_SIZE * 2, 16, QChar('0'));

			qDebug()
				<< __FILE__ << __LINE__
				<< "Entering MainWindow::destroyAlsoAllTargetPlotWidgets with param:"
				<< originPlotWidgetString << originPlotWidget << originPlotWidget->m_name
				<< originPlotWidget->m_desc << "undergoing destruction."
				<< "There are currently " << m_dataPlotWidgetRelationer.m_list.size()
				<< "relations in total.";

			// Get all the relations that have originPlotWidget as origin plot widget:
			QList<DataPlotWidgetRelation *> matchingOriginList;
			m_dataPlotWidgetRelationer.getRelationsByOrigin(originPlotWidget,
					&matchingOriginList);

			qDebug() << __FILE__ << __LINE__ << "The plot widget being destroyed has "
				<< matchingOriginList.size() << "target plot widgets.";

			AbstractPlotWidget * originRootPlotWidget = nullptr;
			DataPlotWidgetRelation *rootRelation = nullptr;

			qDebug() << __FILE__ << __LINE__
				<< "Going to iterate in the list of target plot widgets";

			// Now, for each relation get the *target* widget and make sure we ask
			// it to destroy itself.
			for(int iter = 0; iter < matchingOriginList.size(); ++iter)
			{
				DataPlotWidgetRelation *relation = matchingOriginList.at(iter);

				//qDebug() << __FILE__ << __LINE__ << "Iteration " << iter
				//<< "relation: " << relation;

				// There is a specific case where a relation may have a target plot
				// widget pointer that is nullptr: when the origin plot widget is a root
				// plot widget, that is the plot widget that displays the TIC
				// chromatogram right after loading mass data from file. We need to test
				// this, because we cannot use that nullptr pointer.
				if(relation->mp_targetPlotWidget == nullptr)
				{
					// The currently iterated plot widget is a root plot widget, wait
					// for
					// the end of the for loop, then we'll deal with it. We cannot call
					// destroy plot widget for its target plot widget that does not
					// exist!

					originRootPlotWidget = relation->mp_originPlotWidget;
					rootRelation = relation;

					qDebug()
						<< __FILE__ << __LINE__
						<< "This iterated relation is for an origin root plot widget:"
						<< originRootPlotWidget << originRootPlotWidget->m_name
						<< originRootPlotWidget->m_desc
						<< "Setting aside the data for later handling, when outside the "
						"for loop.";

					continue;
				}

				qDebug() << __FILE__ << __LINE__ << "Iteration " << iter
					<< "relation: " << relation
					<< "has a non-nullptr target plot widget:"
					<< relation->mp_targetPlotWidget
					<< relation->mp_targetPlotWidget->m_name;

				// Upon running the below function, the destroyed widget will become the
				// originPlotWidget of a new function call to this function. This is
				// recursive execution of the function until all the target widgets have
				// lost their target widgets, and thus become origin plot widgets...

				qDebug() << __FILE__ << __LINE__
					<< "Going to call "
					"relation->mp_targetPlotWidget->shouldDestroyPlotWidget"
					<< "on" << relation->mp_targetPlotWidget;

				relation->mp_targetPlotWidget->shouldDestroyPlotWidget();
			}

			// End of
			// for (int iter = 0; iter < matchingOriginList.size(); ++iter)

			qDebug() << __FILE__ << __LINE__ << "Finished iterating in the list of "
				"relations that match the origin plot " "widget.";

			// There is a specific case here: when we delete the origin root plot
			// widget:
			if(originRootPlotWidget != nullptr)
			{
				qDebug() << __FILE__ << __LINE__
					<< "Now handling the origin root plot widget case."
					<< "Deleting the root plot widget and root relation:"
					<< originRootPlotWidget << originRootPlotWidget->m_name << "/"
					<< rootRelation;

				QColor color = originRootPlotWidget->color();

				// The root origin plot widget must have the open spectra dialog list
				// widget item member datum set. Use it so that we can delete that item.

				delete rootRelation->mp_openSpectraListWidgetItem;

				m_dataPlotWidgetRelationer.clearRelation(rootRelation);
				delete originRootPlotWidget;
				originRootPlotWidget = nullptr;

				// Finally, we can recycle the plotting color.
				recyclePlottingColor(color);
			}

			// At this point, we should be able to delete the original plot widget, the
			// one that started the cascade.

			// We want to clear the relation for this plot widget before deleting
			// it. To get its relation, we need to ask for the relations that have
			// their target matching this plot widget:
			QList<DataPlotWidgetRelation *> matchingTargetList;
			m_dataPlotWidgetRelationer.getRelationsByTarget(originPlotWidget,
					&matchingTargetList);

			if(matchingTargetList.size() == 1)
			{
				// Indeed, this widget is the last leaf of a branch, delete it and
				// also clear its relation:

				m_dataPlotWidgetRelationer.clearRelation(matchingTargetList.first());

				qDebug() << __FILE__ << __LINE__
					<< "Now deleting the last leaf plot widget" << originPlotWidget
					<< originPlotWidget->m_name << originPlotWidget->m_desc;

				delete originPlotWidget;
				originPlotWidget = nullptr;
			}
			else if(matchingTargetList.size() == 0)
			{
				qDebug() << __FILE__ << __LINE__ << "No matching target list, the "
					"origin root plot widget was "
					"deleted already.";
			}
		}

	void
		MainWindow::toggleMultiGraphGraph(AbstractPlotWidget *widget, bool recursive)
		{
			if(widget == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			QList<AbstractPlotWidget *> plotWidgetList;

			// Whatever the recursive value, we want to handle the widget:
			plotWidgetList.append(widget);

			if(recursive)
			{
				// The user has asked that the param widget and all its targets have
				// their
				// multigraph graph shown if it was hidden or hidden if it was shown.
				m_dataPlotWidgetRelationer.targetsRecursive(widget, &plotWidgetList);
			}

			// Now that we have seeded the plot wiget list, just work on it.
			for(int iter = 0; iter < plotWidgetList.size(); ++iter)
				plotWidgetList.at(iter)->toggleMultiGraphGraph();
		}

	void
		MainWindow::cancelPushButtonClicked()
		{
			m_isOperationCancelled = true;
			emit cancelOperationSignal();
		}


	void
		MainWindow::setupTicChromPlotWidget(MassSpecDataSet *massSpecDataSet)
		{

			// Each new plot must have a different color than all the others already
			// used, such that we can distinguish the files throughout all the
			// program. We have two lists of colors, the list of available colors
			// (not used currently) and the list of used colors. When a new plot is
			// created, one color is selected from the available colors list and
			// that color is moved to the used colors list so that it cannot be
			// reused. When a graph is closed, its color is migrated from the list
			// of used colors to the list of available colors.

			QColor color;
			QColor *useColor = nullptr;

			// Let's see if there is still a color available:

			if(m_colorList.isEmpty())
			{
				// Get a new color, since we do not have any more available.

				color = QColorDialog::getColor(Qt::white, this,
						"Please select a color for the plot");

				if(!color.isValid())
				{
					useColor = new QColor(Qt::black);
				}
				else
				{
					useColor = new QColor(color);
					m_usedColorList.append(useColor);
				}
			}
			else
			{
				useColor = m_colorList.first();
				m_usedColorList.append(useColor);
				m_colorList.takeFirst();
			}

			// Before computing the TIC and plotting it, let the window know what
			// color we wish for the plot.
			mp_ticChromWnd->m_wishedColor = *useColor;

			// Now compute the TIC and plot it. The function returns the plot widget
			// that displays the chromatogram for this mass spec data set.

			qDebug() << __FILE__ << __LINE__ << "Calculating the TIC chromatogram...";
			statusBarMessage("Calculating the TIC chromatogram...");
			logConsoleMessage("Calculating the TIC chromatogram...\n");

			AbstractPlotWidget *ticChromPlotWidget =
				mp_ticChromWnd->initialTic(massSpecDataSet);

			if(ticChromPlotWidget == nullptr)
			{
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}

			qDebug() << __FILE__ << __LINE__
				<< "Done calculating the TIC chromatogram...";

			statusBarMessage("Done calculating the TIC chromatogram...");
			logConsoleMessage("Done calculating the TIC chromatogram...\n");

			// Add a new item to the open spectrum dialog window widget list such
			// that we'll know which files are open. Note that the function below
			// will create a mapping between the list widget item and the tic chrom plot
			// widget that is set to be the parent of massSpecDataSet.
			QListWidgetItem *item =
				mp_openSpectraDlg->addOpenSpectrum(massSpecDataSet, *useColor);

			m_dataPlotWidgetRelationer.newRelation(
					massSpecDataSet, ticChromPlotWidget /* origin plot widget */,
					nullptr /* target plot widget */, item);

			// At this time, we now can compute the initial color map, thus generating
			// a new plot widget that we will associate to the tic chrom plot widget.

			// The color map that relates the drift time with the m/z with the
			// intensity of the m/z as a color map.
			// To be able to identify the colormap as related to a given tic
			// chromatogram plot widget, we'll set its background to the same color.
			mp_colorMapWnd->m_wishedColor = *useColor;

			// Right before starting the color map calculation, that can be lengthy,
			// make sure we show it on the fore for the user to see it.

			showColorMapWnd();

			bool ok = false;

			AbstractPlotWidget *colorMapPlotWidget =
				mp_colorMapWnd->initialColorMap(massSpecDataSet, &ok);

			if(colorMapPlotWidget == nullptr)
			{
				if(!ok)
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

				qDebug() << __FILE__ << __LINE__
					<< "Returned nullptr color map widget.";
			}
			else
			{
				// The colormap is considered to be a descendant from the TIC:

				m_dataPlotWidgetRelationer.newRelation(
						massSpecDataSet, ticChromPlotWidget /*origin plot widget */,
						colorMapPlotWidget /*target plot widget */, nullptr);
			}
		}


} // namespace msXpertSuite
