/*   msXpertSuite - mass spectrometry software suite
 *   -----------------------------------------------
 *   Copyright(C) 2016 Filippo Rusconi
 *
 *   http://www.{site}.org
 *
 *   This file is part of the msXpertSuite project.
 *
 *   The msXpertSuite project is the successor of the massXpert project. This
 *   project now includes various independent modules:
 *   - massXpert, the original mass spectrometry environment software (greatly
 *     refactored code);
 *   - mineXpert, a powerful TIC chromatogram/mass/drift spectrum viewer/miner;
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


/////////////////////// Std includes


/////////////////////// Qt includes
#include <QFileInfo>
#include <QList>
#include <QStringList>
#include <QDebug>
#include <qmath.h>


/////////////////////// Local includes
#include "CommandLineRunner.hpp"
#include "MassSpecDataFileLoaderSqlite3.hpp"
#include "MassSpecDataFileLoaderMzml.hpp"
#include "MassSpecDataFileFormatAnalyzer.hpp"
#include "MassSpecDataFileLoaderXy.hpp"
#include "globals.hpp"


namespace msXpertSuite
{

	CommandLineRunner::CommandLineRunner()
	{
		m_rtStart = m_mzStart = m_dtStart = qSNaN();
		m_rtEnd = m_mzEnd = m_dtEnd = qSNaN();
	}


	CommandLineRunner::~CommandLineRunner()
	{
	}


	bool
		CommandLineRunner::allStartEndValuesUnset()
		{
			int setValues = 0;

			setValues += (qIsNaN(m_rtStart) ? 0 : 1);
			setValues += (qIsNaN(m_mzStart) ? 0 : 1);
			setValues += (qIsNaN(m_dtStart) ? 0 : 1);

			setValues += (qIsNaN(m_rtEnd) ? 0 : 1);
			setValues += (qIsNaN(m_mzEnd) ? 0 : 1);
			setValues += (qIsNaN(m_dtEnd) ? 0 : 1);

			if(setValues == 0)
			{
				// qDebug() << __FILE__ << __LINE__ << "returning true.";

				return true;
			}

			// qDebug() << __FILE__ << __LINE__ << "returning false.";

			return false;
		}


	bool
		CommandLineRunner::createHistoryItems(QString *errors)
		{
			if(errors == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// We'll check that size again to see if errors occurred here.
			int errorStringSize = errors->size();

			// All the Start/End range values are initialized in the constructor with
			// Not a Number values to make the distinction between a variable that was
			// set by the user and a varialbe that was untouched.

			// Remove all the previously set history items in the member History.
			m_history.freeList();

			if(!qIsNaN(m_rtStart) && !qIsNaN(m_rtEnd))
			{
				// There was retention time integration asked for.

				// qDebug() << __FILE__ << __LINE__ << "Retention time integration
				// range:"
				//<< "start:" << m_rtStart << "end:" << m_rtEnd;

				if(m_rtEnd < m_rtStart)
				{
					double temp = m_rtEnd;
					m_rtEnd     = m_rtStart;
					m_rtStart   = temp;
				}

				if(m_rtStart < 0)
				{
					*errors += "The rtStart value is less than 0.";
				}

				// At this point, create the history item.
				HistoryItem *item = new HistoryItem();
				item->newIntegrationRange(IntegrationType::ANY_TO_RT, m_rtStart,
						m_rtEnd);
				m_history.appendHistoryItem(item);

				// qDebug() << __FILE__ << __LINE__
				//<< "New history item:" << item->asText(false);
			}

			if(!qIsNaN(m_mzStart) && !qIsNaN(m_mzEnd))
			{
				// There was mass-to-charge ratio integration asked for.

				// qDebug() << __FILE__ << __LINE__ << "Mz integration range:"
				//<< "start:" << m_mzStart << "end:" << m_mzEnd;

				if(m_mzEnd < m_mzStart)
				{
					double temp = m_mzEnd;
					m_mzEnd     = m_mzStart;
					m_mzStart   = temp;
				}

				if(m_mzStart < 0)
				{
					*errors += "The mzStart value is less than 0.";
				}

				// At this point, create the history item.
				HistoryItem *item = new HistoryItem();
				item->newIntegrationRange(IntegrationType::ANY_TO_MZ, m_mzStart,
						m_mzEnd);
				m_history.appendHistoryItem(item);

				// qDebug() << __FILE__ << __LINE__
				//<< "New history item:" << item->asText(false);
			}

			if(!qIsNaN(m_dtStart) && !qIsNaN(m_dtEnd))
			{
				// There was drift time integration asked for.

				// qDebug() << __FILE__ << __LINE__ << "Drfit time integration range:"
				//<< "start:" << m_dtStart << "end:" << m_dtEnd;

				if(m_dtEnd < m_dtStart)
				{
					double temp = m_dtEnd;
					m_dtEnd     = m_dtStart;
					m_dtStart   = temp;
				}

				if(m_dtStart < 0)
				{
					*errors += "The dtStart value is less than 0.";
				}

				// At this point, create the history item.
				HistoryItem *item = new HistoryItem();
				item->newIntegrationRange(IntegrationType::ANY_TO_DT, m_dtStart,
						m_dtEnd);
				m_history.appendHistoryItem(item);

				// qDebug() << __FILE__ << __LINE__
				//<< "New history item:" << item->asText(false);
			}

			// At this point, every thing is checked.


			// qDebug() << __FILE__ << __LINE__
			//<< "Finished history item creation. History:"
			//<< m_history.asText();

			if(errors->size() != errorStringSize)
				return false;

			return true;
		}


	bool
		CommandLineRunner::allInputFilesExist(int *existingFiles, int *missingFiles)
		{
			// Initialize the variables to meaningful values

			if(existingFiles != nullptr)
				*existingFiles = 0;

			if(missingFiles != nullptr)
				*missingFiles = 0;

			for(int iter = 0; iter < m_inFileNames.size(); ++iter)
			{
				QFileInfo fileInfo(m_inFileNames.at(iter));

				if(!fileInfo.exists())
				{
					if(missingFiles != nullptr)
						++(*missingFiles);
				}
				else
				{
					if(existingFiles != nullptr)
						++(*existingFiles);
				}
			}

			return (m_inFileNames.size() == *existingFiles);
		}


	bool
		CommandLineRunner::allInputFilesExist(QStringList *existingFiles,
				QStringList *missingFiles)
		{
			if(existingFiles == nullptr || missingFiles == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			bool allExist = true;

			for(int iter = 0; iter < m_inFileNames.size(); ++iter)
			{
				m_currentInputFileName = m_inFileNames.at(iter);
				QFileInfo fileInfo(m_currentInputFileName);

				if(!fileInfo.exists())
				{
					allExist = false;
					missingFiles->append(m_currentInputFileName);
				}
				else
					existingFiles->append(m_currentInputFileName);
			}

			return allExist;
		}


	QStringList
		CommandLineRunner::existingInputFiles()
		{
			QStringList existingFiles;

			for(int iter = 0; iter < m_inFileNames.size(); ++iter)
			{
				m_currentInputFileName = m_inFileNames.at(iter);
				QFileInfo fileInfo(m_currentInputFileName);

				if(fileInfo.exists())
					existingFiles.append(m_currentInputFileName);
			}

			return existingFiles;
		}


	bool
		CommandLineRunner::shouldRunGuiRightAway(QStringList *existingFiles,
				QString *    errors)
		{
			if(existingFiles == nullptr || errors == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// Test all the options and ascertain that we should run the GUI since the
			// parameters do not allow to start a meaningful non-GUI task.

			bool allStartEndValuesAreUnset = allStartEndValuesUnset();

			/*
				 qDebug() << __FILE__ << __LINE__
				 << "allStartEndValuesAreUnset: " << allStartEndValuesAreUnset
				 << "m_integration :" << m_integration
				 << "m_outFileName:" << m_outFileName
				 << "m_inFileNames:" << m_inFileNames;
				 */

			if(allStartEndValuesAreUnset && m_integration.isEmpty() &&
					m_outFileName.isEmpty() && !m_inFileNames.size())
			{
				// It is legitimate to have absolutely no argument to the command
				// line.
				// That means that we only want to start the program with the GUI.

				return true;
			}

			if(allStartEndValuesAreUnset && m_integration.isEmpty() &&
					m_outFileName.isEmpty() && m_inFileNames.size())
			{
				// Only input file names were provided. Check for the existence of the
				// files. Make a report for the files that are provided in the command
				// line but that are not found on disk.

				QStringList missingFiles;

				allInputFilesExist(existingFiles, &missingFiles);

				// Set, for each missing file, a corresponding item in the errors
				// string.
				for(int iter = 0; iter < missingFiles.size(); ++iter)
					*errors += QString("File %1 in input does not exist.\n")
						.arg(missingFiles.at(iter));

				return true;
			}

			return false;
		}


	bool
		CommandLineRunner::run(QStringList *existingFiles, QString *errors,
				bool *shouldRunGui)
		{
			if(existingFiles == nullptr || errors == nullptr || shouldRunGui == nullptr)
				qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

			// Before anything, if the arguments contain as script file name, then
			// return should run gui = true.
			if(!m_scriptFileName.isEmpty())
			{
				*shouldRunGui = true;
				return true;
			}

			// This command returns false if there was at least one error in the
			// command line parameters that impedes to launch the program.  Upon
			// returning, shouldRunGui should have been set to the proper value.

			if(shouldRunGuiRightAway(existingFiles, errors))
			{
				*shouldRunGui = true;
				return true;
			}

			// We'll check that size again to see if there were errors.
			int errorStringSize = errors->size();

			// Is the output to be sent to the standard output or not.
			m_isStdOut = (m_outFileName == "stdout" ? true : false);

			if(!m_integration.isEmpty())
			{
				// One integration process has been asked for. The verification of the
				// input file parameters needs to be more stringent. All need to be
				// found on disk.

				if(!m_inFileNames.size())
				{
					*errors += "There are no files in input.\n";

					*shouldRunGui = false;
				}

				// Now check that the files can all be opened.
				QStringList missingFiles;

				if(!allInputFilesExist(existingFiles, &missingFiles))
				{
					for(int iter = 0; iter < missingFiles.size(); ++iter)
						*errors += QString("File %1 in input does not exist.\n")
							.arg(missingFiles.at(iter));

					*shouldRunGui = false;
				}

				// If an integration operation is required, an output file is needed.
				if(m_outFileName.isEmpty())
				{
					*errors += "The output file name is empty.\n";

					*shouldRunGui = false;
				}

				// Now some special logic: if the m_subtract flag is set, then that
				// means that there must be at least two input file names.
				if(m_subtract)
				{
					if(m_inFileNames.size() < 2)
						*errors += "Only one file in input for a subtract integration.\n";
				}

				// Create all the History items required to document the Start/End
				// parameters that were set by the user.

				bool result = createHistoryItems(errors);
				if(!result)
				{
					*shouldRunGui = false;

					return false;
				}

				// At this point, every thing is checked.

				if(errors->size() != errorStringSize)
				{
					*shouldRunGui = false;

					return false;
				}

				// Finally launch the process.

				if(m_integration == "RT_ARBITRARY")
				{
					if(m_subtract)
						return runSubtract_RT_ARBITRARY(errors);
					else
						return runSum_RT_ARBITRARY(errors);
				}
				else if(m_integration == "MZ_ARBITRARY")
				{
					if(m_subtract)
						return runSubtract_MZ_ARBITRARY(errors);
					else
						return runSum_MZ_ARBITRARY(errors);
				}
				else if(m_integration == "DT_ARBITRARY")
				{
					if(m_subtract)
						return runSubtract_DT_ARBITRARY(errors);
					else
						return runSum_DT_ARBITRARY(errors);
				}
				else
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);
			}

			return false;
		}


	bool
		CommandLineRunner::runSum_RT_ARBITRARY(QString *errors)
		{
			qDebug() << __FILE__ << __LINE__ << "Entering runSum_RT_ARBITRARY().";

			// We'll check that size again to see if there were errors.

			int errorStringSize = errors->size();

			// We need this as a tag for success or not in the search for a value in
			// the map below.

			// Handy shortcut.
			int inFileCount = m_inFileNames.size();

			// The last result of the computation should be a TIC chromatogram.

			QMap<double, double> keyValMap;

			for(int fileCountIter = 0; fileCountIter < inFileCount; ++fileCountIter)
			{
				QVector<double> keyVector;
				QVector<double> valVector;

				m_currentInputFileName = m_inFileNames.at(fileCountIter);

				if(GlobIsSqlite3File(m_currentInputFileName))
				{
					MassSpecDataFileLoaderSqlite3 loader(m_currentInputFileName);

					// Make sure we connect the feedback signal/slot
					connect(&loader,
							&MassSpecDataFileLoaderSqlite3::setupProgressFeedbackSignal, this,
							&CommandLineRunner::setupProgressFeedback);

					connect(&loader,
							&MassSpecDataFileLoaderSqlite3::updateProgressFeedbackSignal, this,
							&CommandLineRunner::updateProgressFeedback);

					loader.streamedRtIntegration(m_history, &keyVector, &valVector);
				}
				else if(GlobIsMzmlFile(m_currentInputFileName))
				{
					MassSpecDataFileLoaderMzml loader(m_currentInputFileName);
					loader.loadData(m_history, IntegrationType::ANY_TO_RT,
							&keyVector, &valVector);
				}
				else // if(GlobIsXyFile(m_currentInputFileName))
				{
					// MassSpecDataFileLoaderXy loader(m_currentInputFileName);
					// loader.streamLoadData(&keyVector, &valVector);

					MassSpecDataFileFormatAnalyzer analyzer(m_currentInputFileName);

					int format = analyzer.canLoadData();
					if(format == MassSpecDataFileFormat::MASS_SPEC_DATA_FILE_FORMAT_XY)
					{
						MassSpecDataFileLoaderXy loader(m_currentInputFileName);
						loader.loadData(&keyVector, &valVector);
					}
				}

				// At this point the currently iterated file's data were loaded in
				// the
				//  vectors. We now need to sum the contents of the  vectors
				// to the map.

				// Sanity check.
				if(keyVector.size() != valVector.size())
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

				for(int iter = 0; iter < keyVector.size(); ++iter)
				{
					double key = keyVector.at(iter);
					double val = valVector.at(iter);

					double oldVal = keyValMap.value(key, qSNaN());

					if(!qIsNaN(oldVal))
					{
						// A value by the same key was found, make the increment.
						val += oldVal;
					}

					keyValMap.insert(key, val);
				}
			}

			if(errors->size() != errorStringSize)
			{
				return false;
			}

			int writtenLines = writeOutput(keyValMap);

			qDebug() << __FILE__ << __LINE__ << "Exiting runSum_RT_ARBITRARY().";

			if(writtenLines != -1)
				return true;

			return false;
		}


	bool
		CommandLineRunner::runSubtract_RT_ARBITRARY(QString *errors)
		{
			qDebug() << __FILE__ << __LINE__ << "Entering runSubtract_RT_ARBITRARY().";

			// We'll check that size again to see if there were errors.

			int errorStringSize = errors->size();

			// We need this as a tag for success or not in the search for a value in
			// the map below.

			// Handy shortcut.
			int inFileCount = m_inFileNames.size();

			// The last result of the computation should be a TIC chromatogram.
			// This time, we want to make a subtraction, that is for example, we
			// want
			// to
			// remove a background signal. In the command line the background
			// signal
			// is
			// the first file (fileCountIter == 0). So we need to set it apart.

			QMap<double, double> keyValMapBackground;
			QMap<double, double> keyValMap;

			for(int fileCountIter = 0; fileCountIter < inFileCount; ++fileCountIter)
			{
				QVector<double> keyVector;
				QVector<double> valVector;

				m_currentInputFileName = m_inFileNames.at(fileCountIter);

				if(GlobIsSqlite3File(m_currentInputFileName))
				{
					MassSpecDataFileLoaderSqlite3 loader(m_currentInputFileName);
					loader.streamedRtIntegration(m_history, &keyVector, &valVector);
				}
				else if(GlobIsMzmlFile(m_currentInputFileName))
				{
					MassSpecDataFileLoaderMzml loader(m_currentInputFileName);
					loader.loadData(m_history, IntegrationType::ANY_TO_RT,
							&keyVector, &valVector);
				}
				else if(GlobIsXyFile(m_currentInputFileName))
				{
					MassSpecDataFileLoaderXy loader(m_currentInputFileName);
					loader.loadData(&keyVector, &valVector);
				}

				// Sanity check.
				if(keyVector.size() != valVector.size())
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

				// Since we are in a subtraction mode, we need to have a special
				// processing for the first file's data. These data are the
				// background
				// data that we'll have to store in a specific map.

				for(int iter = 0; iter < keyVector.size(); ++iter)
				{
					double key = keyVector.at(iter);
					double val = valVector.at(iter);

					double oldVal = 0;

					// Theoretically, this should not be necessary, since we are
					// doing
					// that map insertion only once for fileCountIter == 0. But, in that
					// *first*
					// file, if there are two lines with the same key and another
					// value,
					// we can add the values for the same key. Nit-picking.
					if(fileCountIter == 0)
						oldVal = keyValMapBackground.value(key, qSNaN());
					else
						oldVal = keyValMap.value(key, qSNaN());

					if(!qIsNaN(oldVal))
					{
						// A value by the same key was found, make the increment.
						val += oldVal;
					}

					if(fileCountIter == 0)
						keyValMapBackground.insert(key, val);
					else
					{
						// Remeber we are subtracting file 1 from all the other
						// files.
						// This is the moment to do it.

						oldVal = keyValMapBackground.value(key, qSNaN());

						if(!qIsNaN(oldVal))
						{
							// There was some background signal for that key.
							// Remove
							// it.
							val -= oldVal;
						}

						// Finally, we can insert in the keyValMap the
						// background-reduced
						// value.
						keyValMap.insert(key, val);
					}
				}
			}

			if(errors->size() != errorStringSize)
			{
				return false;
			}

			int writtenLines = writeOutput(keyValMap);

			qDebug() << __FILE__ << __LINE__ << "Exiting runSubtract_RT_ARBITRARY().";

			if(writtenLines != -1)
				return true;

			return false;
		}


	bool
		CommandLineRunner::runSum_MZ_ARBITRARY(QString *errors)
		{
			qDebug() << __FILE__ << __LINE__ << "Entering runSum_MZ_ARBITRARY().";

			// We'll check that size again to see if there were errors.

			int errorStringSize = errors->size();

			// We need this as a tag for success or not in the search for a value in
			// the map below.

			// Handy shortcut.
			int inFileCount = m_inFileNames.size();

			// The last result of the computation should be a mass spectrum.

			QMap<double, double> keyValMap;

			for(int fileCountIter = 0; fileCountIter < inFileCount; ++fileCountIter)
			{
				QVector<double> keyVector;
				QVector<double> valVector;

				m_currentInputFileName = m_inFileNames.at(fileCountIter);

				if(GlobIsSqlite3File(m_currentInputFileName))
				{
					MassSpecDataFileLoaderSqlite3 loader(m_currentInputFileName);

					// Make sure we connect the feedback signal/slot
					connect(&loader,
							&MassSpecDataFileLoaderSqlite3::setupProgressFeedbackSignal, this,
							&CommandLineRunner::setupProgressFeedback);

					connect(&loader,
							&MassSpecDataFileLoaderSqlite3::updateProgressFeedbackSignal, this,
							&CommandLineRunner::updateProgressFeedback);

					loader.streamedMzIntegration(m_history, &keyVector, &valVector);
				}
				else if(GlobIsMzmlFile(m_currentInputFileName))
				{
					MassSpecDataFileLoaderMzml loader(m_currentInputFileName);
					loader.loadData(m_history, IntegrationType::ANY_TO_MZ,
							&keyVector, &valVector);
				}
				else if(GlobIsXyFile(m_currentInputFileName))
				{
					MassSpecDataFileLoaderXy loader(m_currentInputFileName);
					loader.loadData(&keyVector, &valVector);
				}

				// At this point the currently iterated file's data were loaded in
				// the
				//  vectors. We now need to sum the contents of the  vectors
				// to the map.

				// Sanity check.
				if(keyVector.size() != valVector.size())
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

				for(int iter = 0; iter < keyVector.size(); ++iter)
				{
					double key = keyVector.at(iter);
					double val = valVector.at(iter);

					double oldVal = keyValMap.value(key, qSNaN());

					if(!qIsNaN(oldVal))
					{
						// A value by the same key was found, make the increment.
						val += oldVal;
					}

					keyValMap.insert(key, val);
				}
			}

			if(errors->size() != errorStringSize)
			{
				return false;
			}

			int writtenLines = writeOutput(keyValMap);

			qDebug() << __FILE__ << __LINE__ << "Exiting runSum_MZ_ARBITRARY().";

			if(writtenLines != -1)
				return true;

			return true;
		}


	bool
		CommandLineRunner::runSubtract_MZ_ARBITRARY(QString *errors)
		{
			qDebug() << __FILE__ << __LINE__ << "Entering runSubtract_MZ_ARBITRARY().";

			// We'll check that size again to see if there were errors.

			int errorStringSize = errors->size();

			// We need this as a tag for success or not in the search for a value in
			// the map below.

			// Handy shortcut.
			int inFileCount = m_inFileNames.size();

			// The last result of the computation should be a mass spectrum.  This
			// time, we want to make a subtraction, that is for example, we want to
			// remove a background signal. In the command line the background signal is
			// the first file (fileCountIter == 0). So we need to set it apart.

			QMap<double, double> keyValMapBackground;
			QMap<double, double> keyValMap;

			for(int fileCountIter = 0; fileCountIter < inFileCount; ++fileCountIter)
			{
				QVector<double> keyVector;
				QVector<double> valVector;

				m_currentInputFileName = m_inFileNames.at(fileCountIter);

				if(GlobIsSqlite3File(m_currentInputFileName))
				{
					MassSpecDataFileLoaderSqlite3 loader(m_currentInputFileName);

					// Make sure we connect the feedback signal/slot
					connect(&loader,
							&MassSpecDataFileLoaderSqlite3::setupProgressFeedbackSignal, this,
							&CommandLineRunner::setupProgressFeedback);

					connect(&loader,
							&MassSpecDataFileLoaderSqlite3::updateProgressFeedbackSignal, this,
							&CommandLineRunner::updateProgressFeedback);

					int result =
						loader.streamedMzIntegration(m_history, &keyVector, &valVector);
					if(result == 0)
						continue;
					if(result == -1)
						return false;
				}
				else if(GlobIsMzmlFile(m_currentInputFileName))
				{
					MassSpecDataFileLoaderMzml loader(m_currentInputFileName);
					int                     result = loader.loadData(
							m_history, IntegrationType::ANY_TO_MZ, &keyVector, &valVector);
					if(result == 0)
						continue;
					if(result == -1)
						return false;
				}
				else if(GlobIsXyFile(m_currentInputFileName))
				{
					MassSpecDataFileLoaderXy loader(m_currentInputFileName);
					int result = loader.loadData(&keyVector, &valVector);
					if(result == 0)
						continue;
					if(result == -1)
						return false;
				}

				// Sanity check.
				if(keyVector.size() != valVector.size())
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

				// Since we are in a subtraction mode, we need to have a special
				// processing for the first file's data. These data are the
				// background data that we'll have to store in a specific map.

				for(int iter = 0; iter < keyVector.size(); ++iter)
				{
					double key = keyVector.at(iter);
					double val = valVector.at(iter);

					// qDebug() << __FILE__ << __LINE__
					//<< QString("iterating in key: %1 and val: %2")
					//.arg(key, 0, 'f', 12)
					//.arg(val, 0, 'f', 12);

					double oldVal = 0;

					if(fileCountIter == 0)
					{

						// This is the first file of the input files, and thus is
						// considered to be the background data file.


						// Theoretically, this should not be necessary, since we are
						// doing that map insertion only once for fileCountIter == 0.
						// But, in that *first* file, if there are two lines with the
						// same key and another value, we can add the values for the
						// same key.  Nit-picking.

						oldVal = keyValMapBackground.value(key, qSNaN());

						if(!qIsNaN(oldVal))
						{
							// A background value by the same key was found, make the
							// increment.

							val += oldVal;

							// qDebug()
							//<< __FILE__ << __LINE__
							//<< QString(
							//"background file: found oldVal: %1, new val: %2")
							//.arg(oldVal, 0, 'f', 12)
							//.arg(val, 0, 'f', 12);
						}
						else
						{
							// qDebug() << __FILE__ << __LINE__
							//<< QString("background file: add key %1 ; val %2")
							//.arg(key, 0, 'f', 12)
							//.arg(val, 0, 'f', 12);
						}

						keyValMapBackground.insert(key, val);
					}
					else
						// the file is not the background file (that is fileCountIter !=
						// 0)
					{

						// We are not iterating in key/val pairs belonging to the
						// background file. Search for a value in the background file,
						// to see if we have background signal to remove.

						oldVal = keyValMapBackground.value(key, qSNaN());

						// qDebug() << __FILE__ << __LINE__
						//<< QString("for key %1, searched for oldVal in the "
						//"background file ; got %2")
						//.arg(key, 0, 'f', 12)
						//.arg(oldVal, 0, 'f', 12);

						if(!qIsNaN(oldVal))
						{
							// A background value by the same key was found, make the
							// increment.

							val -= oldVal;

							// qDebug()
							//<< __FILE__ << __LINE__
							//<< QString("Found background signal to subtract for "
							//"key %1: %2 ; new val: %3")
							//.arg(key, 0, 'f', 12)
							//.arg(oldVal, 0, 'f', 12)
							//.arg(val, 0, 'f', 12);
						}
						else
						{
							// qDebug()
							//<< __FILE__ << __LINE__
							//<< QString("Found NO background signal to subtract for "
							//"key %1: unchanged val: %2")
							//.arg(key, 0, 'f', 12)
							//.arg(val, 0, 'f', 12);
						}

						// Now that we have the denoised -- that is, the unbackgrounded
						// value, we can check if such a value already occurred in the
						// non-background data files.

						oldVal = keyValMap.value(key, qSNaN());

						if(!qIsNaN(oldVal))
						{
							// A value by the same key was found, make the
							// increment.

							val += oldVal;

							// qDebug() << __FILE__ << __LINE__
							//<< QString("Found previous REAL signal for "
							//"key %1: %2 ; new val: %3")
							//.arg(key, 0, 'f', 12)
							//.arg(oldVal, 0, 'f', 12)
							//.arg(val, 0, 'f', 12);
						}
						else
						{
							// qDebug() << __FILE__ << __LINE__
							//<< QString("Found NO previous REAL signal for "
							//"key %1: unchanged val: %2")
							//.arg(key, 0, 'f', 12)
							//.arg(val, 0, 'f', 12);
						}

						keyValMap.insert(key, val);
					}
					// End else (the file is not the background file.
				}
				// End of
				// for(int iter = 0; iter < keyVector.size(); ++iter)
			}
			// End of
			// for(int fileCountIter = 0; fileCountIter < inFileCount; ++fileCountIter)

			if(errors->size() != errorStringSize)
			{
				return false;
			}

			// At this point we can craft the string to send to the file.
			QList<double> keys = keyValMap.keys();

			int writtenLines = writeOutput(keyValMap);

			qDebug() << __FILE__ << __LINE__ << "Exiting runSubtract_MZ_ARBITRARY().";

			if(writtenLines != -1)
				return true;

			return false;
		}


	bool
		CommandLineRunner::runSum_DT_ARBITRARY(QString *errors)
		{
			qDebug() << __FILE__ << __LINE__ << "Entering runSum_DT_ARBITRARY().";

			// We'll check that size again to see if there were errors.

			int errorStringSize = errors->size();

			// We need this as a tag for success or not in the search for a value in
			// the map below.

			// Handy shortcut.
			int inFileCount = m_inFileNames.size();

			// The last result of the computation should be a drift spectrum.

			QMap<double, double> keyValMap;

			for(int fileCountIter = 0; fileCountIter < inFileCount; ++fileCountIter)
			{
				QVector<double> keyVector;
				QVector<double> valVector;

				m_currentInputFileName = m_inFileNames.at(fileCountIter);

				if(GlobIsSqlite3File(m_currentInputFileName))
				{
					MassSpecDataFileLoaderSqlite3 loader(m_currentInputFileName);
					loader.streamedDtIntegration(m_history, &keyVector, &valVector);
				}
				else if(GlobIsMzmlFile(m_currentInputFileName))
				{
					MassSpecDataFileLoaderMzml loader(m_currentInputFileName);
					loader.loadData(m_history, IntegrationType::ANY_TO_DT,
							&keyVector, &valVector);
				}
				else if(GlobIsXyFile(m_currentInputFileName))
				{
					MassSpecDataFileLoaderXy loader(m_currentInputFileName);
					loader.loadData(&keyVector, &valVector);
				}

				// At this point the currently iterated file's data were loaded in
				// the
				//  vectors. We now need to sum the contents of the  vectors
				// to the map.

				// Sanity check.
				if(keyVector.size() != valVector.size())
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

				for(int iter = 0; iter < keyVector.size(); ++iter)
				{
					double key = keyVector.at(iter);
					double val = valVector.at(iter);

					double oldVal = keyValMap.value(key, qSNaN());

					if(!qIsNaN(oldVal))
					{
						// A value by the same key was found, make the increment.
						val += oldVal;
					}

					keyValMap.insert(key, val);
				}
			}

			if(errors->size() != errorStringSize)
			{
				return false;
			}

			int writtenLines = writeOutput(keyValMap);

			qDebug() << __FILE__ << __LINE__ << "Exiting runSum_DT_ARBITRARY().";

			if(writtenLines != -1)
				return true;

			return false;
		}


	bool
		CommandLineRunner::runSubtract_DT_ARBITRARY(QString *errors)
		{
			qDebug() << __FILE__ << __LINE__ << "Entering runSubtract_DT_ARBITRARY().";

			// We'll check that size again to see if there were errors.

			int errorStringSize = errors->size();

			// We need this as a tag for success or not in the search for a value in
			// the map below.

			// Handy shortcut.
			int inFileCount = m_inFileNames.size();

			// The last result of the computation should be a drift spectrum.
			// This time, we want to make a subtraction, that is for example, we
			// want
			// to
			// remove a background signal. In the command line the background
			// signal
			// is
			// the first file (fileCountIter == 0). So we need to set it apart.

			QMap<double, double> keyValMapBackground;
			QMap<double, double> keyValMap;

			for(int fileCountIter = 0; fileCountIter < inFileCount; ++fileCountIter)
			{
				QVector<double> keyVector;
				QVector<double> valVector;

				QString m_currentInputFileName = m_inFileNames.at(fileCountIter);

				if(GlobIsSqlite3File(m_currentInputFileName))
				{
					MassSpecDataFileLoaderSqlite3 loader(m_currentInputFileName);
					loader.streamedDtIntegration(m_history, &keyVector, &valVector);
				}
				else if(GlobIsMzmlFile(m_currentInputFileName))
				{
					MassSpecDataFileLoaderMzml loader(m_currentInputFileName);
					loader.loadData(m_history, IntegrationType::ANY_TO_DT,
							&keyVector, &valVector);
				}
				else if(GlobIsXyFile(m_currentInputFileName))
				{
					MassSpecDataFileLoaderXy loader(m_currentInputFileName);
					loader.loadData(&keyVector, &valVector);
				}

				// Sanity check.
				if(keyVector.size() != valVector.size())
					qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

				// Since we are in a subtraction mode, we need to have a special
				// processing for the first file's data. These data are the
				// background
				// data that we'll have to store in a specific map.

				for(int iter = 0; iter < keyVector.size(); ++iter)
				{
					double key = keyVector.at(iter);
					double val = valVector.at(iter);

					double oldVal = 0;

					// Theoretically, this should not be necessary, since we are
					// doing
					// that map insertion only once for fileCountIter == 0. But, in that
					// *first*
					// file, if there are two lines with the same key and another
					// value,
					// we can add the values for the same key. Nit-picking.
					if(fileCountIter == 0)
						oldVal = keyValMapBackground.value(key, qSNaN());
					else
						oldVal = keyValMap.value(key, qSNaN());

					if(!qIsNaN(oldVal))
					{
						// A value by the same key was found, make the increment.
						val += oldVal;
					}

					if(fileCountIter == 0)
						keyValMapBackground.insert(key, val);
					else
					{
						// Remeber we are subtracting file 1 from all the other
						// files.
						// This is the moment to do it.

						oldVal = keyValMapBackground.value(key, qSNaN());

						if(!qIsNaN(oldVal))
						{
							// There was some background signal for that key.
							// Remove
							// it.
							val -= oldVal;
						}

						// Finally, we can insert in the keyValMap the
						// background-reduced
						// value.
						keyValMap.insert(key, val);
					}
				}
			}

			if(errors->size() != errorStringSize)
			{
				return false;
			}

			int writtenLines = writeOutput(keyValMap);

			qDebug() << __FILE__ << __LINE__ << "Exiting runSubtract_DT_ARBITRARY().";

			if(writtenLines != -1)
				return true;

			return false;
		}


	void
		CommandLineRunner::setupProgressFeedback(QString msg, int startValue,
				int endValue, int currentValue)
		{
			fputs(m_currentInputFileName.toUtf8().data(), stdout);
			fputs("\n", stdout);

			m_baseMessage = msg;
			m_baseMessage += QString(" - [ %1 -- %2 ]").arg(startValue).arg(endValue);

			fputs(msg.toUtf8().data(), stdout);
			fputs("\n\n\n", stdout);
		}


	void
		CommandLineRunner::updateProgressFeedback(QString msg, int currentValue)
		{
			// remove previous line
			fputs("                                                                    "
					"                 ",
					stdout);

			// Craft new line.
			QString localMsg =
				QString("\r%1 - progress: %2").arg(m_baseMessage).arg(currentValue);
			fputs(localMsg.toUtf8().data(), stdout);
		}


	int
		CommandLineRunner::writeOutput(const QMap<double, double> &keyValMap)
		{
			QList<double> keys = keyValMap.keys();

			QFile outputFile;
			bool  result = false;

			if(m_isStdOut)
				result = outputFile.open(stdout, QIODevice::WriteOnly);
			else
			{
				outputFile.setFileName(m_outFileName);
				result = outputFile.open(QFile::WriteOnly | QFile::Truncate);
			}

			if(!result)
				return -1;

			QTextStream outStream(&outputFile);

			outStream.setRealNumberNotation(QTextStream::FixedNotation);
			outStream.setRealNumberPrecision(12);

			int linesWritten = 0;

			for(int iter = 0; iter < keys.size(); ++iter)
			{
				double key = keys.at(iter);
				double val = keyValMap.value(key);

				outStream << key << " " << val << endl;
				++linesWritten;
			}

			return linesWritten;
		}


	QString
		CommandLineRunner::asText()
		{
			QString text = "integration: " + m_integration + "\n";
			text += "input files: " + m_inFileNames.join(" , ") + "\n";
			text += "output file: " + m_outFileName + "\n";
			text += "script file: " + m_scriptFileName + "\n";
			text += QString("rtStart: %1").arg(m_rtStart, 0, 'f', 10) + "\n";
			text += QString("rtEnd: %1").arg(m_rtEnd, 0, 'f', 10) + "\n";
			text += QString("mzStart: %1").arg(m_mzStart, 0, 'f', 10) + "\n";
			text += QString("mzEnd: %1").arg(m_mzEnd, 0, 'f', 10) + "\n";
			text += QString("dtStart: %1").arg(m_dtStart, 0, 'f', 10) + "\n";
			text += QString("dtEnd: %1").arg(m_dtEnd, 0, 'f', 10) + "\n";

			return text;
		}


} // namespace msXpertSuite
