/*   msXpertSuite - mass spectrometry software suite
 *   -----------------------------------------------
 *   Copyright(C) 2016 Filippo Rusconi
 *
 *   http://www.{site}.org
 *
 *   This file is part of the msXpertSuite project.
 *
 *   The msXpertSuite project is the successor of the massXpert project. This
 *   project now includes various independent modules:
 *   - massXpert, the original mass spectrometry environment software (greatly
 *     refactored code);
 *   - mineXpert, a powerful TIC chromatogram/mass/drift spectrum viewer/miner;
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


/////////////////////// Qt includes
#include <QDebug>

/////////////////////// Local includes
#include "MzRangeList.hpp"


namespace msXpertSuite
{

  MzRangeList::MzRangeList(const QString &mzList)
  {
    // Immediately fill-in the list with the values contained in the mzList
    // string that may either contain ','-separated values or a single value.
    // In this constructor we assume that the values in the mzList are sorted in
    // ascending order. We check that two consecutive values are not
    // identical and that the newly added values is indeed greater than the
    // previous. Otherwise qFatal() is called.

    if(mzList.isEmpty())
      return;

    QStringList mzStringList = mzList.split(',', QString::SkipEmptyParts);

    for(int iter = 0; iter < mzStringList.size(); ++iter)
      {
        bool ok = false;

        double value = mzStringList.at(iter).toDouble(&ok);

        if(!ok)
          qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

        if(last() == value || last() > value)
          qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

        append(value);
      }
  }

  MzRangeList::~MzRangeList()
  {
  }


  void
  MzRangeList::setName(const QString &name)
  {
    m_name = name;
  }


  int
  MzRangeList::insertAscOrder(double value)
  {
    // We need to check if the value is already in *this list. If so we just do
    // nothing and return the index of the preexsiting double value. Othewise we
    // need to insert the new double value in ascending order.

    if(isEmpty())
      {
        append(value);
        return 0;
      }

    if(value < first())
      {
        prepend(value);
        return 0;
      }

    if(value > last())
      {
        append(value);
        return size() - 1;
      }

    for(int iter = 1; iter < size(); ++iter)
      {
        if(value > at(iter))
          continue;
        else
          {
            insert(iter, value);
            return iter;
          }
      }
    // Because we already handled the situation for which we have value >
    // last(),
    // we should never attain this point.

    qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

    return -1;
  }


  int
  MzRangeList::insertAscOrder(const QString &string)
  {
    // Returns the number of items inserted and/or found.

    // 'string' might either be a single value or a ','-separated list.

    // qDebug() << __FILE__ << __LINE__
    // << "string: " << string;

    QStringList valueList = string.split(',', QString::SkipEmptyParts);

    if(valueList.isEmpty())
      return 0;

    // qDebug() << __FILE__ << __LINE__
    // << "valueList: " << valueList;

    int count = 0;

    for(int iter = 0; iter < valueList.size(); ++iter)
      {
        bool ok = false;

        double localValue = valueList.at(iter).toDouble(&ok);

        if(!ok)
          qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

        if(insertAscOrder(localValue) == -1)
          qFatal("Fatal error at %s@%d. Program aborted.", __FILE__, __LINE__);

        ++count;
      }

    return count;
  }


  int
  MzRangeList::combine(const QList<double> &doubleList)
  {
    // Return the count of items inserted in the list. If an error occurred,
    // then -1 is returned. The caller function must test this return
    // value.

    // We get a list of double_s and we need to combine them into *this list.
    // That is, for each double we only insert in *this list if it does not
    // exist already. Also insertion is performed in ascending order sorting.

    int listSize = doubleList.size();
    int count    = 0;

    for(int iter = 0; iter < listSize; ++iter)
      {
        if(insertAscOrder(doubleList.at(iter)) == -1)
          return -1;

        ++count;
      }

    return count;
  }


  QString *
  MzRangeList::text()
  {
    QString *text = new QString(m_name);
    text->append("\n");

    for(int iter = 0; iter < size(); ++iter)
      {
        text->append(QString("%1\n").arg(at(iter), 0, 'f', 6));
      }

    return text;
  }


} // namespace msXpertSuite
